/* NIOS-specific support for 32-bit ELF.
   Copyright (C) 1998, 1999 Free Software Foundation, Inc.

This file is part of BFD, the Binary File Descriptor library.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include "bfd.h"
#include "sysdep.h"
#include "libbfd.h"
#include "elf-bfd.h"
#include "elf/nios.h"

/* Forward declarations.  */
static bfd_reloc_status_type nios_elf_relocate_lo16_lo5
  PARAMS ((bfd *,  Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static bfd_reloc_status_type nios_elf_relocate_lo16_hi11
  PARAMS ((bfd *,  Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static bfd_reloc_status_type nios_elf_relocate_hi16_lo5
  PARAMS ((bfd *,  Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static bfd_reloc_status_type nios_elf_relocate_hi16_hi11
  PARAMS ((bfd *,  Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static bfd_reloc_status_type nios_elf_relocate_h_lo5
  PARAMS ((bfd *,  Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static bfd_reloc_status_type nios_elf_relocate_h_hi11
  PARAMS ((bfd *,  Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static bfd_reloc_status_type nios_elf_relocate_h_xlo5
  PARAMS ((bfd *,  Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static bfd_reloc_status_type nios_elf_relocate_h_xhi11
  PARAMS ((bfd *,  Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static bfd_reloc_status_type nios_elf_relocate_h_16
  PARAMS ((bfd *,  Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static bfd_reloc_status_type nios_elf_relocate_h_32
  PARAMS ((bfd *,  Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static bfd_reloc_status_type nios_elf_relocate_pcrel6
  PARAMS ((bfd *, asection *, Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static bfd_reloc_status_type nios_elf_relocate_pcrel8
  PARAMS ((bfd *, asection *, Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static bfd_reloc_status_type nios_elf_relocate_pcrel11
  PARAMS ((bfd *, asection *, Elf_Internal_Rela *, bfd_byte *, bfd_vma));
static reloc_howto_type * nios_reloc_type_lookup
  PARAMS ((bfd *abfd, bfd_reloc_code_real_type code));
static void nios_info_to_howto_rela 
  PARAMS ((bfd *, arelent *, Elf32_Internal_Rela *));
static boolean nios_elf_relocate_section 
  PARAMS ((bfd *, struct bfd_link_info *, bfd *, asection *, bfd_byte *, Elf_Internal_Rela *, Elf_Internal_Sym *, asection **));
static bfd_reloc_status_type nios_final_link_relocate
  PARAMS ((reloc_howto_type *, bfd *, asection *, bfd_byte *, Elf_Internal_Rela *, bfd_vma));
static boolean nios_elf_gc_sweep_hook
  PARAMS ((bfd *, struct bfd_link_info *, asection *, const Elf_Internal_Rela *));
static asection * nios_elf_gc_mark_hook
  PARAMS ((bfd *, struct bfd_link_info *, Elf_Internal_Rela *, struct elf_link_hash_entry *, Elf_Internal_Sym *));
static bfd_reloc_status_type nios_elf_prepare_pcrel_offset2 
  PARAMS ((bfd *, arelent *, struct symbol_cache_entry *, PTR, asection *, bfd *, char **));
static int elf32_nios_machine PARAMS ((bfd *));
static boolean nios_elf_object_p PARAMS ((bfd *));
static boolean nios_elf_set_private_flags PARAMS ((bfd *, flagword));
static boolean nios_elf_copy_private_bfd_data PARAMS ((bfd *, bfd *));
static boolean nios_elf_merge_private_bfd_data PARAMS ((bfd *, bfd *));
static boolean nios_elf_print_private_bfd_data PARAMS ((bfd *, PTR));

static reloc_howto_type nios_elf_howto_table [] =
{
  /* This reloc does nothing.  */
  HOWTO (R_NIOS_NONE,		/* type */
	 0,			/* rightshift */
	 2,			/* size (0 = byte, 1 = short, 2 = long) */
	 32,			/* bitsize */
	 false,			/* pc_relative */
	 0,			/* bitpos */
	 complain_overflow_bitfield, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_NONE",		/* name */
	 false,			/* partial_inplace */
	 0,			/* src_mask */
	 0,			/* dst_mask */
	 false),		/* pcrel_offset */

  /* A 32 bit absolute relocation.  */
  HOWTO (R_NIOS_32,		/* type */
	 0,			/* rightshift */
	 2,			/* size (0 = byte, 1 = short, 2 = long) */
	 32,			/* bitsize */
	 false,			/* pc_relative */
	 0,			/* bitpos */
	 complain_overflow_bitfield, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_32",		/* name */
	 false,			/* partial_inplace */
	 0x00000000,		/* src_mask */
	 0xffffffff,		/* dst_mask */
	 false),		/* pcrel_offset */

  /* A LO-16 5 bit absolute relocation.  */
  HOWTO (R_NIOS_LO16_LO5,	/* type */
	 0,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 5,			/* bitsize */
	 false,			/* pc_relative */
	 5,			/* bitpos */
	 complain_overflow_dont, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_LO16_LO5",	/* name */
	 false,			/* partial_inplace */
	 0x0000,		/* src_mask */
	 0x03e0,		/* dst_mask */
	 false),		/* pcrel_offset */

  /* A LO-16 top 11 bit absolute relocation.  */
  HOWTO (R_NIOS_LO16_HI11,	/* type */
	 5,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 11,			/* bitsize */
	 false,			/* pc_relative */
	 0,			/* bitpos */
	 complain_overflow_dont, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_LO16_HI11",/* name */
	 false,			/* partial_inplace */
	 0x0000,		/* src_mask */
	 0x07ff,		/* dst_mask */
	 false),		/* pcrel_offset */

  /* A HI-16 5 bit absolute relocation.  */
  HOWTO (R_NIOS_HI16_LO5,	/* type */
	 16,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 5,			/* bitsize */
	 false,			/* pc_relative */
	 5,			/* bitpos */
	 complain_overflow_dont, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_HI16_LO5",	/* name */
	 false,			/* partial_inplace */
	 0x0000,		/* src_mask */
	 0x03e0,		/* dst_mask */
	 false),		/* pcrel_offset */

  /* A HI-16 top 11 bit absolute relocation.  */
  HOWTO (R_NIOS_HI16_HI11,	/* type */
	 21,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 11,			/* bitsize */
	 false,			/* pc_relative */
	 0,			/* bitpos */
	 complain_overflow_dont, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_HI16_HI11",/* name */
	 false,			/* partial_inplace */
	 0x0000,		/* src_mask */
	 0x07ff,		/* dst_mask */
	 false),		/* pcrel_offset */


  /* A 6 bit relative relocation.  */
  HOWTO (R_NIOS_PCREL6,	/* type */
	 1,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 6,			/* bitsize */
	 true,			/* pc_relative */
	 5,			/* bitpos */
	 complain_overflow_bitfield, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_PCREL6",   /* name */
	 false,			/* partial_inplace */
	 0x0000,  		/* src_mask */
	 0x07e0,	        /* dst_mask */
	 true), 		/* pcrel_offset */

  /* An 8 bit relative relocation.  */
  HOWTO (R_NIOS_PCREL8,	/* type */
	 1,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 8,			/* bitsize */
	 true,			/* pc_relative */
	 0,			/* bitpos */
	 complain_overflow_bitfield, /* complain_on_overflow */
	 nios_elf_prepare_pcrel_offset2,	/* special_function */
	 "R_NIOS_PCREL8",   /* name */
	 false,			/* partial_inplace */
	 0x0000,  		/* src_mask */
	 0x00ff,	        /* dst_mask */
	 true), 		/* pcrel_offset */

  /* An 11 bit relative relocation.  */
  HOWTO (R_NIOS_PCREL11,	/* type */
	 1,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 11,			/* bitsize */
	 true,			/* pc_relative */
	 0,			/* bitpos */
	 complain_overflow_dont, /* complain_on_overflow */
	 nios_elf_prepare_pcrel_offset2,	/* special_function */
	 "R_NIOS_PCREL11",     /* name */
	 false,			/* partial_inplace */
	 0x0000,  		/* src_mask */
	 0x07ff,	        /* dst_mask */
	 true), 		/* pcrel_offset */

  /* A 16 bit absolute relocation.  */
  HOWTO (R_NIOS_16,		/* type */
	 0,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 16,			/* bitsize */
	 false,			/* pc_relative */
	 0,			/* bitpos */
	 complain_overflow_bitfield, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_16",		/* name */
	 false,			/* partial_inplace */
	 0x0000,		/* src_mask */
	 0xffff,		/* dst_mask */
	 false),		/* pcrel_offset */

  /* Low 5-bits of absolute relocation in halfwords.  */
  HOWTO (R_NIOS_H_LO5,	        /* type */
	 1,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 5,			/* bitsize */
	 false,			/* pc_relative */
	 5,			/* bitpos */
	 complain_overflow_dont, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_H_LO5",	/* name */
	 false,			/* partial_inplace */
	 0x0000,		/* src_mask */
	 0x03e0,		/* dst_mask */
	 false),		/* pcrel_offset */

  /* Top 11 bits of 16-bit absolute relocation in halfwords.  */
  HOWTO (R_NIOS_H_HI11,	/* type */
	 6,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 11,			/* bitsize */
	 false,			/* pc_relative */
	 5,			/* bitpos */
	 complain_overflow_dont, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_H_HI11",      /* name */
	 false,			/* partial_inplace */
	 0x0000,		/* src_mask */
	 0x07ff,		/* dst_mask */
	 false),		/* pcrel_offset */

  /* Low 5 bits of top 16-bits of 32-bit absolute relocation in halfwords.  */
  HOWTO (R_NIOS_H_XLO5,	/* type */
	 17,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 5,			/* bitsize */
	 false,			/* pc_relative */
	 5,			/* bitpos */
	 complain_overflow_dont, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_H_XLO5",	/* name */
	 false,			/* partial_inplace */
	 0x0000,		/* src_mask */
	 0x03e0,		/* dst_mask */
	 false),		/* pcrel_offset */

  /* Top 11 bits of top 16-bits of 32-bit absolute relocation in halfwords.  */
  HOWTO (R_NIOS_H_XHI11,	/* type */
	 22,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 11,			/* bitsize */
	 false,			/* pc_relative */
	 0,			/* bitpos */
	 complain_overflow_dont, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_H_XHI11",     /* name */
	 false,			/* partial_inplace */
	 0x0000,		/* src_mask */
	 0x07ff,		/* dst_mask */
	 false),		/* pcrel_offset */

  /* Half-word @h value */
  HOWTO (R_NIOS_H_16,	        /* type */
	 1,			/* rightshift */
	 1,			/* size (0 = byte, 1 = short, 2 = long) */
	 16,			/* bitsize */
	 false,			/* pc_relative */
	 0,			/* bitpos */
	 complain_overflow_bitfield, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_H_16",	/* name */
	 false,			/* partial_inplace */
	 0x0000,		/* src_mask */
	 0xffff,		/* dst_mask */
	 false),		/* pcrel_offset */

  /* Word @h value */
  HOWTO (R_NIOS_H_32,	        /* type */
	 1,			/* rightshift */
	 2,			/* size (0 = byte, 1 = short, 2 = long) */
	 32,			/* bitsize */
	 false,			/* pc_relative */
	 0,			/* bitpos */
	 complain_overflow_bitfield, /* complain_on_overflow */
	 bfd_elf_generic_reloc,	/* special_function */
	 "R_NIOS_H_32",        /* name */
	 false,			/* partial_inplace */
	 0x00000000,		/* src_mask */
	 0xffffffff,		/* dst_mask */
	 false),		/* pcrel_offset */
};

/* GNU extension to record C++ vtable hierarchy */
static reloc_howto_type nios_elf_vtinherit_howto =
  HOWTO (R_NIOS_GNU_VTINHERIT,   /* type */
         0,                     /* rightshift */
         2,                     /* size (0 = byte, 1 = short, 2 = long) */
         0,                     /* bitsize */
         false,                 /* pc_relative */
         0,                     /* bitpos */
         complain_overflow_dont, /* complain_on_overflow */
         NULL,                  /* special_function */
         "R_NIOS_GNU_VTINHERIT", /* name */
         false,                 /* partial_inplace */
         0,                     /* src_mask */
         0,                     /* dst_mask */
         false);                /* pcrel_offset */

  /* GNU extension to record C++ vtable member usage */
static reloc_howto_type nios_elf_vtentry_howto =
  HOWTO (R_NIOS_GNU_VTENTRY,     /* type */
         0,                     /* rightshift */
         2,                     /* size (0 = byte, 1 = short, 2 = long) */
         0,                     /* bitsize */
         false,                 /* pc_relative */
         0,                     /* bitpos */
         complain_overflow_dont, /* complain_on_overflow */
         _bfd_elf_rel_vtable_reloc_fn,  /* special_function */
         "R_NIOS_GNU_VTENTRY",   /* name */
         false,                 /* partial_inplace */
         0,                     /* src_mask */
         0,                     /* dst_mask */
         false);                /* pcrel_offset */




/* special routine to prepare pcrel to subtract 2 before using generic howto table logic
   which cannot encode offsets to subtract */
static bfd_reloc_status_type
nios_elf_prepare_pcrel_offset2 (abfd, reloc_entry, symbol, 
				 data, input_section, output_bfd, error_message)
     bfd *abfd;
     arelent *reloc_entry;
     struct symbol_cache_entry *symbol;
     PTR data;
     asection *input_section;
     bfd *output_bfd;
     char **error_message;
{
  bfd_reloc_status_type result;

  /* Kludge - we want to subtract 2 from the relative offset before placing
     in the instruction.  The problem is that this function gets called by
     both the assembler and the linker.  Subtracting from the addend allows
     us to avoid having to write a new bfd_elf_generic_reloc-type function.
     However, when assembling, if the relocation doesn't get resolved, we
     certainly don't want to change the addend as this is a permanent change
     which will affect the linker (i.e. we will again subtract 2 when we
     pass through here to link srecord format and we will be off by 2.
     Thus, we look at the return code from bfd_elf_generic_reloc to see
     whether to replace the 2 we took from the addend or not. */

  reloc_entry->addend -= 2;
  result = bfd_elf_generic_reloc (abfd, reloc_entry, symbol, data, input_section,
				  output_bfd, error_message);

  if (result != bfd_reloc_continue)
    reloc_entry->addend += 2;

  return result;
}

/* Handle a NIOS ELF HI16 LO5 reloc.  */

static bfd_reloc_status_type
nios_elf_relocate_hi16_lo5 (input_bfd, relhi, contents, value)
     bfd *input_bfd;
     Elf_Internal_Rela *relhi;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;

  insn = bfd_get_16 (input_bfd, contents + relhi->r_offset);

  value += relhi->r_addend;
  value = ((value >> 16) & 0x001f);
  value <<= 5;

  insn = ((insn & 0xfc1f) | value);

  bfd_put_16 (input_bfd, insn, contents + relhi->r_offset);
  return bfd_reloc_ok;

}

/* Handle a NIOS ELF H_XLO5 reloc.  */

static bfd_reloc_status_type
nios_elf_relocate_h_xlo5 (input_bfd, relhi, contents, value)
     bfd *input_bfd;
     Elf_Internal_Rela *relhi;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;

  insn = bfd_get_16 (input_bfd, contents + relhi->r_offset);

  value += relhi->r_addend;
  value = ((value >> 17) & 0x001f);
  value <<= 5;

  insn = ((insn & 0xfc1f) | value);

  bfd_put_16 (input_bfd, insn, contents + relhi->r_offset);
  return bfd_reloc_ok;

}

/* Handle a NIOS ELF HI16 HI11 reloc.  */

static bfd_reloc_status_type
nios_elf_relocate_hi16_hi11 (input_bfd, relhi, contents, value)
     bfd *input_bfd;
     Elf_Internal_Rela *relhi;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;

  insn = bfd_get_16 (input_bfd, contents + relhi->r_offset);

  value += relhi->r_addend;
  value = (value >> 21);

  insn = ((insn & 0xf800) | value);

  bfd_put_16 (input_bfd, insn, contents + relhi->r_offset);
  return bfd_reloc_ok;

}

/* Handle a NIOS ELF H_XHI11 reloc.  */

static bfd_reloc_status_type
nios_elf_relocate_h_xhi11 (input_bfd, relhi, contents, value)
     bfd *input_bfd;
     Elf_Internal_Rela *relhi;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;

  insn = bfd_get_16 (input_bfd, contents + relhi->r_offset);

  value += relhi->r_addend;
  value = (value >> 22);

  insn = ((insn & 0xf800) | value);

  bfd_put_16 (input_bfd, insn, contents + relhi->r_offset);
  return bfd_reloc_ok;

}

/* Handle a NIOS ELF LO16 LO5 reloc.  */

static bfd_reloc_status_type
nios_elf_relocate_lo16_lo5 (input_bfd, relhi, contents, value)
     bfd *input_bfd;
     Elf_Internal_Rela *relhi;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;

  insn = bfd_get_16 (input_bfd, contents + relhi->r_offset);

  value += relhi->r_addend;
  value = (value & 0x001f);

  insn = ((insn & 0xfc1f) | (value << 5));

  bfd_put_16 (input_bfd, insn, contents + relhi->r_offset);
  return bfd_reloc_ok;

}

/* Handle a NIOS ELF H_LO5 reloc.  */

static bfd_reloc_status_type
nios_elf_relocate_h_lo5 (input_bfd, relhi, contents, value)
     bfd *input_bfd;
     Elf_Internal_Rela *relhi;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;

  insn = bfd_get_16 (input_bfd, contents + relhi->r_offset);

  value += relhi->r_addend;
  value = ((value >> 1) & 0x001f);

  insn = ((insn & 0xfc1f) | (value << 5));

  bfd_put_16 (input_bfd, insn, contents + relhi->r_offset);
  return bfd_reloc_ok;

}

/* Handle a NIOS ELF LO16 HI11 reloc.  */

static bfd_reloc_status_type
nios_elf_relocate_lo16_hi11 (input_bfd, relhi, contents, value)
     bfd *input_bfd;
     Elf_Internal_Rela *relhi;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;

  insn = bfd_get_16 (input_bfd, contents + relhi->r_offset);

  value += relhi->r_addend;
  value = ((value & 0xffff) >> 5);

  insn = (insn & 0xf800) | value;

  bfd_put_16 (input_bfd, insn, contents + relhi->r_offset);
  return bfd_reloc_ok;

}

/* Handle a NIOS ELF H_HI11 reloc.  */

static bfd_reloc_status_type
nios_elf_relocate_h_hi11 (input_bfd, relhi, contents, value)
     bfd *input_bfd;
     Elf_Internal_Rela *relhi;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;

  insn = bfd_get_16 (input_bfd, contents + relhi->r_offset);

  value += relhi->r_addend;
  value = ((value & 0x1fffe) >> 6);

  insn = (insn & 0xf800) | value;

  bfd_put_16 (input_bfd, insn, contents + relhi->r_offset);
  return bfd_reloc_ok;

}

/* Perform the relocation for NIOS LDC instruction (PCREL 6) */
static bfd_reloc_status_type
nios_elf_relocate_pcrel6 (input_bfd, input_section, rello, contents, value)
     bfd *input_bfd;
     asection *input_section;
     Elf_Internal_Rela *rello;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;
  const bfd_arch_info_type *arch_info; 

  /* Grab the instruction */
  insn = bfd_get_16 (input_bfd, contents + rello->r_offset);

  value -= input_section->output_section->vma + input_section->output_offset;
  value -= rello->r_offset;
  value += rello->r_addend;
  value -= 4;

  /* insn uses relocation value times register size */
  arch_info = bfd_get_arch_info (input_bfd);
  if (arch_info->arch == bfd_mach_nios32)
    value = value >> 2;
  else
    value = value >> 1;

  if ((long) value > 0x3f || (long) value < 0)
    return bfd_reloc_overflow;

  value &= 0x3f;
  insn = insn & 0xf800;
  insn = insn | (value << 5);

  bfd_put_16 (input_bfd, insn, contents + rello->r_offset);

  return bfd_reloc_ok;
}

/* Perform the relocation for NIOS JMPC/CALLC instructions (PCREL 8) */
static bfd_reloc_status_type
nios_elf_relocate_pcrel8 (input_bfd, input_section, rello, contents, value)
     bfd *input_bfd;
     asection *input_section;
     Elf_Internal_Rela *rello;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;
  const bfd_arch_info_type *arch_info; 

  /* Grab the instruction */
  insn = bfd_get_16 (input_bfd, contents + rello->r_offset);

  value -= input_section->output_section->vma + input_section->output_offset;
  value -= rello->r_offset;
  value += rello->r_addend;
  value -= 2;

  value = value >> 1;

  if ((long) value > 0xff || (long) value < 0)
    return bfd_reloc_overflow;

  insn = insn & 0x701f;
  insn = insn | (value << 5);

  bfd_put_16 (input_bfd, insn, contents + rello->r_offset);

  return bfd_reloc_ok;
}


/* Perform the relocation for NIOS BR/BSR instructions (PCREL11) */
static bfd_reloc_status_type
nios_elf_relocate_pcrel11 (input_bfd, input_section, rello, contents, value)
     bfd *input_bfd;
     asection *input_section;
     Elf_Internal_Rela *rello;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;

  /* Grab the instruction */
  insn = bfd_get_16 (input_bfd, contents + rello->r_offset);

  value -= input_section->output_section->vma + input_section->output_offset;
  value -= rello->r_offset;
  value += rello->r_addend - 2;

  /* insn uses bottom 11 bits of relocation value times 2 */
  value = (value & 0xfff) >> 1;

  if (   ((long) value > 0x3ff)
      || ((unsigned long) value > 0x7ff))
    return bfd_reloc_overflow;

  insn = insn & 0xf800;
  insn = insn | value;

  bfd_put_16 (input_bfd, insn, contents + rello->r_offset);

  return bfd_reloc_ok;
}

/* Utility to actually perform an R_NIOS_H_32 reloc.  */

static bfd_reloc_status_type
nios_elf_relocate_h_32 (input_bfd, relhi, contents, value)
     bfd *input_bfd;
     Elf_Internal_Rela *relhi;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;

  value += relhi->r_addend;
  value >>= 1;

  bfd_put_32 (input_bfd, value, contents + relhi->r_offset);
  return bfd_reloc_ok;
}


/* Utility to actually perform an R_NIOS_H_16 reloc.  */

static bfd_reloc_status_type
nios_elf_relocate_h_16 (input_bfd, relhi, contents, value)
     bfd *input_bfd;
     Elf_Internal_Rela *relhi;
     bfd_byte *contents;
     bfd_vma value;
{
  bfd_vma insn;

  value += relhi->r_addend;
  value >>= 1;

  if (value > 0xffff)
    return bfd_reloc_overflow;

  bfd_put_16 (input_bfd, value, contents + relhi->r_offset);
  return bfd_reloc_ok;
}


/* Map BFD reloc types to NIOS ELF reloc types.  */

struct nios_reloc_map
{
  bfd_reloc_code_real_type bfd_reloc_val;
  unsigned int nios_reloc_val;
};

static const struct nios_reloc_map nios_reloc_map [] =
{
  { BFD_RELOC_NONE,                  R_NIOS_NONE },
  { BFD_RELOC_32,                    R_NIOS_32 },
  { BFD_RELOC_NIOS_LO16_LO5,        R_NIOS_LO16_LO5 },
  { BFD_RELOC_NIOS_LO16_HI11,       R_NIOS_LO16_HI11 },
  { BFD_RELOC_NIOS_HI16_LO5,        R_NIOS_HI16_LO5 },
  { BFD_RELOC_NIOS_HI16_HI11,       R_NIOS_HI16_HI11 },
  { BFD_RELOC_NIOS_PCREL6,          R_NIOS_PCREL6 },
  { BFD_RELOC_NIOS_PCREL8,          R_NIOS_PCREL8 },
  { BFD_RELOC_NIOS_PCREL11,         R_NIOS_PCREL11 },
  { BFD_RELOC_16,                    R_NIOS_16 },
  { BFD_RELOC_NIOS_H_LO5,           R_NIOS_H_LO5 },
  { BFD_RELOC_NIOS_H_HI11,          R_NIOS_H_HI11 },
  { BFD_RELOC_NIOS_H_XLO5,          R_NIOS_H_XLO5 },
  { BFD_RELOC_NIOS_H_XHI11,         R_NIOS_H_XHI11 },
  { BFD_RELOC_NIOS_H_16,            R_NIOS_H_16 },
  { BFD_RELOC_NIOS_H_32,            R_NIOS_H_32 },
  { BFD_RELOC_VTABLE_INHERIT,        R_NIOS_GNU_VTINHERIT },
  { BFD_RELOC_VTABLE_ENTRY,          R_NIOS_GNU_VTENTRY },
};

static reloc_howto_type *
nios_reloc_type_lookup (abfd, code)
     bfd * abfd ATTRIBUTE_UNUSED;
     bfd_reloc_code_real_type code;
{
  unsigned int i;
  switch (code)
    {
    case BFD_RELOC_NONE:
      return &nios_elf_howto_table[ (int) R_NIOS_NONE];
  
    case BFD_RELOC_32:
    case BFD_RELOC_CTOR:
      return &nios_elf_howto_table[ (int) R_NIOS_32];

    case BFD_RELOC_NIOS_H_32:
      return &nios_elf_howto_table[ (int) R_NIOS_H_32];

    case BFD_RELOC_16:
      return &nios_elf_howto_table[ (int) R_NIOS_16];

    case BFD_RELOC_NIOS_H_16:
      return &nios_elf_howto_table[ (int) R_NIOS_H_16];

    case BFD_RELOC_NIOS_LO16_LO5:
      return &nios_elf_howto_table[ (int) R_NIOS_LO16_LO5];

    case BFD_RELOC_NIOS_LO16_HI11:
      return &nios_elf_howto_table[ (int) R_NIOS_LO16_HI11];

    case BFD_RELOC_NIOS_HI16_LO5:
      return &nios_elf_howto_table[ (int) R_NIOS_HI16_LO5];

    case BFD_RELOC_NIOS_HI16_HI11:
      return &nios_elf_howto_table[ (int) R_NIOS_HI16_HI11];

    case BFD_RELOC_NIOS_H_LO5:
      return &nios_elf_howto_table[ (int) R_NIOS_H_LO5];

    case BFD_RELOC_NIOS_H_HI11:
      return &nios_elf_howto_table[ (int) R_NIOS_H_HI11];

    case BFD_RELOC_NIOS_H_XLO5:
      return &nios_elf_howto_table[ (int) R_NIOS_H_XLO5];

    case BFD_RELOC_NIOS_H_XHI11:
      return &nios_elf_howto_table[ (int) R_NIOS_H_XHI11];

    case BFD_RELOC_NIOS_PCREL6:
      return &nios_elf_howto_table[ (int) R_NIOS_PCREL6];

    case BFD_RELOC_NIOS_PCREL8:
      return &nios_elf_howto_table[ (int) R_NIOS_PCREL8];

    case BFD_RELOC_NIOS_PCREL11:
      return &nios_elf_howto_table[ (int) R_NIOS_PCREL11];

    case BFD_RELOC_VTABLE_INHERIT:
      return &nios_elf_vtinherit_howto;

    case BFD_RELOC_VTABLE_ENTRY:
      return &nios_elf_vtentry_howto;
    }

  return NULL;
}

/* Set the howto pointer for an NIOS ELF reloc.  */

static void
nios_info_to_howto_rela (abfd, cache_ptr, dst)
     bfd * abfd ATTRIBUTE_UNUSED;
     arelent * cache_ptr;
     Elf32_Internal_Rela * dst;
{
  unsigned int r_type;

  r_type = ELF32_R_TYPE (dst->r_info);
  switch (r_type)
    {
    case R_NIOS_GNU_VTINHERIT:
      cache_ptr->howto = & nios_elf_vtinherit_howto;
      break;

    case R_NIOS_GNU_VTENTRY:
      cache_ptr->howto = & nios_elf_vtentry_howto;
      break;

    default:
      cache_ptr->howto = & nios_elf_howto_table [r_type];
      break;

    }
}

/* Perform a single relocation.  By default we use the standard BFD
   routines, but a few relocs, we have to do them ourselves.  */

static bfd_reloc_status_type
nios_final_link_relocate (howto, input_bfd, input_section, contents, rel, relocation)
     reloc_howto_type *  howto;
     bfd *               input_bfd;
     asection *          input_section;
     bfd_byte *          contents;
     Elf_Internal_Rela * rel;
     bfd_vma             relocation;
{
  bfd_reloc_status_type r = bfd_reloc_ok;
  bfd_vma               x;
  bfd_signed_vma	srel;
  
  switch (howto->type)
    {
    default:
      r = _bfd_final_link_relocate (howto, input_bfd, input_section,
				    contents, rel->r_offset,
				    relocation, rel->r_addend);
    }

  return r;
}


/* Relocate an NIOS ELF section.
   There is some attempt to make this function usable for many architectures,
   both USE_REL and USE_RELA ['twould be nice if such a critter existed],
   if only to serve as a learning tool.

   The RELOCATE_SECTION function is called by the new ELF backend linker
   to handle the relocations for a section.

   The relocs are always passed as Rela structures; if the section
   actually uses Rel structures, the r_addend field will always be
   zero.

   This function is responsible for adjusting the section contents as
   necessary, and (if using Rela relocs and generating a relocateable
   output file) adjusting the reloc addend as necessary.

   This function does not have to worry about setting the reloc
   address or the reloc symbol index.

   LOCAL_SYMS is a pointer to the swapped in local symbols.

   LOCAL_SECTIONS is an array giving the section in the input file
   corresponding to the st_shndx field of each local symbol.

   The global hash table entry for the global symbols can be found
   via elf_sym_hashes (input_bfd).

   When generating relocateable output, this function must handle
   STB_LOCAL/STT_SECTION symbols specially.  The output symbol is
   going to be the section symbol corresponding to the output
   section, which means that the addend must be adjusted
   accordingly.  */

static boolean
nios_elf_relocate_section (output_bfd, info, input_bfd, input_section,
			   contents, relocs, local_syms, local_sections)
     bfd *                   output_bfd ATTRIBUTE_UNUSED;
     struct bfd_link_info *  info;
     bfd *                   input_bfd;
     asection *              input_section;
     bfd_byte *              contents;
     Elf_Internal_Rela *     relocs;
     Elf_Internal_Sym *      local_syms;
     asection **             local_sections;
{
  Elf_Internal_Shdr *           symtab_hdr;
  struct elf_link_hash_entry ** sym_hashes;
  Elf_Internal_Rela *           rel;
  Elf_Internal_Rela *           relend;

  symtab_hdr = & elf_tdata (input_bfd)->symtab_hdr;
  sym_hashes = elf_sym_hashes (input_bfd);
  relend     = relocs + input_section->reloc_count;

  for (rel = relocs; rel < relend; rel ++)
    {
      reloc_howto_type *           howto;
      unsigned long                r_symndx;
      Elf_Internal_Sym *           sym;
      asection *                   sec;
      struct elf_link_hash_entry * h;
      bfd_vma                      relocation;
      bfd_reloc_status_type        r;
      const char *                 name = NULL;
      int                          r_type;
      
      r_type = ELF32_R_TYPE (rel->r_info);
      
      if (   r_type == R_NIOS_GNU_VTINHERIT
	  || r_type == R_NIOS_GNU_VTENTRY)
	continue;
      
      r_symndx = ELF32_R_SYM (rel->r_info);

      if (info->relocateable)
	{
	  /* This is a relocateable link.  We don't have to change
             anything, unless the reloc is against a section symbol,
             in which case we have to adjust according to where the
             section symbol winds up in the output section.  */
	  if (r_symndx < symtab_hdr->sh_info)
	    {
	      sym = local_syms + r_symndx;
	      
	      if (ELF_ST_TYPE (sym->st_info) == STT_SECTION)
		{
		  sec = local_sections [r_symndx];
		  rel->r_addend += sec->output_offset + sym->st_value;
		}
	    }

	  continue;
	}

      /* This is a final link.  */
      howto  = nios_elf_howto_table + ELF32_R_TYPE (rel->r_info);
      h      = NULL;
      sym    = NULL;
      sec    = NULL;
      
      if (r_symndx < symtab_hdr->sh_info)
	{
	  sym = local_syms + r_symndx;
	  sec = local_sections [r_symndx];
	  relocation = (sec->output_section->vma
			+ sec->output_offset
			+ sym->st_value);
	  
	  name = bfd_elf_string_from_elf_section
	    (input_bfd, symtab_hdr->sh_link, sym->st_name);
	  name = (name == NULL) ? bfd_section_name (input_bfd, sec) : name;
#if 0
	  fprintf (stderr, "local: sec: %s, sym: %s (%d), value: %x + %x + %x addend %x\n",
		   sec->name, name, sym->st_name,
		   sec->output_section->vma, sec->output_offset,
		   sym->st_value, rel->r_addend);
#endif
	}
      else
	{
	  h = sym_hashes [r_symndx];
	  
	  while (h->root.type == bfd_link_hash_indirect
		 || h->root.type == bfd_link_hash_warning)
	    h = (struct elf_link_hash_entry *) h->root.u.i.link;

	  name = h->root.root.string;
	  
	  if (h->root.type == bfd_link_hash_defined
	      || h->root.type == bfd_link_hash_defweak)
	    {
	      sec = h->root.u.def.section;
	      relocation = (h->root.u.def.value
			    + sec->output_section->vma
			    + sec->output_offset);
#if 0
	      fprintf (stderr,
		       "defined: sec: %s, name: %s, value: %x + %x + %x gives: %x\n",
		       sec->name, name, h->root.u.def.value,
		       sec->output_section->vma, sec->output_offset, relocation);
#endif
	    }
	  else if (h->root.type == bfd_link_hash_undefweak)
	    {
#if 0
	      fprintf (stderr, "undefined: sec: %s, name: %s\n",
		       sec->name, name);
#endif
	      relocation = 0;
	    }
	  else
	    {
	      if (! ((*info->callbacks->undefined_symbol)
		     (info, h->root.root.string, input_bfd,
		      input_section, rel->r_offset)))
		return false;
#if 0
	      fprintf (stderr, "unknown: name: %s\n", name);
#endif
	      relocation = 0;
	    }
	}

      switch (r_type)
	{
	case R_NIOS_LO16_LO5:
	  r = nios_elf_relocate_lo16_lo5 (input_bfd, rel, contents, relocation);
	  break;
	case R_NIOS_LO16_HI11:
	  r = nios_elf_relocate_lo16_hi11 (input_bfd, rel, contents, relocation);
	  break;
	case R_NIOS_HI16_LO5:
	  r = nios_elf_relocate_hi16_lo5 (input_bfd, rel, contents, relocation);
	  break;
	case R_NIOS_HI16_HI11:
	  r = nios_elf_relocate_hi16_hi11 (input_bfd, rel, contents, relocation);
	  break;
	case R_NIOS_H_LO5:
	  r = nios_elf_relocate_h_lo5 (input_bfd, rel, contents, relocation);
	  break;
	case R_NIOS_H_HI11:
	  r = nios_elf_relocate_h_hi11 (input_bfd, rel, contents, relocation);
	  break;
	case R_NIOS_H_XLO5:
	  r = nios_elf_relocate_h_xlo5 (input_bfd, rel, contents, relocation);
	  break;
	case R_NIOS_H_XHI11:
	  r = nios_elf_relocate_h_xhi11 (input_bfd, rel, contents, relocation);
	  break;
	case R_NIOS_H_16:
	  r = nios_elf_relocate_h_16 (input_bfd, rel, contents, relocation);
	  break;
	case R_NIOS_H_32:
	  r = nios_elf_relocate_h_32 (input_bfd, rel, contents, relocation);
	  break;
	case R_NIOS_PCREL6:
	  r = nios_elf_relocate_pcrel6 (input_bfd, input_section, rel, 
					 contents, relocation);
	  break;
	case R_NIOS_PCREL8:
	  r = nios_elf_relocate_pcrel8 (input_bfd, input_section, rel, 
					 contents, relocation);
	  break;
	case R_NIOS_PCREL11:
	  r = nios_elf_relocate_pcrel11 (input_bfd, input_section, rel, 
					  contents, relocation);
	  break;
	default:
	  r = nios_final_link_relocate (howto, input_bfd, input_section,
					 contents, rel, relocation);
	  break;
	}

      if (r != bfd_reloc_ok)
	{
	  const char * msg = (const char *) NULL;

	  switch (r)
	    {
	    case bfd_reloc_overflow:
	      r = info->callbacks->reloc_overflow
		(info, name, howto->name, (bfd_vma) 0,
		 input_bfd, input_section, rel->r_offset);
	      break;
	      
	    case bfd_reloc_undefined:
	      r = info->callbacks->undefined_symbol
		(info, name, input_bfd, input_section, rel->r_offset);
	      break;
	      
	    case bfd_reloc_outofrange:
	      msg = _("internal error: out of range error");
	      break;

	    case bfd_reloc_notsupported:
	      msg = _("internal error: unsupported relocation error");
	      break;

	    case bfd_reloc_dangerous:
	      msg = _("internal error: dangerous relocation");
	      break;

	    default:
	      msg = _("internal error: unknown error");
	      break;
	    }

	  if (msg)
	    r = info->callbacks->warning
	      (info, msg, name, input_bfd, input_section, rel->r_offset);

	  if (! r)
	    return false;
	}
    }

  return true;
}

/* Return the section that should be marked against GC for a given
   relocation.  */

static asection *
nios_elf_gc_mark_hook (abfd, info, rel, h, sym)
     bfd *                        abfd;
     struct bfd_link_info *       info ATTRIBUTE_UNUSED;
     Elf_Internal_Rela *          rel;
     struct elf_link_hash_entry * h;
     Elf_Internal_Sym *           sym;
{
  if (h != NULL)
    {
      switch (ELF32_R_TYPE (rel->r_info))
	{
	case R_NIOS_GNU_VTINHERIT:
	case R_NIOS_GNU_VTENTRY:
	  break;

	default:
	  switch (h->root.type)
	    {
	    case bfd_link_hash_defined:
	    case bfd_link_hash_defweak:
	      return h->root.u.def.section;

	    case bfd_link_hash_common:
	      return h->root.u.c.p->section;

	    default:
	      break;
	    }
	}
    }
  else
    {
      if (!(elf_bad_symtab (abfd)
	    && ELF_ST_BIND (sym->st_info) != STB_LOCAL)
	  && ! ((sym->st_shndx <= 0 || sym->st_shndx >= SHN_LORESERVE)
		&& sym->st_shndx != SHN_COMMON))
	{
	  return bfd_section_from_elf_index (abfd, sym->st_shndx);
	}
    }

  return NULL;
}

/* Update the got entry reference counts for the section being removed.  */

static boolean
nios_elf_gc_sweep_hook (abfd, info, sec, relocs)
     bfd *                     abfd ATTRIBUTE_UNUSED;
     struct bfd_link_info *    info ATTRIBUTE_UNUSED;
     asection *                sec ATTRIBUTE_UNUSED;
     const Elf_Internal_Rela * relocs ATTRIBUTE_UNUSED;
{
  return true;
}

/* Look through the relocs for a section during the first phase.
   Since we don't do .gots or .plts, we just need to consider the
   virtual table relocs for gc.  */
 
static boolean
nios_elf_check_relocs (abfd, info, sec, relocs)
     bfd *abfd;
     struct bfd_link_info *info;
     asection *sec;
     const Elf_Internal_Rela *relocs;
{
  Elf_Internal_Shdr *symtab_hdr;
  struct elf_link_hash_entry **sym_hashes, **sym_hashes_end;
  const Elf_Internal_Rela *rel;
  const Elf_Internal_Rela *rel_end;
 
  if (info->relocateable)
    return true;
 
  symtab_hdr = &elf_tdata (abfd)->symtab_hdr;
  sym_hashes = elf_sym_hashes (abfd);
  sym_hashes_end = sym_hashes + symtab_hdr->sh_size/sizeof(Elf32_External_Sym);
  if (!elf_bad_symtab (abfd))
    sym_hashes_end -= symtab_hdr->sh_info;
 
  rel_end = relocs + sec->reloc_count;
  for (rel = relocs; rel < rel_end; rel++)
    {
      struct elf_link_hash_entry *h;
      unsigned long r_symndx;
 
      r_symndx = ELF32_R_SYM (rel->r_info);
      if (r_symndx < symtab_hdr->sh_info)
        h = NULL;
      else
        h = sym_hashes[r_symndx - symtab_hdr->sh_info];
 
      switch (ELF32_R_TYPE (rel->r_info))
        {
        /* This relocation describes the C++ object vtable hierarchy.
           Reconstruct it for later use during GC.  */
        case R_NIOS_GNU_VTINHERIT:
          if (!_bfd_elf32_gc_record_vtinherit (abfd, sec, h, rel->r_offset))
            return false;
          break;
 
        /* This relocation describes which C++ vtable entries are actually
           used.  Record for later use during GC.  */
        case R_NIOS_GNU_VTENTRY:
          if (!_bfd_elf32_gc_record_vtentry (abfd, sec, h, rel->r_addend))
            return false;
          break;
        }
    }
 
  return true;
}
/* Return the MACH for an e_flags value.  */

static int
elf32_nios_machine (abfd)
     bfd *abfd;
{
  switch (elf_elfheader (abfd)->e_flags & EF_NIOS_CPU_MASK)
    {
    case EF_NIOS_CPU_16:  return bfd_mach_nios16;
    case EF_NIOS_CPU_32:  return bfd_mach_nios32;
    }

  return bfd_mach_nios16;
}


/* Function to set the ELF flag bits */

static boolean
nios_elf_set_private_flags (abfd, flags)
     bfd *abfd;
     flagword flags;
{
  elf_elfheader (abfd)->e_flags = flags;
  elf_flags_init (abfd) = true;
  return true;
}

/* Copy backend specific data from one object module to another */
static boolean
nios_elf_copy_private_bfd_data (ibfd, obfd)
     bfd *ibfd;
     bfd *obfd;
{
  if (bfd_get_flavour (ibfd) != bfd_target_elf_flavour
      || bfd_get_flavour (obfd) != bfd_target_elf_flavour)
    return true;

  BFD_ASSERT (!elf_flags_init (obfd)
	      || elf_elfheader (obfd)->e_flags == elf_elfheader (ibfd)->e_flags);

  elf_elfheader (obfd)->e_flags = elf_elfheader (ibfd)->e_flags;
  elf_flags_init (obfd) = true;
  return true;
}

/* Merge backend specific data from an object file to the output
   object file when linking */
static boolean
nios_elf_merge_private_bfd_data (ibfd, obfd)
     bfd *ibfd;
     bfd *obfd;
{
  flagword old_flags, old_partial;
  flagword new_flags, new_partial;
  boolean error = false;
  char new_opt[80];
  char old_opt[80];

  new_opt[0] = old_opt[0] = '\0';
  new_flags = elf_elfheader (ibfd)->e_flags;
  old_flags = elf_elfheader (obfd)->e_flags;

#ifdef DEBUG
  (*_bfd_error_handler) ("old_flags = 0x%.8lx, new_flags = 0x%.8lx, init = %s, filename = %s",
			 old_flags, new_flags, elf_flags_init (obfd) ? "yes" : "no",
			 bfd_get_filename (ibfd));
#endif

  if (!elf_flags_init (obfd))			/* First call, no flags set */
    {
      elf_flags_init (obfd) = true;
      elf_elfheader (obfd)->e_flags = new_flags;
    }

  else if (new_flags == old_flags)		/* Compatible flags are ok */
    ;

  else						/* Possibly incompatible flags */
    {
      /* Warn if different cpu is used (allow a specific cpu to override
	 the generic cpu).  */
      new_partial = (new_flags & EF_NIOS_CPU_MASK);
      old_partial = (old_flags & EF_NIOS_CPU_MASK);
      if (new_partial == old_partial)
	;

      else
	{
	  switch (new_partial)
	    {
	    default:		        strcat (new_opt, " -m16");  break;
	    case EF_NIOS_CPU_16:  strcat (new_opt, " -m16");  break;
	    case EF_NIOS_CPU_32:  strcat (new_opt, " -m32");  break;
	    }

	  switch (old_partial)
	    {
	    default:		        strcat (old_opt, " -m16");  break;
	    case EF_NIOS_CPU_16:  strcat (old_opt, " -m16");  break;
	    case EF_NIOS_CPU_32:  strcat (old_opt, " -m32");  break;
	    }
	}
      
      /* Print out any mismatches from above */
      if (new_opt[0])
	{
	  error = true;
	  (*_bfd_error_handler)
	    (_("%s: compiled with %s and linked with modules compiled with %s"),
	     bfd_get_filename (ibfd), new_opt, old_opt);
	}

      new_flags &= ~ EF_NIOS_ALL_FLAGS;
      old_flags &= ~ EF_NIOS_ALL_FLAGS;

      /* Warn about any other mismatches */
      if (new_flags != old_flags)
	{
	  error = true;
	  (*_bfd_error_handler)
	    (_("%s: uses different e_flags (0x%lx) fields than previous modules (0x%lx)"),
	     bfd_get_filename (ibfd), (long)new_flags, (long)old_flags);
	}
    }

  if (error)
    bfd_set_error (bfd_error_bad_value);

  return !error;
}


boolean
nios_elf_print_private_bfd_data (abfd, ptr)
     bfd *abfd;
     PTR ptr;
{
  FILE *file = (FILE *) ptr;
  flagword flags;

  BFD_ASSERT (abfd != NULL && ptr != NULL);

  /* Print normal ELF private data.  */
  _bfd_elf_print_private_bfd_data (abfd, ptr);

  flags = elf_elfheader (abfd)->e_flags;
  fprintf (file, _("private flags = 0x%lx:"), (long)flags);

  switch (flags & EF_NIOS_CPU_MASK)
    {
    default:							break;
    case EF_NIOS_CPU_16:  fprintf (file, " -m16");	break;
    case EF_NIOS_CPU_32:  fprintf (file, " -m32");	break;
    }

  fputc ('\n', file);
}

static
boolean
nios_elf_object_p (abfd)
     bfd *abfd;
{
  /* Irix 5 and 6 is broken.  Object file symbol tables are not always
     sorted correctly such that local symbols precede global symbols,
     and the sh_info field in the symbol table is not always right.  */
  elf_bad_symtab (abfd) = true;

  bfd_default_set_arch_mach (abfd, bfd_arch_nios,
			     elf32_nios_machine (abfd));
  return true;
}

/* The final processing done just before writing out a NIOS ELF object
   file.  This gets the NIOS architecture right based on the machine
   number.  */

/*ARGSUSED*/
void
nios_elf_final_write_processing (abfd, linker)
     bfd *abfd;
     boolean linker ATTRIBUTE_UNUSED;
{
  /* empty */
}


#define ELF_ARCH		bfd_arch_nios
#define ELF_MACHINE_CODE	EM_CYGNUS_NIOS
#define ELF_MAXPAGESIZE		0x1000

#define TARGET_LITTLE_SYM          bfd_elf32_nios_vec
#define TARGET_LITTLE_NAME		"elf32-nios"

#define elf_info_to_howto_rel			NULL
#define elf_info_to_howto			nios_info_to_howto_rela
#define elf_backend_relocate_section		nios_elf_relocate_section
#define elf_backend_gc_mark_hook		nios_elf_gc_mark_hook
#define elf_backend_gc_sweep_hook		nios_elf_gc_sweep_hook
#define elf_backend_check_relocs                nios_elf_check_relocs
#define elf_backend_object_p		        nios_elf_object_p
#define elf_backend_final_write_processing      nios_elf_final_write_processing

#define elf_backend_can_gc_sections		1

#define bfd_elf32_bfd_reloc_type_lookup		nios_reloc_type_lookup
#define bfd_elf32_bfd_set_private_flags		nios_elf_set_private_flags
#define bfd_elf32_bfd_copy_private_bfd_data	nios_elf_copy_private_bfd_data
#define bfd_elf32_bfd_merge_private_bfd_data	nios_elf_merge_private_bfd_data
#define bfd_elf32_bfd_print_private_bfd_data	nios_elf_print_private_bfd_data

#include "elf32-target.h"
