/* {{{ Target definitions for the Altera Nios chip.

   Copyright (C) 1999, 2000 Free Software Foundation, Inc.
   Contributed by Cygnus Solutions, a Red Hat company

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2, or (at your option) any later version.

GNU CC is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along
with GNU CC; see the file COPYING.  If not, write to the Free Software
Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/* }}} */

/* {{{ General purpose macro and data structure definitions.  */

/* Set up System V.4 (aka ELF) defaults.  */
#include "elfos.h"

/* Forward structure declarations for use in prototypes.  */
#ifdef BUFSIZ		/* stdio.h has been included, ok to use FILE * */
#define STDIO_PROTO(ARGS) PROTO(ARGS)
#else
#define STDIO_PROTO(ARGS) ()
#endif

/* Opcode table used in nios_output_insn. These are two operand opcodes
   that share the same operand formats: reg,reg or reg,imm5.  */
enum nios_2op_opcodes {
    NIOS_ADD = 0,
    NIOS_AND,
    NIOS_ASR,
    NIOS_CMP,
    NIOS_LSL,
    NIOS_LSR,
    NIOS_MOV,
    NIOS_OR,
    NIOS_SUB,
    NIOS_XOR,

    NIOS_LAST_2OP_OPCODE	/* End of list marker, do not remove.  */
};


/* Opcode names corresponding to the opcode table above. Each entry in the
   table consists of two strings: the operand name when using reg,reg
   operands and the operand name when using reg,imm5 operands.  */
extern char *nios_2op_opcode_names[][2];

#define NIOS_2OP_OPCODE_NAMES	\
{				\
  {"add",	"addi"},	\
  {"and",	"and"},		\
  {"asr",	"asri"},	\
  {"cmp",	"cmpi"},	\
  {"lsl",	"lsli"},	\
  {"lsr",	"lsri"},	\
  {"mov",	"movi"},	\
  {"or",	"or"},		\
  {"sub",	"subi"},	\
  {"xor",	"xor"},		\
  {"",		""}		\
};


/* Opcode table used in nios_output_user_opcode. */
enum nios_user_opcodes {
    NIOS_USER_OPCODE_MUL = 0,
    NIOS_USER_OPCODE_DIV,
    NIOS_USER_OPCODE_UDIV,
    NIOS_USER_OPCODE_MOD,
    NIOS_USER_OPCODE_UMOD,
    NIOS_USER_OPCODE_FADD,
    NIOS_USER_OPCODE_FSUB,
    NIOS_USER_OPCODE_FMUL,
    NIOS_USER_OPCODE_SMIN,
    NIOS_USER_OPCODE_UMIN,
    NIOS_USER_OPCODE_SMAX,
    NIOS_USER_OPCODE_UMAX,
    NIOS_USER_OPCODE_FFS,
    NIOS_USER_OPCODE_EXTV,
    NIOS_USER_OPCODE_EXTZV,
    NIOS_USER_OPCODE_INSV,
    NIOS_NUM_USER_OPCODES
};

struct nios_user_opcode {
  const char *name;	/* name of gcc operation */
  const char *s;	/* value part of -muser-opcode-???= command option */
  int pfx;		/* immediate value of prefix, or 0 if no prefix */
  int usr;		/* which USRx opcode; 0..4 */
};

typedef struct nios_user_opcode NIOS_USER_OPCODE;

extern NIOS_USER_OPCODE nios_user_opcodes[NIOS_NUM_USER_OPCODES];

extern char *nios_max_address;
extern int nios_suppress_address_xlo;
extern int nios_suppress_address_xhi;

/* Nios general purpose macros.  */
/* Align an address.  */
#define ADDR_ALIGN(addr,align) (((addr) + (align) - 1) & ~((align) - 1))

/* Return true if a value is inside a range.  */
#define IN_RANGE_P(VALUE, LOW, HIGH) ((VALUE) >= (LOW) && (VALUE) <= (HIGH))

/* Truncate and sign extend a value to 16 bits.  */
#define SIGN_EXTEND_SHORT(VALUE) ((int)((((VALUE) & 0xffff) ^ 0x8000) - 0x8000))

/* }}} */

/* {{{ Driver configuration.  */

/* A C expression which determines whether the option `-CHAR' takes
   arguments. The value should be the number of arguments that option
   takes-zero, for many options.

   By default, this macro is defined to handle the standard options
   properly. You need not define it unless you wish to add additional
   options which take arguments.

   Defined in svr4.h.  */

#undef SWITCH_TAKES_ARG
#define SWITCH_TAKES_ARG(CHAR)                                          \
  (DEFAULT_SWITCH_TAKES_ARG (CHAR) || (CHAR) == 'G')


/* A C string constant that tells the GNU CC driver program options to pass
   to CPP.  It can also specify how to translate options you give to GNU CC
   into options for GNU CC to pass to the CPP.

   Do not define this macro if it does not need to do anything.  */

#define CPP_SPEC "							\
  %{m16: -D__nios16__ -D__INT_MAX__=32767}				\
  %{m32: -D__nios32__}							\
  %{!m32: -D__nios16__ -D__INT_MAX__=32767}"


/* A C string constant that tells the GNU CC driver program options to pass
   to `cc1'.  It can also specify how to translate options you give to GNU
   CC into options for GNU CC to pass to the `cc1'.

   Do not define this macro if it does not need to do anything.  */

#define CC1_SPEC "							\
  %{G*}"


/* A C string constant that tells the GNU CC driver program options to pass
   to the assembler.  It can also specify how to translate options you give
   to GNU CC into options for GNU CC to pass to the assembler.  See the
   file `sun3.h' for an example of this.

   Defined in svr4.h.  */

#undef ASM_SPEC
#define ASM_SPEC "							\
  %{m16:-m16}								\
  %{m32:-m32}"

#undef STARTFILE_SPEC
#undef ENDFILE_SPEC
/* }}} */

/* {{{ Run-time target specifications.  */

/* Define this to be a string constant containing `-D' options to define
   the predefined macros that identify this machine and system.  These
   macros will be predefined unless the `-ansi' option is specified.

   In addition, a parallel set of macros are predefined, whose names are
   made by appending `__' at the beginning and at the end.  These `__'
   macros are permitted by the ANSI standard, so they are predefined
   regardless of whether `-ansi' is specified. */

#define CPP_PREDEFINES "-D__nios__ -Amachine(nios)"


/* This declaration should be present.  */
extern int target_flags;

/* This series of macros is to allow compiler command arguments to enable
   or disable the use of optional features of the target machine.  For
   example, one machine description serves both the 68000 and the 68020; a
   command argument tells the compiler whether it should use 68020-only
   instructions or not.  This command argument works by means of a macro
   `TARGET_68020' that tests a bit in `target_flags'.

   Define a macro `TARGET_FEATURENAME' for each such option.  Its
   definition should test a bit in `target_flags'; for example:

        #define TARGET_68020 (target_flags & 1)

   One place where these macros are used is in the condition-expressions of
   instruction patterns.  Note how `TARGET_68020' appears frequently in the
   68000 machine description file, `m68k.md'.  Another place they are used
   is in the definitions of the other macros in the `MACHINE.h' file.  */

#define MASK_32BIT	 	0x00000001	/* 32-bit regs/addrs, 16-bit insns */

/* Target debug masks */
#define MASK_DEBUG_ARG	 	0x40000000	/* debug argument handling */
#define MASK_DEBUG_ADDR	 	0x20000000	/* debug go_if_legitimate_address */
#define MASK_DEBUG_STACK 	0x10000000	/* debug stack frame */
#define MASK_DEBUG	 	0x08000000	/* general debugging switch */
#define MASK_DEBUG_LOC	 	0x04000000	/* optimize line # table */
#define MASK_FEWER_OPCODES	0x02000000	/* Compiler with full set of */
						/* instructions or reduced */
#define MASK_ANDN		0x00800000	/* some bit masks use andn */
#define MASK_NEG_CONST		0x00400000	/* better negative constants */
#define MASK_ZERO_EXTEND	0x00200000	/* match zero_extend patterns */
#define MASK_NEW_MULHISI	0x00100000	/* correct bug in mulhisi3 */
#define MASK_NIOS_FLAT          0x00080000      /* never use SAVE or RESTORE */
#define MASK_DCACHE             0x00040000      /* create PFXIO instructions */

#define MASK_DEFAULT		(-MASK_32BIT)

#define	TARGET_32BIT		((target_flags & MASK_32BIT) != 0)

#define TARGET_DEBUG_ARG	((target_flags & MASK_DEBUG_ARG) != 0)
#define TARGET_DEBUG_ADDR	((target_flags & MASK_DEBUG_ADDR) != 0)
#define TARGET_DEBUG_STACK	((target_flags & MASK_DEBUG_STACK) != 0)
#define TARGET_DEBUG		((target_flags & MASK_DEBUG) != 0)
#define TARGET_DEBUG_LOC	((target_flags & MASK_DEBUG_LOC) != 0)
#define TARGET_FEWER_OPCODES	((target_flags & MASK_FEWER_OPCODES) != 0)
#define TARGET_ANDN		((target_flags & MASK_ANDN) != 0)
#define TARGET_NEG_CONST	((target_flags & MASK_NEG_CONST) != 0)
#define TARGET_ZERO_EXTEND	((target_flags & MASK_ZERO_EXTEND) != 0)
#define TARGET_NEW_MULHISI	((target_flags & MASK_NEW_MULHISI) != 0)
#define TARGET_NIOS_FLAT        ((target_flags & MASK_NIOS_FLAT) != 0)
#define TARGET_DCACHE           ((target_flags & MASK_DCACHE) != 0)

#define TARGET_USER_OPCODE_MUL	\
	(nios_user_opcodes[NIOS_USER_OPCODE_MUL].s != NULL)
#define TARGET_USER_OPCODE_DIV	\
	(nios_user_opcodes[NIOS_USER_OPCODE_DIV].s != NULL)
#define TARGET_USER_OPCODE_UDIV	\
	(nios_user_opcodes[NIOS_USER_OPCODE_UDIV].s != NULL)
#define TARGET_USER_OPCODE_MOD	\
	(nios_user_opcodes[NIOS_USER_OPCODE_MOD].s != NULL)
#define TARGET_USER_OPCODE_UMOD	\
	(nios_user_opcodes[NIOS_USER_OPCODE_UMOD].s != NULL)
#define TARGET_USER_OPCODE_FADD	\
	(nios_user_opcodes[NIOS_USER_OPCODE_FADD].s != NULL)
#define TARGET_USER_OPCODE_FSUB	\
	(nios_user_opcodes[NIOS_USER_OPCODE_FSUB].s != NULL)
#define TARGET_USER_OPCODE_FMUL	\
	(nios_user_opcodes[NIOS_USER_OPCODE_FMUL].s != NULL)
#define TARGET_USER_OPCODE_SMIN	\
	(nios_user_opcodes[NIOS_USER_OPCODE_SMIN].s != NULL)
#define TARGET_USER_OPCODE_UMIN	\
	(nios_user_opcodes[NIOS_USER_OPCODE_UMIN].s != NULL)
#define TARGET_USER_OPCODE_SMAX	\
	(nios_user_opcodes[NIOS_USER_OPCODE_SMAX].s != NULL)
#define TARGET_USER_OPCODE_UMAX	\
	(nios_user_opcodes[NIOS_USER_OPCODE_UMAX].s != NULL)
#define TARGET_USER_OPCODE_FFS	\
	(nios_user_opcodes[NIOS_USER_OPCODE_FFS].s != NULL)
#define TARGET_USER_OPCODE_EXTV	\
	(nios_user_opcodes[NIOS_USER_OPCODE_EXTV].s != NULL)
#define	TARGET_USER_OPCODE_EXTZV \
	(nios_user_opcodes[NIOS_USER_OPCODE_EXTZV].s != NULL)
#define TARGET_USER_OPCODE_INSV	\
	(nios_user_opcodes[NIOS_USER_OPCODE_INSV].s != NULL)



/* This macro defines names of command options to set and clear bits in
   `target_flags'.  Its definition is an initializer with a subgrouping for
   each command option.

   Each subgrouping contains a string constant, that defines the option
   name, a number, which contains the bits to set in `target_flags', and an
   optional second string which is the textual description that will be
   displayed when the user passes --help on the command line.  If the
   number entry is negative then the specified bits will be cleared instead
   of being set.  If the second string entry is present but empty, then no
   help information will be displayed for that option, but it will not
   count as an undocumented option.  The actual option name, asseen on the
   command line is made by appending `-m' to the specified name.

   One of the subgroupings should have a null string.  The number in this
   grouping is the default value for `target_flags'.  Any target options
   act starting with that value.

   Here is an example which defines `-m68000' and `-m68020' with opposite
   meanings, and picks the latter as the default:

        #define TARGET_SWITCHES \
          { { "68020",  1, ""},      \
            { "68000", -1, "Compile for the m68000"},     \
            { "",       1, }}

   This declaration must be present.  */

#define TARGET_SWITCHES							    \
{									    \
  { "debug-arg",	MASK_DEBUG_ARG,		"Internal debug switch" },  \
  { "debug-addr",	MASK_DEBUG_ADDR,	"Internal debug switch" },  \
  { "debug-stack",	MASK_DEBUG_STACK,	"Internal debug switch" },  \
  { "debug",		MASK_DEBUG,		"Internal debug switch" },  \
  { "debug-loc",	MASK_DEBUG_LOC,		"Internal debug switch" },  \
  { "fewer-opcodes",	MASK_FEWER_OPCODES,	"Small Instruction Set" },  \
  { "no-new-mulhisi",	-MASK_NEW_MULHISI,				    \
			"Prior functionality" },			    \
  { "new-mulhisi",	MASK_NEW_MULHISI,				    \
			"Correct bug in mulhisi_internal (set by -O2)" },   \
  { "no-zero-extend",	-MASK_ZERO_EXTEND,				    \
			"Use lsli/lsri (default)" },			    \
  { "zero-extend",	MASK_ZERO_EXTEND,				    \
			"Use ext8s and ext16s" },			    \
  { "no-neg-const",	-MASK_NEG_CONST,				    \
			"Larger negative constants" },			    \
  { "neg-const",	MASK_NEG_CONST,					    \
			"Smaller negative constants (set by -O2)" },	    \
  { "no-andn",		-MASK_ANDN,					    \
			"Omit andn instructions" },  			    \
  { "andn",		MASK_ANDN,					    \
			"Some bit masks use andn (set by -O2)" },	    \
  { "no-dcache",	-MASK_DCACHE,					    \
			"Nios has no data cache (default)" },		    \
  { "dcache",		MASK_DCACHE,					    \
			"Nios has data cache.  Volatile loads use pfxio." },\
  { "16",	       	-MASK_32BIT,		"Generate 16-bit code" },   \
  { "32",		MASK_32BIT,		"Generate 32-bit code" },   \
  { "flat",             MASK_NIOS_FLAT,                                     \
                        "Never use SAVE / RESTORE instuctions"},            \
  { "",			MASK_DEFAULT,		"" }			    \
}

/* Default target_flags if no switches are specified  */

#ifdef IN_LIBGCC2
#undef TARGET_32BIT
/* Make this compile time constant for libgcc2 */
#ifdef __nios32__
#define TARGET_32BIT		1
#else
#define TARGET_32BIT		0
#endif
#endif /* IN_LIBGCC2 */


/* This macro is similar to `TARGET_SWITCHES' but defines names of
   command options that have values.  Its definition is an
   initializer with a subgrouping for each command option.

   Each subgrouping contains a string constant, that defines the
   fixed part of the option name, and the address of a variable. 
   The variable, type `char *', is set to the variable part of the
   given option if the fixed part matches.  The actual option name
   is made by appending `-m' to the specified name.

   Here is an example which defines `-mshort-data-NUMBER'.  If the
   given option is `-mshort-data-512', the variable `m88k_short_data'
   will be set to the string `"512"'.

	extern char *m88k_short_data;
	#define TARGET_OPTIONS { { "short-data-", &m88k_short_data } }  */

#define TARGET_OPTIONS 							 \
{									 \
  {"max-address=",	   &nios_max_address,				 \
		"Limit range of all generated addresses"},		 \
  {"user-opcode-ffs=",	   &nios_user_opcodes[NIOS_USER_OPCODE_FFS].s,   \
		"USRx instruction for find first set bit"},		 \
  {"user-opcode-umin=",	   &nios_user_opcodes[NIOS_USER_OPCODE_UMIN].s,  \
		"USRx instruction for unsigned integral minimum"},	 \
  {"user-opcode-smin=",	   &nios_user_opcodes[NIOS_USER_OPCODE_SMIN].s,  \
		"USRx instruction for signed integral minimum"},	 \
  {"user-opcode-umax=",	   &nios_user_opcodes[NIOS_USER_OPCODE_UMAX].s,  \
		"USRx instruction for unsigned integral maximum"},	 \
  {"user-opcode-smax=",	   &nios_user_opcodes[NIOS_USER_OPCODE_SMAX].s,  \
		"USRx instruction for signed integral maximum"},	 \
  {"user-opcode-insv=",	   &nios_user_opcodes[NIOS_USER_OPCODE_INSV].s,  \
		"USRx instruction for bit-field insertion"},		 \
  {"user-opcode-extzv=",   &nios_user_opcodes[NIOS_USER_OPCODE_EXTZV].s, \
		"USRx instruction for unsigned bit-field extract"},	 \
  {"user-opcode-extv=",	   &nios_user_opcodes[NIOS_USER_OPCODE_EXTV].s,  \
		"USRx instruction for signed bit-field extract"},	 \
  {"user-opcode-fmul=",	   &nios_user_opcodes[NIOS_USER_OPCODE_FMUL].s,  \
		"USRx instruction for 32-bit floating-point multiply"},	 \
  {"user-opcode-fsub=",	   &nios_user_opcodes[NIOS_USER_OPCODE_FSUB].s,  \
		"USRx instruction for 32-bit floating-point subtraction"},	 \
  {"user-opcode-fadd=",	   &nios_user_opcodes[NIOS_USER_OPCODE_FADD].s,  \
		"USRx instruction for 32-bit floating-point addition"},	 \
  {"user-opcode-umod=",	   &nios_user_opcodes[NIOS_USER_OPCODE_UMOD].s,  \
		"USRx instruction for unsigned integral modulus"},	 \
  {"user-opcode-mod=",	   &nios_user_opcodes[NIOS_USER_OPCODE_MOD].s,   \
		"USRx instruction for signed integral modulus"},	 \
  {"user-opcode-udiv=",	   &nios_user_opcodes[NIOS_USER_OPCODE_UDIV].s,  \
		"USRx instruction for unsigned integral division"},	 \
  {"user-opcode-div=",	   &nios_user_opcodes[NIOS_USER_OPCODE_DIV].s,   \
		"USRx instruction for signed integral division"},	 \
  {"user-opcode-mul=",	   &nios_user_opcodes[NIOS_USER_OPCODE_MUL].s,   \
		"USRx instruction for signed integral multiplication"},  \
}


/* This macro is a C statement to print on `stderr' a string describing the
   particular machine description choice.  Every machine description should
   define `TARGET_VERSION'.  */

#define TARGET_VERSION fprintf (stderr, " (nios)");


/* Sometimes certain combinations of command options do not make sense on a
   particular target machine.  You can define a macro `OVERRIDE_OPTIONS' to
   take account of this.  This macro, if defined, is executed once just
   after all the command options have been parsed.

   Don't use this macro to turn on various extra optimizations for `-O'.
   That is what `OPTIMIZATION_OPTIONS' is for.  */

#define OVERRIDE_OPTIONS nios_override_options ()


/* Some machines may desire to change what optimizations are performed for
   various optimization levels.  This macro, if defined, is executed once just
   after the optimization level is determined and before the remainder of the
   command options have been parsed.  Values set in this macro are used as the
   default values for the other command line options.

   LEVEL is the optimization level specified; 2 if `-O2' is specified, 1 if
   `-O' is specified, and 0 if neither is specified.

   SIZE is non-zero if `-Os' is specified, 0 otherwise.  

   You should not use this macro to change options that are not
   machine-specific.  These should uniformly selected by the same optimization
   level on all supported machines.  Use this macro to enable machbine-specific
   optimizations.

   *Do not examine `write_symbols' in this macro!* The debugging options are
   *not supposed to alter the generated code.  */
#define OPTIMIZATION_OPTIONS(LEVEL,SIZE) nios_optimization_options(LEVEL,SIZE)

/* }}} */

/* {{{ Small Data Area Support.  */

/* Maximum size of variables that go in .sdata/.sbss.
   The -msdata=foo switch also controls how small variables are handled. */

#ifndef SDATA_DEFAULT_SIZE
#define SDATA_DEFAULT_SIZE 8
#endif

extern int g_switch_value;        /* value of the -G xx switch */
extern int g_switch_set;          /* whether -G xx was passed.  */

/* }}} */

/* {{{ Storage Layout.  */

/* Define this to set the endianness to use in libgcc2.c, which can
   not depend on target_flags.  */

#define LIBGCC2_WORDS_BIG_ENDIAN 0

/* Define this macro to have the value 1 if the most significant bit in a
   byte has the lowest number; otherwise define it to have the value zero.
   This means that bit-field instructions count from the most significant
   bit. If the machine has no bit-field instructions, then this must still
   be defined, but it doesn't matter which value it is defined to.  This
   macro need not be a constant.

   This macro does not affect the way structure fields are packed into
   bytes or words; that is controlled by `BYTES_BIG_ENDIAN'.  */

#define BITS_BIG_ENDIAN 0


/* Define this macro to have the value 1 if the most significant byte in a
   word has the lowest number.  This macro need not be a constant.  */

#define BYTES_BIG_ENDIAN 0


/* Define this macro to have the value 1 if, in a multiword object, the
   most significant word has the lowest number.  This applies to both
   memory locations and registers; GNU CC fundamentally assumes that the
   order of words in memory is the same as the order in registers.  This
   macro need not be a constant.  */

#define WORDS_BIG_ENDIAN 0


/* Define this macro to be the number of bits in an addressable storage unit
   (byte); normally 8.  */

#define BITS_PER_UNIT 8


/* Number of bits in a word; normally 32.  */

#define BITS_PER_WORD ((TARGET_32BIT) ? 32 : 16)

/* Maximum number of bits in a word.  If this is undefined, the default is
   `BITS_PER_WORD'.  Otherwise, it is the constant value that is the largest
   value that `BITS_PER_WORD' can have at run-time. 
   
   If BITS_PER_WORD is not always a constant (e.g., if it depends on
   target_flags) this macro must be defined. */

#define MAX_BITS_PER_WORD 32


/* Number of storage units in a word; normally 4.  */

#define UNITS_PER_WORD ((TARGET_32BIT) ? 4 : 2)


/* Minimum number of units in a word.  If this is undefined, the default is
   `UNITS_PER_WORD'.  Otherwise, it is the constant value that is the smallest
   value that `UNITS_PER_WORD' can have at run-time.
   
   If UNITS_PER_WORD is not always a constant (e.g., if it depends on
   target_flags) this macro must be defined. */

#define MIN_UNITS_PER_WORD 2


/* Width of a pointer, in bits.  You must specify a value no wider than the
   width of `Pmode'.  If it is not equal to the width of `Pmode', you must
   define `POINTERS_EXTEND_UNSIGNED'.  */

#define POINTER_SIZE ((TARGET_32BIT) ? 32 : 16)


/* Alignment required for a function entry point, in bits.  */

#define FUNCTION_BOUNDARY 16


/* Biggest alignment that any data type can require on this machine,
   in bits.  */

#define BIGGEST_ALIGNMENT 32


/* If defined, a C expression to compute the alignment for a static
   variable. TYPE is the data type, and ALIGN is the alignment that the
   object would ordinarily have.  The value of this macro is used instead
   of that alignment to align the object.

   If this macro is not defined, then ALIGN is used.

   One use of this macro is to increase alignment of medium-size data to
   make it all fit in fewer cache lines.  Another is to cause character
   arrays to be word-aligned so that `strcpy' calls that copy constants to
   character arrays can be done inline.  */

#define DATA_ALIGNMENT(TYPE, ALIGN)					\
  ((TREE_CODE (TYPE) == ARRAY_TYPE					\
    && TYPE_MODE (TREE_TYPE (TYPE)) == QImode				\
    && (ALIGN) < BITS_PER_WORD)						\
   ? BITS_PER_WORD 							\
   : (ALIGN))


/* If defined, a C expression to compute the alignment given to a constant
   that is being placed in memory.  CONSTANT is the constant and ALIGN is
   the alignment that the object would ordinarily have.  The value of this
   macro is used instead of that alignment to align the object.

   If this macro is not defined, then ALIGN is used.

   The typical use of this macro is to increase alignment for string
   constants to be word aligned so that `strcpy' calls that copy constants
   can be done inline.  */

#define CONSTANT_ALIGNMENT(EXP, ALIGN)  				\
  (TREE_CODE (EXP) == STRING_CST					\
   && (ALIGN) < BITS_PER_WORD ? BITS_PER_WORD : (ALIGN))


/* Number of bits which any structure or union's size must be a multiple
   of. Each structure or union's size is rounded up to a multiple of this.

   If you do not define this macro, the default is the same as
   `BITS_PER_UNIT'.  */

#define STRUCTURE_SIZE_BOUNDARY BITS_PER_WORD


/* Define this macro to be the value 1 if instructions will fail to work if
   given data not on the nominal alignment.  If instructions will merely go
   slower in that case, define this macro as 0.  */

#define STRICT_ALIGNMENT 1


/* Define this if you wish to imitate the way many other C compilers handle
   alignment of bitfields and the structures that contain them.

   The behavior is that the type written for a bitfield (`int', `short', or
   other integer type) imposes an alignment for the entire structure, as if
   the structure really did contain an ordinary field of that type.  In
   addition, the bitfield is placed within the structure so that it would
   fit within such a field, not crossing a boundary for it.

   Thus, on most machines, a bitfield whose type is written as `int' would
   not cross a four-byte boundary, and would force four-byte alignment for
   the whole structure.  (The alignment used may not be four bytes; it is
   controlled by the other alignment parameters.)

   If the macro is defined, its definition should be a C expression; a
   nonzero value for the expression enables this behavior.

   Note that if this macro is not defined, or its value is zero, some
   bitfields may cross more than one alignment boundary.  The compiler can
   support such references if there are `insv', `extv', and `extzv' insns
   that can directly reference memory.

   The other known way of making bitfields work is to define
   `STRUCTURE_SIZE_BOUNDARY' as large as `BIGGEST_ALIGNMENT'.  Then every
   structure can be accessed with fullwords.

   Unless the machine has bitfield instructions or you define
   `STRUCTURE_SIZE_BOUNDARY' that way, you must define
   `PCC_BITFIELD_TYPE_MATTERS' to have a nonzero value.

   Defined in svr4.h.  */

#define PCC_BITFIELD_TYPE_MATTERS 1


/* GNU CC supports two ways of implementing C++ vtables: traditional or
   with so-called "thunks".  The flag `-fvtable-thunk' chooses between
   them. Define this macro to be a C expression for the default value of
   that flag.  If `DEFAULT_VTABLE_THUNKS' is 0, GNU CC uses the traditional
   implementation by default.  The "thunk" implementation is more efficient
   (especially if you have provided an implementation of
   `ASM_OUTPUT_MI_THUNK', but is not binary compatible with code compiled
   using the traditional implementation.  If you are writing a new ports,
   define `DEFAULT_VTABLE_THUNKS' to 1.

   If you do not define this macro, the default for `-fvtable-thunk' is 0.  */

#define DEFAULT_VTABLE_THUNKS 1

/* }}} */

/* {{{ Layout of Source Language Data Types.  */

/* A C expression for the size in bits of the type `int' on the target
   machine. If you don't define this, the default is one word.  */

#define INT_TYPE_SIZE ((TARGET_32BIT) ? 32 : 16)


/* Maximum number for the size in bits of the type `int' on the target
   machine. If this is undefined, the default is `INT_TYPE_SIZE'.
   Otherwise, it is the constant value that is the largest value that
   `INT_TYPE_SIZE' can have at run-time.  This is used in `cpp'.  */

#define MAX_INT_TYPE_SIZE 32


/* A C expression for the size in bits of the type `short' on the target
   machine.  If you don't define this, the default is half a word.  (If
   this would be less than one storage unit, it is rounded up to one unit.) */

#define SHORT_TYPE_SIZE	16


/* A C expression for the size in bits of the type `long' on the target
   machine.  If you don't define this, the default is one word.  */

#define LONG_TYPE_SIZE 32


/* A C expression for the size in bits of the type `long long' on the
   target machine.  If you don't define this, the default is two words.  If
   you want to support GNU Ada on your machine, the value of macro must be
   at least 64.  */

#define LONG_LONG_TYPE_SIZE 64


/* A C expression for the size in bits of the type `char' on the target
   machine.  If you don't define this, the default is one quarter of a
   word. (If this would be less than one storage unit, it is rounded up to
   one unit.)  */

#define CHAR_TYPE_SIZE	8


/* A C expression for the size in bits of the type `float' on the target
   machine.  If you don't define this, the default is one word.  */

#define FLOAT_TYPE_SIZE 32


/* A C expression for the size in bits of the type `double' on the target
   machine.  If you don't define this, the default is two words.  */

#define DOUBLE_TYPE_SIZE 64


/* A C expression for the size in bits of the type `long double' on the
   target machine.  If you don't define this, the default is two words.  */

#define LONG_DOUBLE_TYPE_SIZE 64


/* An expression whose value is 1 or 0, according to whether the type
   `char' should be signed or unsigned by default.  The user can always
   override this default with the options `-fsigned-char' and
   `-funsigned-char'.  */

#define DEFAULT_SIGNED_CHAR 1


/* A C expression for the size in bits of the data type for wide
   characters. This is used in `cpp', which cannot make use of
   `WCHAR_TYPE'.

   Defined in svr4.h.  */

#undef WCHAR_TYPE_SIZE
#define WCHAR_TYPE_SIZE	16

#define TARGET_BELL     0x7     /*  '\a'  */
#define TARGET_BS       0x8     /*  '\b'  */
#define TARGET_TAB      0x9     /*  '\t'  */
#define TARGET_NEWLINE  0xa     /*  '\n'  */
#define TARGET_VT       0xb     /*  '\v'  */
#define TARGET_FF       0xc     /*  '\f'  */
#define TARGET_CR       0xd     /*  '\r'  */

/* }}} */

/* {{{ Register Basics.  */

/* General purpose registers.  */

#define GPR_FIRST       0                       /* First gpr */
#define GPR_LAST        (GPR_FIRST + 31)        /* Last gpr */

#define GPR_R0          GPR_FIRST               /* Regular-old register,
						   not always 0 */
#define GPR_FP          (GPR_FIRST + 30)        /* Frame pointer */
#define GPR_RETURN	(GPR_FIRST + 31)	/* Return address as seen
						   by the callee */
#define GPR_LINK	(GPR_FIRST + 15)	/* Saved return address as
						   seen  by the caller */
#define GPR_SP          (GPR_FIRST + 14)        /* Stack pointer */


/* Base registers.  */
#define BP_REG_FIRST	(GPR_FIRST + 16)
#define BP_REG_LAST	(BP_REG_FIRST + 3)

#define GPR_BP0		BP_REG_FIRST		/* Base pointer 0 for STP/LDP */
#define GPR_BP1		(GPR_BP0   +  1)	/* Base pointer 1 for STP/LDP */
#define GPR_BP2		(GPR_BP0   +  2)	/* Base pointer 2 for STP/LDP */
#define GPR_BP3		(GPR_BP0   +  3)	/* Base pointer 3 for STP/LDP */


/* Control registers. */

#define	CT_REG_FIRST	(GPR_FIRST + 32)
#define CT_REG_LAST	(CT_REG_FIRST  +  5)

#define	CT10_STATUS	 CT_REG_FIRST		/* STATUS reg */
#define	CT11_ISTATUS	(CT_REG_FIRST  +  1)	/* ISTATUS reg */
#define CT12_WVALID	(CT_REG_FIRST  +  2)	/* WVALID reg */
#define CT13_CD_BANK	(CT_REG_FIRST  +  3)	/* CD_BANK reg (16-bit only) */
#define CT14_ST_BANK	(CT_REG_FIRST  +  4)	/* ST_BANK reg (16-bit only) */
#define CT15_LD_BANK	(CT_REG_FIRST  +  5)	/* LD_BANK reg (16-bit only) */

#define GPR_P(R)	IN_RANGE_P (R, GPR_FIRST, GPR_LAST)
#define BP_REG_P(R)	IN_RANGE_P (R, BP_REG_FIRST, BP_REG_LAST)
#define CC_REG_P(R)	IN_RANGE_P (R, CT_REG_FIRST, CT_REG_LAST)


/* Number of hardware registers known to the compiler.  They receive
   numbers 0 through `FIRST_PSEUDO_REGISTER-1'; thus, the first pseudo
   register's number really is assigned the number `FIRST_PSEUDO_REGISTER'. */

#define FIRST_PSEUDO_REGISTER CT_REG_LAST

#define IS_PSEUDO_P(R)	(REGNO (R) >= FIRST_PSEUDO_REGISTER)

/* The first/last register that can contain the arguments to a function.  */

#define FIRST_ARG_REGNUM  (        \
    (TARGET_NIOS_FLAT)             \
    ? (GPR_FIRST + 8)  /* %o0 */   \
    : (GPR_FIRST + 24) /* %i0 */   \
)

#define LAST_ARG_REGNUM		(FIRST_ARG_REGNUM + NIOS_NUM_ARG_REGS - 1)


/* An initializer that says which registers are used for fixed purposes all
   throughout the compiled code and are therefore not available for general
   allocation.  These would include the stack pointer, the frame pointer
   (except on machines where that can be used as a general register when no
   frame pointer is needed), the program counter on machines where that is
   considered one of the addressable registers, and any other numbered
   register with a standard use.

   This information is expressed as a sequence of numbers, separated by
   commas and surrounded by braces.  The Nth number is 1 if register N is
   fixed, 0 otherwise.

   The table initialized from this macro, and the table initialized by the
   following one, may be overridden at run time either automatically, by
   the actions of the macro `CONDITIONAL_REGISTER_USAGE', or by the user
   with the command options `-ffixed-REG', `-fcall-used-REG' and
   `-fcall-saved-REG'.  */

/* gpr0		Regular-old register (not always 0)
   gpr14	Stack pointer
   gpr15	LINK (saved address of CALL instruction)
   gpr16	Base pointer 0 for STP/LDP
   gpr17	Base pointer 1 for STP/LDP
   gpr18	Base pointer 2 for STP/LDP
   gpr19	Base pointer 3 for STP/LDP
   gpr30	FP (Frame Pointer)
   gpr31	Return address  */

#define FIXED_REGISTERS							\
{	/* General purpose registers */					\
	1, 0, 0, 0, 0, 0, 0, 0,		/* 000-007, %g0 - %g7 */	\
	0, 0, 0, 0, 0, 0, 1, 1,		/* 008-015, %o0 - %o7 */	\
	0, 0, 0, 0, 0, 0, 0, 0,		/* 016-023, %l0 - %l7 */	\
	0, 0, 0, 0, 0, 0, 1, 1,		/* 024-031, %i0 - %i7 */	\
	/* Condition Code Registers */					\
	1, 1, 1, 1, 1, 1		/* 032-037, ct10 - ct15 */	\
}


/* Like `FIXED_REGISTERS' but has 1 for each register that is clobbered (in
   general) by function calls as well as for fixed registers.  This macro
   therefore identifies the registers that are not available for general
   allocation of values that must live across function calls.

   If a register has 0 in `CALL_USED_REGISTERS', the compiler automatically
   saves it on function entry and restores it on function exit, if the
   register is used within the function.  */

#define CALL_USED_REGISTERS						\
{	/* General purpose registers */					\
	1, 1, 1, 1, 1, 1, 1, 1,		/* 000-007, %g0 - %g7 */	\
	1, 1, 1, 1, 1, 1, 1, 1,		/* 008-015, %o0 - %o7 */	\
	0, 0, 0, 0, 0, 0, 0, 0,		/* 016-023, %l0 - %l7 */	\
	0, 0, 0, 0, 0, 0, 1, 1,		/* 024-031, %i0 - %i7 */	\
	/* Condition Code Registers */					\
	1, 1, 1, 1, 1, 1		/* 032-037, ct10 - ct15 */	\
}

#define CONDITIONAL_REGISTER_USAGE do {                         \
    if (TARGET_NIOS_FLAT)                                       \
    {                                                           \
        nios_flat_conditional_register_usage();                 \
    }                                                           \
} while (0)

/* Define this macro if the target machine has register windows.  This C
   expression returns the register number as seen by the called function
   corresponding to the register number OUT as seen by the calling
   function. Return OUT if register number OUT is not an outbound register. */

#define INCOMING_REGNO(OUT) (      		        \
    (TARGET_NIOS_FLAT)                                  \
    ? (OUT)  /* <- no translation in flat mode */       \
    : (((OUT) < 8 || (OUT) > 15) ? (OUT) : (OUT) + 16)  \
)

/* Define this macro if the target machine has register windows.  This C
   expression returns the register number as seen by the calling function
   corresponding to the register number IN as seen by the called function.
   Return IN if register number IN is not an inbound register.  */

#define OUTGOING_REGNO(IN) (     		       \
    (TARGET_NIOS_FLAT)                                 \
    ? (IN)  /* <- no translation in flat mode */       \
    : (((IN) < 24 || (IN) > 31) ? (IN) : (IN) - 16)    \
)

/* Argument passing regs.  */
#define NIOS_OUTGOING_INT_ARG_FIRST 8
#define NIOS_INCOMING_INT_ARG_FIRST (TARGET_NIOS_FLAT ? 8 : 24)

/* }}} */

/* {{{ How Values Fit in Registers.  */

/* A C expression for the number of consecutive hard registers, starting at
   register number REGNO, required to hold a value of mode MODE.  */

#define HARD_REGNO_NREGS(REGNO, MODE) 					\
  ((GET_MODE_SIZE (MODE) + UNITS_PER_WORD - 1) / UNITS_PER_WORD)


/* A C expression that is nonzero if it is permissible to store a value of
   mode MODE in hard register number REGNO (or in several registers
   starting with that one).  */

#define HARD_REGNO_MODE_OK(REGNO, MODE) nios_hard_regno_mode_ok (REGNO, MODE)


/* A C expression that is nonzero if it is desirable to choose register
   allocation so as to avoid move instructions between a value of mode
   MODE1 and a value of mode MODE2.

   If `HARD_REGNO_MODE_OK (R, MODE1)' and `HARD_REGNO_MODE_OK (R, MODE2)'
   are ever different for any R, then `MODES_TIEABLE_P (MODE1, MODE2)' must
   be zero.  */

#define MODES_TIEABLE_P(MODE1, MODE2) nios_modes_tieable_p (MODE1, MODE2)


/* Define this macro if the compiler should avoid copies to/from CCmode
   registers.  You should only define this macro if support for copying
   to/from CCmode is incomplete.  */

#define AVOID_CCMODE_COPIES

/* }}} */

/* {{{ Handling Leaf Functions.  */

/* A C initializer for a vector, indexed by hard register number, which
   contains 1 for a register that is allowable in a candidate for leaf
   function treatment.

   If leaf function treatment involves renumbering the registers, then the
   registers marked here should be the ones before renumbering--those that
   GNU CC would ordinarily allocate.  The registers which will actually be
   used in the assembler code, after renumbering, should not be marked with
   1 in this vector.

   Define this macro only if the target machine offers a way to optimize
   the treatment of leaf functions.  */

#define LEAF_REGISTERS 		\
  { 1, 1, 1, 1, 1, 1, 1, 1,	\
    0, 0, 0, 0, 0, 0, 1, 0,	\
    0, 0, 0, 0, 0, 0, 0, 0,	\
    1, 1, 1, 1, 1, 1, 0, 1,	\
  }


/* A C expression whose value is the register number to which REGNO should
   be renumbered, when a function is treated as a leaf function.

   If REGNO is a register number which should not appear in a leaf function
   before renumbering, then the expression should yield -1, which will
   cause the compiler to abort.

   Define this macro only if the target machine offers a way to optimize
   the treatment of leaf functions, and registers need to be renumbered to
   do this.  */

extern char leaf_reg_remap[];
#define LEAF_REG_REMAP(REGNO) (leaf_reg_remap[REGNO])

/* }}} */

/* {{{ Register Classes.  */

/* An enumeral type that must be defined with all the register class names
   as enumeral values.  `NO_REGS' must be first.  `ALL_REGS' must be the
   last register class, followed by one more enumeral value,
   `LIM_REG_CLASSES', which is not a register class but rather tells how
   many classes there are.

   Each register class has a number, which is the value of casting the
   class name to type `int'.  The number serves as an index in many of the
   tables described below.

   All things being equal, the register allocator prefers the higher
   register class, so put all of the general registers higher than
   the accumulators and CRs.  */

enum reg_class
{
  NO_REGS,
  CT_REGS,
  LN_REGS,
  BP_REGS,
  DATA_REGS,
  GPR_REGS,
  ALL_REGS,
  LIM_REG_CLASSES
};


/* The name GENERAL_REGS must be the name of a class (or an alias for
   another name such as ALL_REGS).  This is the class of registers that is
   allowed by "g" or "r" in a register constraint. Also, registers outside
   this class are allocated only when instructions express preferences for
   them.  */

#define GENERAL_REGS GPR_REGS


/* The number of distinct register classes, defined as follows:

        #define N_REG_CLASSES (int) LIM_REG_CLASSES  */

#define N_REG_CLASSES ((int) LIM_REG_CLASSES)


/* An initializer containing the names of the register classes as C string
   constants.  These names are used in writing some of the debugging dumps.  */

#define REG_CLASS_NAMES {						\
   "NO_REGS",								\
   "CT_REGS",								\
   "LN_REGS",								\
   "BP_REGS",								\
   "DATA_REGS",								\
   "GPR_REGS",								\
   "ALL_REGS"								\
}


/* An initializer containing the contents of the register classes, as integers
   which are bit masks.  The Nth integer specifies the contents of class N.
   The way the integer MASK is interpreted is that register R is in the class
   if `MASK & (1 << R)' is 1.

   When the machine has more than 32 registers, an integer does not suffice.
   Then the integers are replaced by sub-initializers, braced groupings
   containing several integers.  Each sub-initializer must be suitable as an
   initializer for the type `HARD_REG_SET' which is defined in
   `hard-reg-set.h'.  */

#define REG_CLASS_CONTENTS						\
  {/* gpr31-gpr0  ct15-ct10 */						\
    { 0x00000000, 0x00000000 },/* NO_REGS  */				\
    { 0x00000000, 0x00000005 },/* CT_REGS  */				\
    { 0x80008000, 0x00000000 },/* LN_REGS  */				\
    { 0x400f4000, 0x00000000 },/* BP_REGS  */				\
    { 0x3ff03fff, 0x00000000 },/* DATA_REGS */				\
    { 0xffffffff, 0x00000000 },/* GPR_REGS */				\
    { 0xffffffff, 0xffffffff },/* ALL_REGS */				\
  }


/* A C expression whose value is a register class containing hard register
   REGNO.  In general there is more than one such class; choose a class which
   is "minimal", meaning that no smaller class also contains the register.  */

extern enum reg_class regno_reg_class[];
#define REGNO_REG_CLASS(REGNO) regno_reg_class [REGNO]


/* A macro whose definition is the name of the class to which a valid base
   register must belong.  A base register is one used in an address which is
   the register value plus a displacement.  */

#define BASE_REG_CLASS BP_REGS


/* A macro whose definition is the name of the class to which a valid index
   register must belong.  An index register is one used in an address where its
   value is either multiplied by a scale factor or added to another register
   (as well as added to a displacement).  */

#define INDEX_REG_CLASS NO_REGS


/* A C expression which defines the machine-dependent operand constraint
   letters for register classes.  If CHAR is such a letter, the value should be
   the register class corresponding to it.  Otherwise, the value should be
   `NO_REGS'.  The register letter `r', corresponding to class `GENERAL_REGS',
   will not be passed to this macro; you do not need to handle it.

   The following letters are unavailable, because they are already used as
   constraints:
	'0'..'9'
	'<', '>'
	'E', 'F', 'G', 'H'
	'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P'
	'Q', 'R', 'S', 'T', 'U'
	'V', 'X'
	'g', 'i', 'm', 'n', 'o', 'p', 'r', 's' */

extern enum reg_class reg_class_from_letter[];
#define REG_CLASS_FROM_LETTER(CHAR) reg_class_from_letter[(CHAR)]


/* A C expression which is nonzero if register number NUM is suitable for
   use as a base register in operand addresses.  It may be either a
   suitable hard register or a pseudo register that has been allocated such
   a hard register.  */

#define REGNO_OK_FOR_BASE_P(NUM)				\
  ((NUM) < FIRST_PSEUDO_REGISTER				\
   ? BP_REG_P (NUM)						\
   : (reg_renumber[NUM] >= 0 && BP_REG_P (reg_renumber[NUM])))


/* A C expression which is nonzero if register number NUM is suitable for
   use as an index register in operand addresses.  It may be either a
   suitable hard register or a pseudo register that has been allocated such
   a hard register.

   The difference between an index register and a base register is that the
   index register may be scaled.  If an address involves the sum of two
   registers, neither one of them scaled, then either one may be labeled
   the "base" and the other the "index"; but whichever labeling is used
   must fit the machine's constraints of which registers may serve in each
   capacity. The compiler will try both labelings, looking for one that is
   valid, and will reload one or both registers only if neither labeling
   works.  */

#define REGNO_OK_FOR_INDEX_P(NUM) 0


/* Is X, a REG, an in or global register?  i.e. is regno 0..7 or 24..31 */
#define IN_OR_GLOBAL_P(X)					\
    (REGNO (X) < 8 || (REGNO (X) >= 24 && REGNO (X) <= 31))


/* A C expression that places additional restrictions on the register class
   to use when it is necessary to copy value X into a register in class
   CLASS. The value is a register class; perhaps CLASS, or perhaps another,
   smaller class.  On many machines, the following definition is safe:

        #define PREFERRED_RELOAD_CLASS(X,CLASS) CLASS

   For nios we reload subword modes into data registers because to extract
   a byte or half-word using ext8d or ext16d we need the original base
   pointer intact, so we can't risk having the same base pointer register
   as a source and target of a move operation.  */

#define PREFERRED_RELOAD_CLASS(X, CLASS)				\
  ((GET_MODE_SIZE (GET_MODE (X)) < UNITS_PER_WORD)			\
   ? DATA_REGS								\
   : CLASS)


/* A C expression that places additional restrictions on the register class
   to use when it is necessary to be able to hold a value of mode MODE in a
   reload register for which class CLASS would ordinarily be used.

   Unlike `PREFERRED_RELOAD_CLASS', this macro should be used when there
   are certain modes that simply can't go in certain reload classes.

   The value is a register class; perhaps CLASS, or perhaps another,
   smaller class.

   Don't define this macro unless the target machine has limitations which
   require the macro to do something nontrivial.  */

#define LIMIT_RELOAD_CLASS(MODE, CLASS)					\
  ((GET_MODE_SIZE (MODE) < UNITS_PER_WORD)				\
   ? DATA_REGS								\
   : CLASS)


/* Many machines have some registers that cannot be copied directly to or
   from memory or even from other types of registers.  An example is the
   `MQ' register, which on most machines, can only be copied to or from
   general registers, but not memory.  Some machines allow copying all
   registers to and from memory, but require a scratch register for stores
   to some memory locations (e.g., those with symbolic address on the RT,
   and those with certain symbolic address on the Sparc when compiling
   PIC).  In some cases, both an intermediate and a scratch register are
   required.

   You should define these macros to indicate to the reload phase that it
   may need to allocate at least one register for a reload in addition to
   the register to contain the data.  Specifically, if copying X to a
   register CLASS in MODE requires an intermediate register, you should
   define `SECONDARY_INPUT_RELOAD_CLASS' to return the largest register
   class all of whose registers can be used as intermediate registers or
   scratch registers.

   If copying a register CLASS in MODE to X requires an intermediate or
   scratch register, `SECONDARY_OUTPUT_RELOAD_CLASS' should be defined to
   return the largest register class required.  If the requirements for
   input and output reloads are the same, the macro
   `SECONDARY_RELOAD_CLASS' should be used instead of defining both macros
   identically.

   The values returned by these macros are often `GENERAL_REGS'.  Return
   `NO_REGS' if no spare register is needed; i.e., if X can be directly
   copied to or from a register of CLASS in MODE without requiring a
   scratch register. Do not define this macro if it would always return
   `NO_REGS'.

   If a scratch register is required (either with or without an
   intermediate register), you should define patterns for `reload_inM' or
   `reload_outM', as required..  These patterns, which will normally be
   implemented with a `define_expand', should be similar to the `movM'
   patterns, except that operand 2 is the scratch register.

   Define constraints for the reload register and scratch register that
   contain a single register class.  If the original reload register (whose
   class is CLASS) can meet the constraint given in the pattern, the
   value returned by these macros is used for the class of the scratch
   register.  Otherwise, two additional reload registers are required.
   Their classes are obtained from the constraints in the insn pattern.

   X might be a pseudo-register or a `subreg' of a pseudo-register, which
   could either be in a hard register or in memory.  Use `true_regnum' to
   find out; it will return -1 if the pseudo is in memory and the hard
   register number if it is in a register.

   These macros should not be used in the case where a particular class of
   registers can only be copied to memory and not to another class of
   registers.  In that case, secondary reload registers are not needed and
   would not be helpful.  Instead, a stack location must be used to perform
   the copy and the `movM' pattern should use memory as a intermediate
   storage. This case often occurs between floating-point and general
   registers.  */
#define SECONDARY_RELOAD_CLASS(CLASS, MODE, X)				\
    nios_secondary_reload_class (CLASS, MODE, X)


/* A C expression for the maximum number of consecutive registers of
   class CLASS needed to hold a value of mode MODE.

   This is closely related to the macro `HARD_REGNO_NREGS'.  In fact, the
   value of the macro `CLASS_MAX_NREGS (CLASS, MODE)' should be the maximum
   value of `HARD_REGNO_NREGS (REGNO, MODE)' for all REGNO values in the
   class CLASS.

   This macro helps control the handling of multiple-word values in the
   reload pass.

   This declaration is required.  */

#define CLASS_MAX_NREGS(CLASS, MODE) 					\
  ((GET_MODE_SIZE (MODE) + UNITS_PER_WORD - 1) / UNITS_PER_WORD)


/* If defined, a C expression for a class that contains registers which the
   compiler must always access in a mode that is the same size as the mode
   in which it loaded the register.

   For the example, loading 32-bit integer or floating-point objects into
   floating-point registers on the Alpha extends them to 64-bits.
   Therefore loading a 64-bit object and then storing it as a 32-bit object
   does not store the low-order 32-bits, as would be the case for a normal
   register. Therefore, `alpha.h' defines this macro as `FLOAT_REGS'.  */

/* #define CLASS_CANNOT_CHANGE_SIZE */


/* Negative of 16-bit unsigned immediate.  */
#define CONST_OK_FOR_I(VALUE)	IN_RANGE_P (VALUE, -65535, 0)

/* 4-bit unsigned immediate.  */
#define CONST_OK_FOR_J(VALUE)	IN_RANGE_P(VALUE, 0, 15)

/* 8-bit unsigned immediate.  */
#define CONST_OK_FOR_K(VALUE)	IN_RANGE_P(VALUE, 0, 255)

/* 5-bit unsigned immediate.  */
#define CONST_OK_FOR_L(VALUE)	IN_RANGE_P(VALUE, 0, 31)

/* Negative of 5-bit unsigned immediate.  */
#define CONST_OK_FOR_M(VALUE)	IN_RANGE_P (VALUE, -31, 0)

/* 16-bit unsigned immediate.  */
#define CONST_OK_FOR_N(VALUE)	IN_RANGE_P (VALUE, 0, 65535)

/* 32-bit value okay for AND->ANDN conversion.  */
#define CONST_OK_FOR_O(VALUE)	IN_RANGE_P (VALUE, -65536, -1)

/* Power of two constants.  */
#define CONST_OK_FOR_P(VALUE)	(nios_power_of_two (VALUE))

/* 9-bit unsigned immediate.  */
#define CONST_OK_FOR_Q(VALUE)	IN_RANGE_P (VALUE, 0, 511)

/* 10-bit unsigned immediate.  */
#define CONST_OK_FOR_R(VALUE)	IN_RANGE_P (VALUE, 0, 1023)

/* 16-bit signed immediate.  */
#define CONST_OK_FOR_T(VALUE)	IN_RANGE_P(VALUE, -32768, 32767)

/* 11 bit signed immediate.  */
#define CONST_OK_FOR_U(VALUE)	IN_RANGE_P(VALUE, -1024, 1023)


/* A C expression that defines the machine-dependent operand constraint
   letters (`I', `J', `K', .. 'P') that specify particular ranges of
   integer values. If C is one of those letters, the expression should
   check that VALUE, an integer, is in the appropriate range and return 1
   if so, 0 otherwise.  If C is not one of those letters, the value should
   be 0 regardless of VALUE.  */

/* Since Nios requires more than 8 ranges, we use some of the unused
   machine-dependent operand constraint letters in Q through U.
   recog.c requires that any integer-range letters found in nios.md
   are from (`I' .. `P'), but other letters can be used in nios.c
   without harm.  */


#define CONST_OK_FOR_LETTER_P(VALUE, C)					\
  ((C) == 'I' ? CONST_OK_FOR_I (VALUE)					\
 : (C) == 'J' ? CONST_OK_FOR_J (VALUE)					\
 : (C) == 'K' ? CONST_OK_FOR_K (VALUE)					\
 : (C) == 'L' ? CONST_OK_FOR_L (VALUE)					\
 : (C) == 'M' ? CONST_OK_FOR_M (VALUE)					\
 : (C) == 'N' ? CONST_OK_FOR_N (VALUE)					\
 : (C) == 'O' ? CONST_OK_FOR_O (VALUE)					\
 : (C) == 'P' ? CONST_OK_FOR_P (VALUE)					\
 : (C) == 'Q' ? CONST_OK_FOR_Q (VALUE)					\
 : (C) == 'R' ? CONST_OK_FOR_R (VALUE)					\
 : (C) == 'T' ? CONST_OK_FOR_T (VALUE)					\
 : (C) == 'U' ? CONST_OK_FOR_U (VALUE)					\
 : 0)


/* A C expression that defines the machine-dependent operand constraint
   letters (`G', `H') that specify particular ranges of `const_double'
   values.

   If C is one of those letters, the expression should check that VALUE, an
   RTX of code `const_double', is in the appropriate range and return 1 if
   so, 0 otherwise.  If C is not one of those letters, the value should be
   0 regardless of VALUE.

   `const_double' is used for all floating-point constants and for `DImode'
   fixed-point constants.  A given letter can accept either or both kinds
   of values.  It can use `GET_MODE' to distinguish between these kinds. */

#define CONST_DOUBLE_OK_FOR_G(VALUE)					\
((GET_MODE (VALUE) == VOIDmode ||					\
  GET_MODE (VALUE) == SFmode ||						\
  GET_MODE (VALUE) == DFmode)						\
 && CONST_DOUBLE_LOW (VALUE) == 0					\
 && CONST_DOUBLE_HIGH (VALUE) == 0)

#define CONST_DOUBLE_OK_FOR_H(VALUE) 0

#define CONST_DOUBLE_OK_FOR_LETTER_P(VALUE, C)				\
((C) == 'G' ? CONST_DOUBLE_OK_FOR_G (VALUE)				\
 : (C) == 'H' ? CONST_DOUBLE_OK_FOR_H (VALUE)				\
 : 0)


/* A C expression that defines the optional machine-dependent constraint
   letters (`Q', `R', `S', `T', `U') that can be used to segregate specific
   types of operands, usually memory references, for the target machine.
   Normally this macro will not be defined.  If it is required for a
   particular target machine, it should return 1 if VALUE corresponds to
   the operand type represented by the constraint letter C.  If C is not
   defined as an extra constraint, the value returned should be 0
   regardless of VALUE.

   For example, on the ROMP, load instructions cannot have their output in
   r0 if the memory reference contains a symbolic address.  Constraint
   letter `Q' is defined as representing a memory address that does *not*
   contain a symbolic address.  An alternative is specified with a `Q'
   constraint on the input and `r' on the output.  The next alternative
   specifies `m' on the input and a register class that does not include r0
   on the output.  */

/* unused, possibly quad word memory ops that take one instruction */
#define EXTRA_CONSTRAINT_FOR_Q(VALUE) 0

/* double word memory ops that take one instruction.  */
#define EXTRA_CONSTRAINT_FOR_R(VALUE) 0

/* SYMBOL_REF */
#define EXTRA_CONSTRAINT_FOR_S(VALUE) (GET_CODE (VALUE) == SYMBOL_REF)

/* double word memory ops that take two instructions.  */
#define EXTRA_CONSTRAINT_FOR_T(VALUE) 0

/* unused */
#define EXTRA_CONSTRAINT_FOR_U(VALUE) 0

#define EXTRA_CONSTRAINT(VALUE, C)					\
  ((C) == 'Q' ? EXTRA_CONSTRAINT_FOR_Q (VALUE)				\
 : (C) == 'R' ? EXTRA_CONSTRAINT_FOR_R (VALUE)				\
 : (C) == 'S' ? EXTRA_CONSTRAINT_FOR_S (VALUE)				\
 : (C) == 'T' ? EXTRA_CONSTRAINT_FOR_T (VALUE)				\
 : (C) == 'U' ? EXTRA_CONSTRAINT_FOR_U (VALUE)				\
 : 0)

/* }}} */

/* {{{ Basic Stack Layout.  */

/* Normal alignment required for function parameters on the stack, in bits.
   All stack parameters receive at least this much alignment regardless of data
   type.  On most machines, this is the same as the size of an integer.  */

#define PARM_BOUNDARY ((TARGET_32BIT) ? 32 : 16)


/* Define this macro if you wish to preserve a certain alignment for the stack
   pointer.  The definition is a C expression for the desired alignment
   (measured in bits).

   If `PUSH_ROUNDING' is not defined, the stack will always be aligned to the
   specified boundary.  If `PUSH_ROUNDING' is defined and specifies a less
   strict alignment than `STACK_BOUNDARY', the stack may be momentarily
   unaligned while pushing arguments.  */

#define STACK_BOUNDARY ((TARGET_32BIT) ? 32 : 16)


/* Align frames on word boundaries.  */

#define NIOS_STACK_ALIGN(LOC)						\
    ((TARGET_32BIT)							\
     ? (((LOC) + 3) & ~3)						\
     : (((LOC) + 1) & ~1))


/* Maximum number of words that can be reserved with the 'save' instruction.  */

#define NIOS_MAX_SAVE_SIZE 255


/* Define this macro if pushing a word onto the stack moves the stack pointer
   to a smaller address.

   When we say, "define this macro if ...," it means that the compiler
   checks this macro only with `#ifdef' so the precise definition used does
   not matter.  */

#define STACK_GROWS_DOWNWARD 1


/* Define this macro if the addresses of local variable slots are at negative
   offsets from the frame pointer.  */

#define FRAME_GROWS_DOWNWARD 1


/* Define this macro if successive arguments to a function occupy decreasing
   addresses on the stack.  */
/* #define ARGS_GROW_DOWNWARD  */


/* Offset from the frame pointer to the first local variable slot to be
   allocated.

   If `FRAME_GROWS_DOWNWARD', find the next slot's offset by subtracting
   the first slot's length from `STARTING_FRAME_OFFSET'.  Otherwise, it is
   found by adding the length of the first slot to the value
   `STARTING_FRAME_OFFSET'.  */

#define STARTING_FRAME_OFFSET	0


/* Offset from the stack pointer register to the first location at which
   outgoing arguments are placed.  If not specified, the default value of
   zero is used.  This is the proper value for most machines.

   If `ARGS_GROW_DOWNWARD', this is the offset to the location above the
   first location at which outgoing arguments are placed.  */

#define STACK_POINTER_OFFSET FIRST_PARM_OFFSET(0)


/* Offset from the argument pointer register to the first argument's
   address. This is 32 (16-bit nios) or 64 (32-bit nios) for the ins and
   locals, plus 2 (16-bit nios) or 4 (32-bit nios) for the struct-return
   reg even if this function isn't going to use it.  */

#define FIRST_PARM_OFFSET(FUNDECL) (STRUCT_VALUE_OFFSET + UNITS_PER_WORD)


/* A C expression whose value is RTL representing the value of the return
   address for the frame COUNT steps up from the current frame, after the
   prologue.  FRAMEADDR is the frame pointer of the COUNT frame, or the
   frame pointer of the COUNT - 1 frame if `RETURN_ADDR_IN_PREVIOUS_FRAME'
   is defined.

   The value of the expression must always be the correct address when
   COUNT is zero, but may be `NULL_RTX' if there is not way to determine
   the return address of other frames.  */

#define RETURN_ADDR_RTX(COUNT, FRAMEADDR) 				\
    nios_return_addr_rtx (COUNT, FRAMEADDR)


/* A C expression whose value is RTL representing the location of the
   incoming return address at the beginning of any function, before the
   prologue. This RTL is either a `REG', indicating that the return value
   is saved in `REG', or a `MEM' representing a location in the stack.

   You only need to define this macro if you want to support call frame
   debugging information like that provided by DWARF 2.  */

/* #define INCOMING_RETURN_ADDR_RTX gen_rtx_REG (SImode, GPR_RETURN) */


/* A C expression whose value is an integer giving the offset, in bytes,
   from the value of the stack pointer register to the top of the stack
   frame at the beginning of any function, before the prologue.  The top of
   the frame is defined to be the value of the stack pointer in the
   previous frame, just before the call instruction.

   You only need to define this macro if you want to support call frame
   debugging information like that provided by DWARF 2.  */

#define INCOMING_FRAME_SP_OFFSET 0

/* }}} */

/* {{{ Register That Address the Stack Frame.  */

/* The register number of the stack pointer register, which must also be a
   fixed register according to `FIXED_REGISTERS'.  On most machines, the
   hardware determines which register this is.  */

#define STACK_POINTER_REGNUM GPR_SP


/* Offset from the stack pointer register to the first location at which
   outgoing arguments are placed.  If not specified, the default value of zero
   is used.  This is the proper value for most machines.

   If `ARGS_GROW_DOWNWARD', this is the offset to the location above the first
   location at which outgoing arguments are placed.  */

#define STACK_POINTER_OFFSET FIRST_PARM_OFFSET(0)


/* The register number of the frame pointer register, which is used to access
   automatic variables in the stack frame.  On some machines, the hardware
   determines which register this is.  On other machines, you can choose any
   register you wish for this purpose.  */

#define FRAME_POINTER_REGNUM GPR_FP


/* The register number of the arg pointer register, which is used to access
   the function's argument list.  On some machines, this is the same as the
   frame pointer register.  On some machines, the hardware determines which
   register this is.  On other machines, you can choose any register you
   wish for this purpose.  If this is not the same register as the frame
   pointer register, then you must mark it as a fixed register according to
   `FIXED_REGISTERS', or arrange to be able to eliminate it.  */

#define ARG_POINTER_REGNUM FRAME_POINTER_REGNUM


/* Register numbers used for passing a function's static chain pointer.  If
   register windows are used, the register number as seen by the called
   function is `STATIC_CHAIN_INCOMING_REGNUM', while the register number as
   seen by the calling function is `STATIC_CHAIN_REGNUM'.  If these
   registers are the same, `STATIC_CHAIN_INCOMING_REGNUM' need not be
   defined.

   The static chain register need not be a fixed register.

   If the static chain is passed in memory, these macros should not be
   defined; instead, the next two macros should be defined.  */

#define STATIC_CHAIN_REGNUM 2

/* }}} */

/* {{{ Eliminating the Frame Pointer and the Arg Pointer.  */

/* A C expression which is nonzero if a function must have and use a frame
   pointer.  This expression is evaluated in the reload pass.  If its value is
   nonzero the function will have a frame pointer.

   The expression can in principle examine the current function and decide
   according to the facts, but on most machines the constant 0 or the constant
   1 suffices.  Use 0 when the machine allows code to be generated with no
   frame pointer, and doing so saves some time or space.  Use 1 when there is
   no possible advantage to avoiding a frame pointer.

   In certain cases, the compiler does not know how to produce valid code
   without a frame pointer.  The compiler recognizes those cases and
   automatically gives the function a frame pointer regardless of what
   `FRAME_POINTER_REQUIRED' says.  You don't need to worry about them.

   In a function that does not require a frame pointer, the frame pointer
   register can be allocated for ordinary usage, unless you mark it as a fixed
   register.  See `FIXED_REGISTERS' for more information.  */

#define FRAME_POINTER_REQUIRED					0


/* A C statement to store in the variable DEPTH_VAR the difference between
   the frame pointer and the stack pointer values immediately after the
   function prologue.  The value would be computed from information such as
   the result of `get_frame_size ()' and the tables of registers
   `regs_ever_live' and `call_used_regs'.

   If `ELIMINABLE_REGS' is defined, this macro will not be used and need
   not be defined.  Otherwise, it must be defined even if
   `FRAME_POINTER_REQUIRED' is defined to always be true; in that case, you
   may set DEPTH_VAR to anything.  */

#define INITIAL_FRAME_POINTER_OFFSET(DEPTH_VAR)			\
    ((DEPTH_VAR) =						\
	nios_compute_frame_size (get_frame_size (), 		\
	                         leaf_function_p () && only_leaf_regs_used ()))

/* }}} */

/* {{{ Passing Function Arguments on the Stack.  */

/* Define this macro if an argument declared in a prototype as an integral
   type smaller than `int' should actually be passed as an `int'.  In
   addition to avoiding errors in certain cases of mismatch, it also makes
   for better code on certain machines.  */

#define PROMOTE_PROTOTYPES 1


/* If defined, the maximum amount of space required for outgoing arguments
   will be computed and placed into the variable
   `current_function_outgoing_args_size'.  No space will be pushed onto the
   stack for each call; instead, the function prologue should increase the
   stack frame size by this amount.

   Defining both `PUSH_ROUNDING' and `ACCUMULATE_OUTGOING_ARGS' is not
   proper.  */

#define ACCUMULATE_OUTGOING_ARGS


/* Define this macro if functions should assume that stack space has been
   allocated for arguments even when their values are passed in registers.

   The value of this macro is the size, in bytes, of the area reserved for
   arguments passed in registers for the function represented by FNDECL.

   This space can be allocated by the caller, or be a part of the
   machine-dependent stack frame: `OUTGOING_REG_PARM_STACK_SPACE' says
   which.  
 
   For nios we must reserve space for the 6 register arguments (%o0 -
   %o5).  */

#define REG_PARM_STACK_SPACE(FNDECL) (NIOS_NUM_ARG_REGS * UNITS_PER_WORD)


/* A C expression that should indicate the number of bytes of its own
   arguments that a function pops on returning, or 0 if the function pops
   no arguments and the caller must therefore pop them all after the
   function returns.  */

#define RETURN_POPS_ARGS(FUNDECL, FUNTYPE, STACK_SIZE) 0

/* }}} */

/* {{{ Function Arguments in Registers.  */

/* The number of register assigned to holding function arguments.  */

#define NIOS_NUM_ARG_REGS        	6


/* Number of registers in the register window.  */

#define NIOS_REGISTER_WINDOW_SIZE	16


/* A C expression that controls whether a function argument is passed in a
   register, and which register.

   The arguments are CUM, of type CUMULATIVE_ARGS, which summarizes (in a
   way defined by INIT_CUMULATIVE_ARGS and FUNCTION_ARG_ADVANCE) all of the
   previous arguments so far passed in registers; MODE, the machine mode of
   the argument; TYPE, the data type of the argument as a tree node or 0 if
   that is not known (which happens for C support library functions); and
   NAMED, which is 1 for an ordinary argument and 0 for nameless arguments
   that correspond to `...' in the called function's prototype.

   The value of the expression should either be a `reg' RTX for the hard
   register in which to pass the argument, or zero to pass the argument on
   the stack.

   For machines like the Vax and 68000, where normally all arguments are
   pushed, zero suffices as a definition.

   The usual way to make the ANSI library `stdarg.h' work on a machine
   where some arguments are usually passed in registers, is to cause
   nameless arguments to be passed on the stack instead.  This is done by
   making `FUNCTION_ARG' return 0 whenever NAMED is 0.

   You may use the macro `MUST_PASS_IN_STACK (MODE, TYPE)' in the
   definition of this macro to determine if this argument is of a type that
   must be passed in the stack.  If `REG_PARM_STACK_SPACE' is not defined
   and `FUNCTION_ARG' returns non-zero for such an argument, the compiler
   will abort.  If `REG_PARM_STACK_SPACE' is defined, the argument will be
   computed in the stack and then loaded into a register.  */

#define FUNCTION_ARG(CUM, MODE, TYPE, NAMED)                    \
    nios_function_arg (&(CUM), (MODE), (TYPE), (NAMED), FALSE)


/* Define this macro if the target machine has "register windows", so that
   the register in which a function sees an arguments is not necessarily
   the same as the one in which the caller passed the argument.

   For such machines, `FUNCTION_ARG' computes the register in which the
   caller passes the value, and `FUNCTION_INCOMING_ARG' should be defined
   in a similar fashion to tell the function being called where the
   arguments will arrive.

   If `FUNCTION_INCOMING_ARG' is not defined, `FUNCTION_ARG' serves both
   purposes.  */

#define FUNCTION_INCOMING_ARG(CUM, MODE, TYPE, NAMED)			\
    nios_function_arg (&CUM, MODE, TYPE, NAMED, TRUE)


/* A C expression for the number of words, at the beginning of an argument,
   must be put in registers.  The value must be zero for arguments that are
   passed entirely in registers or that are entirely pushed on the stack.

   On some machines, certain arguments must be passed partially in
   registers and partially in memory.  On these machines, typically the
   first N words of arguments are passed in registers, and the rest on the
   stack.  If a multi-word argument (a `double' or a structure) crosses
   that boundary, its first few words must be passed in registers and the
   rest must be pushed. This macro tells the compiler when this occurs, and
   how many of the words should go in registers.

   `FUNCTION_ARG' for these arguments should return the first register to be
   used by the caller for this argument; likewise `FUNCTION_INCOMING_ARG', for
   the called function.  */

#define FUNCTION_ARG_PARTIAL_NREGS(CUM, MODE, TYPE, NAMED)		\
    nios_function_arg_partial_nregs (&CUM, MODE, TYPE, NAMED)


/* A C expression that indicates when an argument must be passed by
   reference. If nonzero for an argument, a copy of that argument is made
   in memory and a pointer to the argument is passed instead of the
   argument itself. The pointer is passed in whatever way is appropriate
   for passing a pointer to that type.

   On machines where `REG_PARM_STACK_SPACE' is not defined, a suitable
   definition of this macro might be
        #define FUNCTION_ARG_PASS_BY_REFERENCE(CUM, MODE, TYPE, NAMED)  \
          MUST_PASS_IN_STACK (MODE, TYPE)  */

#define FUNCTION_ARG_PASS_BY_REFERENCE(CUM, MODE, TYPE, NAMED)		\
    nios_function_arg_pass_by_reference (&CUM, MODE, TYPE, NAMED)


/* A C type for declaring a variable that is used as the first argument of
   `FUNCTION_ARG' and other related values.  For some target machines, the
   type `int' suffices and can hold the number of bytes of argument so far.

   There is no need to record in `CUMULATIVE_ARGS' anything about the
   arguments that have been passed on the stack.  The compiler has other
   variables to keep track of that.  For target machines on which all
   arguments are passed on the stack, there is no need to store anything in
   `CUMULATIVE_ARGS'; however, the data structure must exist and should not
   be empty, so use `int'.

   For Nios, this is a single integer, which is the number of words of
   arguments scanned so far (including the invisible argument, if any,
   which holds the structure-value-address). Thus 7 or more means all
   following args should go on the stack. Note: we use a struct to
   accomodate for future changes (7-Mar-2000).  */

struct nios_args {
  int words;       /* number of words passed so far.  */
};

typedef struct nios_args CUMULATIVE_ARGS;


/* A C statement (sans semicolon) for initializing the variable CUM for the
   state at the beginning of the argument list.  The variable has type
   `CUMULATIVE_ARGS'.  The value of FNTYPE is the tree node for the data
   type of the function which will receive the args, or 0 if the args are
   to a compiler support library function.  The value of INDIRECT is
   nonzero when processing an indirect call, for example a call through a
   function pointer. The value of INDIRECT is zero for a call to an
   explicitly named function, a library function call, or when
   `INIT_CUMULATIVE_ARGS' is used to find arguments for the function being
   compiled.

   When processing a call to a compiler support library function, LIBNAME
   identifies which one.  It is a `symbol_ref' rtx which contains the name
   of the function, as a string.  LIBNAME is 0 when an ordinary C function
   call is being processed.  Thus, each time this macro is called, either
   LIBNAME or FNTYPE is nonzero, but never both of them at once.  */

#define INIT_CUMULATIVE_ARGS(CUM, FNTYPE, LIBNAME, INDIRECT)		\
    nios_init_cumulative_args (&CUM, FNTYPE, LIBNAME, INDIRECT, FALSE)


/* Like `INIT_CUMULATIVE_ARGS' but overrides it for the purposes of finding
   the arguments for the function being compiled.  If this macro is
   undefined, `INIT_CUMULATIVE_ARGS' is used instead.

   The value passed for LIBNAME is always 0, since library routines with
   special calling conventions are never compiled with GNU CC.  The argument
   LIBNAME exists for symmetry with `INIT_CUMULATIVE_ARGS'.  */

#define INIT_CUMULATIVE_INCOMING_ARGS(CUM, FNTYPE, LIBNAME)		\
    nios_init_cumulative_args (&CUM, FNTYPE, LIBNAME, FALSE, TRUE)


/* A C statement (sans semicolon) to update the summarizer variable CUM to
   advance past an argument in the argument list.  The values MODE, TYPE and
   NAMED describe that argument.  Once this is done, the variable CUM is
   suitable for analyzing the *following* argument with `FUNCTION_ARG', etc.

   This macro need not do anything if the argument in question was passed on
   the stack.  The compiler knows how to track the amount of stack space used
   for arguments without any special help.  */

#define FUNCTION_ARG_ADVANCE(CUM, MODE, TYPE, NAMED)			\
    nios_function_arg_advance (&CUM, MODE, TYPE, NAMED)


/* If defined, a C expression which determines whether, and in which
   direction, to pad out an argument with extra space.  The value should be
   of type `enum direction': either `upward' to pad above the argument,
   `downward' to pad below, or `none' to inhibit padding.

   The *amount* of padding is always just enough to reach the next multiple
   of `FUNCTION_ARG_BOUNDARY'; this macro does not control it.

   This macro has a default definition which is right for most systems.
   For little-endian machines, the default is to pad upward.  For
   big-endian machines, the default is to pad downward for an argument of
   constant size shorter than an `int', and upward otherwise.  */
/* #define FUNCTION_ARG_PADDING(MODE, TYPE) */

/* If defined, a C expression that gives the alignment boundary, in bits,
   of an argument with the specified mode and type.  If it is not defined,
   `PARM_BOUNDARY' is used for all arguments.  */

#define FUNCTION_ARG_BOUNDARY(MODE, TYPE)				\
    nios_function_arg_boundary (MODE, TYPE)


/* A C expression that is nonzero if REGNO is the number of a hard register
   in which function arguments are sometimes passed.  This does *not*
   include implicit arguments such as the static chain and the
   structure-value address. On many machines, no registers can be used for
   this purpose since all function arguments are pushed on the stack.  */

#define FUNCTION_ARG_REGNO_P(REGNO)					\
  ((REGNO) >= FIRST_ARG_REGNUM && ((REGNO) <= LAST_ARG_REGNUM))

/* }}} */

/* {{{ How Scalar Function Values are Returned.  */

/* The number of the hard register that is used to return a scalar value
   from a function call (as seen by the callee). On nios we return scalar
   values in register %i0 (%r24) which is register %o0 (%r8) in our caller. */

#define RETURN_VALUE_REGNUM_CALLEE	((TARGET_NIOS_FLAT) ? 8 : 24)
#define RETURN_VALUE_REGNUM_CALLER	8
     

/* Scalar value as seen by the caller. The caller expects to see this value
   in its first output register (%o0)  */

#define FUNCTION_VALUE(VALTYPE, FUNC)					\
  gen_rtx (REG, TYPE_MODE (VALTYPE), RETURN_VALUE_REGNUM_CALLER)


/* Scalar value about to be returned by the callee. The callee stores this
   value in its first input register (%i0)  */

#define FUNCTION_OUTGOING_VALUE(VALTYPE, FUNC)				\
  gen_rtx (REG, TYPE_MODE (VALTYPE), RETURN_VALUE_REGNUM_CALLEE)


/* A C expression to create an RTX representing the place where a library
   function returns a value of mode MODE.

   Note that "library function" in this context means a compiler support
   routine, used to perform arithmetic, whose name is known specially by
   the compiler and was not mentioned in the C code being compiled.

   The definition of `LIBRARY_VALUE' need not be concerned aggregate data
   types, because none of the library functions returns such types.  */

#define LIBCALL_VALUE(MODE)						\
  gen_rtx (REG, MODE, RETURN_VALUE_REGNUM_CALLER)


/* A C expression that is nonzero if REGNO is the number of a hard register
   in which the values of a called function may come back.

   A register whose use for returning values is limited to serving as the
   second of a pair (for a value of type `double', say) need not be
   recognized by this macro.  So for most machines, this definition
   suffices:

        #define FUNCTION_VALUE_REGNO_P(N) ((N) == RETURN_VALUE_REGNUM)

   If the machine has register windows, so that the caller and the called
   function use different registers for the return value, this macro should
   recognize only the caller's register numbers.  */

#define FUNCTION_VALUE_REGNO_P(REGNO) ((REGNO) == RETURN_VALUE_REGNUM_CALLER)

/* }}} */

/* {{{ How Large Values are Returned.  */

/* A C expression which can inhibit the returning of certain function
   values in registers, based on the type of value.  A nonzero value says
   to return the function value in memory, just as large structures are
   always returned. Here TYPE will be a C expression of type `tree',
   representing the data type of the value.

   Note that values of mode `BLKmode' must be explicitly handled by this
   macro. Also, the option `-fpcc-struct-return' takes effect regardless of
   this macro.  On most systems, it is possible to leave the macro
   undefined; this causes a default definition to be used, whose value is
   the constant 1 for `BLKmode' values, and 0 otherwise.

   Do not use this macro to indicate that structures and unions should
   always be returned in memory.  You should instead use
   `DEFAULT_PCC_STRUCT_RETURN' to indicate this.  */

#define RETURN_IN_MEMORY(TYPE)	(int_size_in_bytes (TYPE) > UNITS_PER_WORD)


/* Define this macro to be 1 if all structure and union return values must
   be in memory.  Since this results in slower code, this should be defined
   only if needed for compatibility with other compilers or with an ABI. If
   you define this macro to be 0, then the conventions used for structure
   and union return values are decided by the `RETURN_IN_MEMORY' macro.

   If not defined, this defaults to the value 1.  */

#define DEFAULT_PCC_STRUCT_RETURN 0


/* If the structure value address is passed in a register, then
   `STRUCT_VALUE_REGNUM' should be the number of that register.  */

#define STRUCT_VALUE_REGNUM 3


/* Functions which return large structures get the address to place the
   wanted value at offset 32 (16-bit nios) or 64 (32-bit nios) from the
   frame. Must reserve STRUCT_VALUE_OFFSET bytes for the in and local
   registers.  */

#define STRUCT_VALUE_OFFSET	NIOS_REGISTER_WINDOW_SIZE * UNITS_PER_WORD


/* If the structure value address is not passed in a register, define
   `STRUCT_VALUE' as an expression returning an RTX for the place where the
   address is passed.  If it returns 0, the address is passed as an
   "invisible" first argument.  */

#define STRUCT_VALUE						\
   gen_rtx_MEM (Pmode, plus_constant (stack_pointer_rtx,	\
				      STRUCT_VALUE_OFFSET))


/* If the incoming location is not a register, then you should define
   `STRUCT_VALUE_INCOMING' as an expression for an RTX for where the called
   function should find the value.  If it should find the value on the stack,
   define this to create a `mem' which refers to the frame pointer.  A
   definition of 0 means that the address is passed as an "invisible" first
   argument.  */

#define STRUCT_VALUE_INCOMING \
   gen_rtx_MEM (Pmode, plus_constant (frame_pointer_rtx,	\
				      STRUCT_VALUE_OFFSET))

/* }}} */

/* {{{ Function Entry and Exit.  */

/* A C compound statement that outputs the assembler code for entry to a
   function.  The prologue is responsible for setting up the stack frame,
   initializing the frame pointer register, saving registers that must be
   saved, and allocating SIZE additional bytes of storage for the local
   variables.  SIZE is an integer.  FILE is a stdio stream to which the
   assembler code should be output.

   The label for the beginning of the function need not be output by this
   macro.  That has already been done when the macro is run.

   To determine which registers to save, the macro can refer to the array
   `regs_ever_live': element R is nonzero if hard register R is used
   anywhere within the function.  This implies the function prologue should
   save register R, provided it is not one of the call-used registers.
   (`FUNCTION_EPILOGUE' must likewise use `regs_ever_live'.)

   On machines that have "register windows", the function entry code does
   not save on the stack the registers that are in the windows, even if
   they are supposed to be preserved by function calls; instead it takes
   appropriate steps to "push" the register stack, if any non-call-used
   registers are used in the function.

   On machines where functions may or may not have frame-pointers, the
   function entry code must vary accordingly; it must set up the frame
   pointer if one is wanted, and not otherwise.  To determine whether a
   frame pointer is in wanted, the macro can refer to the variable
   `frame_pointer_needed'.  The variable's value will be 1 at run time in a
   function that needs a frame pointer.

   The function entry code is responsible for allocating any stack space
   required for the function.  This stack space consists of the regions
   listed below.  In most cases, these regions are allocated in the order
   listed, with the last listed region closest to the top of the stack (the
   lowest address if `STACK_GROWS_DOWNWARD' is defined, and the highest
   address if it is not defined).  You can use a different order for a
   machine if doing so is more convenient or required for compatibility
   reasons.  Except in cases where required by standard or by a debugger,
   there is no reason why the stack layout used by GCC need agree with that
   used by other compilers for a machine.

      * A region of `current_function_pretend_args_size' bytes of
        uninitialized space just underneath the first argument
        arriving on the stack.  (This may not be at the very start of
        the allocated stack region if the calling sequence has pushed
        anything else since pushing the stack arguments.  But
        usually, on such machines, nothing else has been pushed yet,
        because the function prologue itself does all the pushing.)
        This region is used on machines where an argument may be
        passed partly in registers and partly in memory, and, in some
        cases to support the features in `varargs.h' and `stdargs.h'.

      * An area of memory used to save certain registers used by the
        function.  The size of this area, which may also include
        space for such things as the return address and pointers to
        previous stack frames, is machine-specific and usually
        depends on which registers have been used in the function.
        Machines with register windows often do not require a save
        area.

      * A region of at least SIZE bytes, possibly rounded up to an
        allocation boundary, to contain the local variables of the
        function.  On some machines, this region and the save area
        may occur in the opposite order, with the save area closer to
        the top of the stack.

      * Optionally, when `ACCUMULATE_OUTGOING_ARGS' is defined, a
        region of `current_function_outgoing_args_size' bytes to be
        used for outgoing argument lists of the function.

   Normally, it is necessary for the macros `FUNCTION_PROLOGUE' and
   `FUNCTION_EPILOGUE' to treat leaf functions specially.  The C variable
   `leaf_function' is nonzero for such a function.  */

#define FUNCTION_PROLOGUE(FILE, SIZE) (                             \
    (TARGET_NIOS_FLAT)                                              \
    ? nios_flat_function_prologue ((FILE), (SIZE),                  \
                              current_function_uses_only_leaf_regs) \
    : nios_function_prologue ((FILE), (SIZE),                       \
                              current_function_uses_only_leaf_regs) \
)


/* Define this macro as a C expression that is nonzero if the return
   instruction or the function epilogue ignores the value of the stack
   pointer; in other words, if it is safe to delete an instruction to
   adjust the stack pointer before a return from the function.

   Note that this macro's value is relevant only for functions for which
   frame pointers are maintained.  It is never safe to delete a final stack
   adjustment in a function that has no frame pointer, and the compiler
   knows this regardless of `EXIT_IGNORE_STACK'.  */

#define EXIT_IGNORE_STACK	1


/* Define this macro as a C expression that is nonzero for registers
   are used by the epilogue or the `return' pattern.  The stack and frame
   pointer registers are already be assumed to be used as needed.  */

#define EPILOGUE_USES(REGNO) ((REGNO) == GPR_RETURN)


/* A C compound statement that outputs the assembler code for exit from a
   function.  The epilogue is responsible for restoring the saved registers
   and stack pointer to their values when the function was called, and
   returning control to the caller.  This macro takes the same arguments as
   the macro `FUNCTION_PROLOGUE', and the registers to restore are
   determined from `regs_ever_live' and `CALL_USED_REGISTERS' in the same
   way.

   On some machines, there is a single instruction that does all the work
   of returning from the function.  On these machines, give that
   instruction the name `return' and do not define the macro
   `FUNCTION_EPILOGUE' at all.

   Do not define a pattern named `return' if you want the
   `FUNCTION_EPILOGUE' to be used.  If you want the target switches to
   control whether return instructions or epilogues are used, define a
   `return' pattern with a validity condition that tests the target
   switches appropriately.  If the `return' pattern's validity condition is
   false, epilogues will be used.

   On machines where functions may or may not have frame-pointers, the
   function exit code must vary accordingly.  Sometimes the code for these
   two cases is completely different.  To determine whether a frame pointer
   is wanted, the macro can refer to the variable `frame_pointer_needed'.
   The variable's value will be 1 when compiling a function that needs a
   frame pointer.

   Normally, `FUNCTION_PROLOGUE' and `FUNCTION_EPILOGUE' must treat leaf
   functions specially.  The C variable `leaf_function' is nonzero for such
   a function.

   On some machines, some functions pop their arguments on exit while
   others leave that for the caller to do.  For example, the 68020 when
   given `-mrtd' pops arguments in functions that take a fixed number of
   arguments.

   Your definition of the macro `RETURN_POPS_ARGS' decides which functions
   pop their own arguments.  `FUNCTION_EPILOGUE' needs to know what was
   decided. The variable that is called `current_function_pops_args' is the
   number of bytes of its arguments that a function should pop.  */

#define FUNCTION_EPILOGUE(FILE, SIZE) (  				\
    (TARGET_NIOS_FLAT)                                                  \
    ? nios_flat_function_epilogue                                       \
            (FILE, SIZE, current_function_uses_only_leaf_regs)          \
    : nios_function_epilogue                                            \
            (FILE, SIZE, current_function_uses_only_leaf_regs)          \
)

/* Define this macro if the function epilogue contains delay slots to which
   instructions from the rest of the function can be "moved".  The
   definition should be a C expression whose value is an integer
   representing the number of delay slots there.  */
/* #define DELAY_SLOTS_FOR_EPILOGUE	1 */

#define DELAY_SLOTS_FOR_EPILOGUE (TARGET_NIOS_FLAT ? 1 : 0)

/* A C expression that returns 1 if INSN can be placed in delay slot number
   N of the epilogue.

   The argument N is an integer which identifies the delay slot now being
   considered (since different slots may have different rules of
   eligibility). It is never negative and is always less than the number of
   epilogue delay slots (what `DELAY_SLOTS_FOR_EPILOGUE' returns).  If you
   reject a particular insn for a given delay slot, in principle, it may be
   reconsidered for a subsequent delay slot.  Also, other insns may (at
   least in principle) be considered for the so far unfilled delay slot.

   The insns accepted to fill the epilogue delay slots are put in an RTL
   list made with `insn_list' objects, stored in the variable
   `current_function_epilogue_delay_list'.  The insn for the first delay
   slot comes first in the list.  Your definition of the macro
   `FUNCTION_EPILOGUE' should fill the delay slots by outputting the insns
   in this list, usually by calling `final_scan_insn'.

   You need not define this macro if you did not define
   `DELAY_SLOTS_FOR_EPILOGUE'.  */

/* #define ELIGIBLE_FOR_EPILOGUE_DELAY(trial, slots_filled)		\
   nios_eligible_for_epilogue_delay (trial, slots_filled) */

/* }}} */

#define ELIGIBLE_FOR_EPILOGUE_DELAY(trial, slots) (                \
    (TARGET_NIOS_FLAT                                              \
    && nios_flat_eligible_for_epilogue_delay((trial), (slots)))    \
)

/* {{{ Generating Code for Profiling.  */

/* A C statement or compound statement to output to FILE some assembler
   code to call the profiling subroutine `mcount'.  Before calling, the
   assembler code must load the address of a counter variable into a
   register where `mcount' expects to find the address.  The name of this
   variable is `LP' followed by the number LABELNO, so you would generate
   the name using `LP%d' in a `fprintf'.

   The details of how the address should be passed to `mcount' are
   determined by your operating system environment, not by GNU CC.  To
   figure them out, compile a small program for profiling using the
   system's installed C compiler and look at the assembler code that
   results.

   This declaration must be present, but it can be an abort if profiling is
   not implemented.  */

#define FUNCTION_PROFILER(FILE, LABELNO)	  \
nios_function_profiler(FILE, LABELNO) 
/* abort () */
/* CBV - Replaced call to abort () to call to nios_function_profiler */
/*       Defined in nios.c                                           */

/* }}} */

/* {{{ Implementing the Varargs Macros.  */

/* If defined, is a C expression that produces the machine-specific code
   for a call to `__builtin_saveregs'.  This code will be moved to the very
   beginning of the function, before any parameter access are made.  The
   return value of this function should be an RTX that contains the value
   to use as the return of `__builtin_saveregs'.

   If this macro is not defined, the compiler will output an ordinary call
   to the library function `__builtin_saveregs'.  */

#define EXPAND_BUILTIN_SAVEREGS() nios_builtin_saveregs ()


/* Define this macro if the location where a function argument is passed
   depends on whether or not it is a named argument.

   This macro controls how the NAMED argument to `FUNCTION_ARG' is set for
   varargs and stdarg functions.  With this macro defined, the NAMED
   argument is always true for named arguments, and false for unnamed
   arguments.  If this is not defined, but `SETUP_INCOMING_VARARGS' is
   defined, then all arguments are treated as named.  Otherwise, all named
   arguments except the last are treated as named.  */

/* #define STRICT_ARGUMENT_NAMING  1 */


/* Build up the stdarg/varargs va_list type tree, assinging it to NODE.  If
   not defined, it is assumed that va_list is a void * pointer.  */

/* #define BUILD_VA_LIST_TYPE(NODE) */


/* Implement the stdarg/varargs va_start macro.  STDARG_P is non-zero if
   this is stdarg.h instead of varargs.h.  VALIST is the tree of the
   va_list variable to initialize.  NEXTARG is the machine independent
   notion of the 'next' argument after the variable arguments.  If not
   defined, a standard implementation will be defined that works for
   arguments passed on the stack.  */
#define EXPAND_BUILTIN_VA_START(stdarg, valist, nextarg)		\
  nios_va_start (stdarg, valist, nextarg)


/* Implement the stdarg/varargs va_arg macro.  VALIST is the variable of type
   va_list as a tree, TYPE is the type passed to va_arg.  */

#define EXPAND_BUILTIN_VA_ARG(valist, type)				\
  nios_va_arg (valist, type)


/* Implement the stdarg/varargs va_end macro.
   VALIST is the variable of type va_list as a tree.  */

/* #define EXPAND_BUILTIN_VA_END(VALIST) */

/* }}} */

/* {{{ Trampolines for Nested Functions.  */

/* A C expression for the size in bytes of the trampoline, as an integer.  */

#define TRAMPOLINE_SIZE nios_trampoline_size ()


/* A C statement to initialize the variable parts of a trampoline.  ADDR is
   an RTX for the address of the trampoline; FNADDR is an RTX for the
   address of the nested function; STATIC_CHAIN is an RTX for the static
   chain value that should be passed to the function when it is called.  */

#define INITIALIZE_TRAMPOLINE(ADDR, FNADDR, STATIC_CHAIN) \
  nios_initialize_trampoline (ADDR, FNADDR, STATIC_CHAIN)

/* }}} */

/* {{{ Addressing Modes.  */

/* A C expression that is 1 if the RTX X is a constant which is a valid
   address.  On most machines, this can be defined as `CONSTANT_P (X)', but
   a few machines are more restrictive in which constant addresses are
   supported.

   We cannot have constant addresses on nios.  */

#define CONSTANT_ADDRESS_P(X)	0

 
#ifdef REG_OK_STRICT
#define REG_OK_STRICT_P 1
#else
#define REG_OK_STRICT_P 0
#endif


/* A C expression that is nonzero if X (assumed to be a `reg' RTX) is valid
   for use as a base register.  For hard registers, it should always accept
   those which the hardware permits and reject the others.  Whether the
   macro accepts or rejects pseudo registers must be controlled by
   `REG_OK_STRICT' as described above.  This usually requires two variant
   definitions, of which `REG_OK_STRICT' controls the one actually used. */

#define REG_OK_FOR_BASE_P(X) nios_reg_ok_for_base_p (X, REG_OK_STRICT_P)


/* A C expression that is nonzero if X (assumed to be a `reg' RTX) is valid
   for use as an index register.

   The difference between an index register and a base register is that the
   index register may be scaled.  If an address involves the sum of two
   registers, neither one of them scaled, then either one may be labeled
   the "base" and the other the "index"; but whichever labeling is used
   must fit the machine's constraints of which registers may serve in each
   capacity. The compiler will try both labelings, looking for one that is
   valid, and will reload one or both registers only if neither labeling
   works.  */

#define REG_OK_FOR_INDEX_P(X) nios_reg_ok_for_index_p(X, REG_OK_STRICT_P)


/* A number, the maximum number of registers that can appear in a valid
   memory address.  Note that it is up to you to specify a value equal to
   the maximum number that `GO_IF_LEGITIMATE_ADDRESS' would ever accept. */

#define MAX_REGS_PER_ADDRESS 1


/* A C compound statement with a conditional `goto LABEL;' executed if X
   (an RTX) is a legitimate memory address on the target machine for a
   memory operand of mode MODE.

   It usually pays to define several simpler macros to serve as subroutines
   for this one.  Otherwise it may be too complicated to understand.

   This macro must exist in two variants: a strict variant and a non-strict
   one.  The strict variant is used in the reload pass.  It must be defined
   so that any pseudo-register that has not been allocated a hard register
   is considered a memory reference.  In contexts where some kind of
   register is required, a pseudo-register with no hard register must be
   rejected.

   The non-strict variant is used in other passes.  It must be defined to
   accept all pseudo-registers in every context where some kind of register
   is required.

   Compiler source files that want to use the strict variant of this macro
   define the macro `REG_OK_STRICT'.  You should use an `#ifdef
   REG_OK_STRICT' conditional to define the strict variant in that case and
   the non-strict variant otherwise.

   Subroutines to check for acceptable registers for various purposes (one
   for base registers, one for index registers, and so on) are typically
   among the subroutines used to define `GO_IF_LEGITIMATE_ADDRESS'.  Then
   only these subroutine macros need have two variants; the higher levels
   of macros may be the same whether strict or not.

   Normally, constant addresses which are the sum of a `symbol_ref' and an
   integer are stored inside a `const' RTX to mark them as constant.
   Therefore, there is no need to recognize such sums specifically as
   legitimate addresses.  Normally you would simply recognize any `const'
   as legitimate.

   Usually `PRINT_OPERAND_ADDRESS' is not prepared to handle constant sums
   that are not marked with `const'.  It assumes that a naked `plus'
   indicates indexing.  If so, then you *must* reject such naked constant
   sums as illegitimate addresses, so that none of them will be given to
   `PRINT_OPERAND_ADDRESS'.

   On some machines, whether a symbolic address is legitimate depends on
   the section that the address refers to.  On these machines, define the
   macro `ENCODE_SECTION_INFO' to store the information into the
   `symbol_ref', and then check for it here.  When you see a `const', you
   will have to look inside it to find the `symbol_ref' in order to
   determine the section.

   The best way to modify the name string is by adding text to the
   beginning, with suitable punctuation to prevent any ambiguity.  Allocate
   the new name in `saveable_obstack'.  You will have to modify
   `ASM_OUTPUT_LABELREF' to remove and decode the added text and output the
   name accordingly, and define `STRIP_NAME_ENCODING' to access the
   original name string.

   You can check the information stored here into the `symbol_ref' in the
   definitions of the macros `GO_IF_LEGITIMATE_ADDRESS' and
   `PRINT_OPERAND_ADDRESS'.  */

#define GO_IF_LEGITIMATE_ADDRESS(MODE, X, ADDR)    			\
  do {									\
    if (nios_go_if_legitimate_address (MODE, X, REG_OK_STRICT_P))	\
      goto ADDR;							\
  } while (0)

    
/* A C compound statement that attempts to replace X with a valid memory
   address for an operand of mode MODE.  WIN will be a C statement label
   elsewhere in the code; the macro definition may use

        GO_IF_LEGITIMATE_ADDRESS (MODE, X, WIN);

   to avoid further processing if the address has become legitimate.

   X will always be the result of a call to `break_out_memory_refs', and
   OLDX will be the operand that was given to that function to produce X.

   The code generated by this macro should not alter the substructure of X.
   If it transforms X into a more legitimate form, it should assign X
   (which will always be a C variable) a new value.

   It is not necessary for this macro to come up with a legitimate address.
   The compiler has standard ways of doing so in all cases.  In fact, it is
   safe for this macro to do nothing.  But often a machine-dependent
   strategy can generate better code.  */

#define LEGITIMIZE_ADDRESS(X, OLDX, MODE, WIN)


/* A C statement or compound statement with a conditional `goto LABEL;'
   executed if memory address X (an RTX) can have different meanings
   depending on the machine mode of the memory reference it is used for or
   if the address is valid for some modes but not others.

   Autoincrement and autodecrement addresses typically have mode-dependent
   effects because the amount of the increment or decrement is the size of
   the operand being addressed.  Some machines have other mode-dependent
   addresses. Many RISC machines have no mode-dependent addresses.

   You may assume that ADDR is a valid address for the machine.  */

#define GO_IF_MODE_DEPENDENT_ADDRESS(ADDR, LABEL)


/* A C expression that is nonzero if X is a legitimate constant for an
   immediate operand on the target machine.  You can assume that X
   satisfies `CONSTANT_P', so you need not check this.  In fact, `1' is a
   suitable definition for this macro on machines where anything
   `CONSTANT_P' is valid.  */

#define LEGITIMATE_CONSTANT_P(X) nios_legitimate_constant_p (X)

/* }}} */

/* {{{ Condition Code Status.  */

/* A C compound statement to set the components of `cc_status'
   appropriately for an insn INSN whose body is EXP.  It is this macro's
   responsibility to recognize insns that set the condition code as a
   byproduct of other activity as well as those that explicitly set
   `(cc0)'.

   This macro is not used on machines that do not use `cc0'.

   If there are insns that do not set the condition code but do alter other
   machine registers, this macro must check to see whether they invalidate
   the expressions that the condition code is recorded as reflecting.  For
   example, on the 68000, insns that store in address registers do not set
   the condition code, which means that usually `NOTICE_UPDATE_CC' can
   leave `cc_status' unaltered for such insns.  But suppose that the
   previous insn set the condition code based on location `a4@(102)' and
   the current insn stores a new value in `a4'.  Although the condition
   code is not changed by this, it will no longer be true that it reflects
   the contents of `a4@(102)'. Therefore, `NOTICE_UPDATE_CC' must alter
   `cc_status' in this case to say that nothing is known about the
   condition code value.

   The definition of `NOTICE_UPDATE_CC' must be prepared to deal with the
   results of peephole optimization: insns whose patterns are `parallel'
   RTXs containing various `reg', `mem' or constants which are just the
   operands. The RTL structure of these insns is not sufficient to indicate
   what the insns actually do.  What `NOTICE_UPDATE_CC' should do when it
   sees one is just to run `CC_STATUS_INIT'.

   A possible definition of `NOTICE_UPDATE_CC' is to call a function that
   looks at an attribute named, for example, `cc'.  This avoids having
   detailed information about patterns in two places, the `md' file and in
   `NOTICE_UPDATE_CC'.  */

#define NOTICE_UPDATE_CC(EXP, INSN) { CC_STATUS_INIT; }

/* }}} */

/* {{{ Describing Relative Costs of Operations.  */

/* A part of a C `switch' statement that describes the relative costs of
   constant RTL expressions.  It must contain `case' labels for expression
   codes `const_int', `const', `symbol_ref', `label_ref' and
   `const_double'. Each case must ultimately reach a `return' statement to
   return the relative cost of the use of that kind of constant value in an
   expression.  The cost may depend on the precise value of the constant,
   which is available for examination in X, and the rtx code of the
   expression in which it is contained, found in OUTER_CODE.

   CODE is the expression code--redundant, since it can be obtained with
   `GET_CODE (X)'.  */

#define CONST_COSTS(X, CODE, OUTER_CODE)				\
  case CONST:								\
  case LABEL_REF:							\
  case SYMBOL_REF:							\
  case CONST_DOUBLE:							\
    return COSTS_N_INSNS (2);						\
									\
  case CONST_INT:							\
    /* Make 5 bit integers really cheap */				\
    return IN_RANGE_P (INTVAL (X), 0, 31) ? 0 : COSTS_N_INSNS (2);	\


/* A C expression for the cost of moving data from a register in class FROM
   to one in class TO.  The classes are expressed using the enumeration
   values such as `GENERAL_REGS'.  A value of 4 is the default; other
   values are interpreted relative to that.

   It is not required that the cost always equal 2 when FROM is the same as
   TO; on some machines it is expensive to move between registers if they
   are not general registers.

   If reload sees an insn consisting of a single `set' between two hard
   registers, and if `REGISTER_MOVE_COST' applied to their classes returns
   a value of 2, reload does not check to ensure that the constraints of
   the insn are met.  Setting a cost of other than 2 will allow reload to
   verify that the constraints are met.  You should do this if the `movM'
   pattern's constraints do not allow such copying.  */

#define REGISTER_MOVE_COST(FROM, TO) 1


/* A C expression for the cost of moving data of mode M between a register
   and memory.  A value of 2 is the default; this cost is relative to those
   in `REGISTER_MOVE_COST'.

   If moving between registers and memory is more expensive than between
   two registers, you should define this macro to express the relative
   cost.  */

#define MEMORY_MOVE_COST(M,C,I) 4


/* A C expression for the cost of a branch instruction.  A value of 1 is
   the default; other values are interpreted relative to that.

   Here are additional macros which do not specify precise relative costs, but
   only that certain actions are more expensive than GNU CC would ordinarily
   expect.  */

#ifndef DEFAULT_BRANCH_COST
#define DEFAULT_BRANCH_COST 1
#endif

#define BRANCH_COST nios_branch_cost_int

/* Define this macro as a C expression which is nonzero if accessing less
   than a word of memory (i.e. a `char' or a `short') is no faster than
   accessing a word of memory, i.e., if such access require more than one
   instruction or if there is no difference in cost between byte and
   (aligned) word loads.

   When this macro is not defined, the compiler will access a field by
   finding the smallest containing object; when it is defined, a fullword
   load will be used if alignment permits.  Unless bytes accesses are
   faster than word accesses, using word accesses is preferable since it
   may eliminate subsequent memory access if subsequent accesses occur to
   other fields in the same word of the structure, but to different bytes.  */

#define SLOW_BYTE_ACCESS 1


/* Define this macro if it is as good or better to call a constant function
   address than to call an address kept in a register.  */

#define NO_FUNCTION_CSE


/* Define this macro if it is as good or better for a function to call
   itself with an explicit address than to call an address kept in a
   register.  */

#define NO_RECURSIVE_FUNCTION_CSE

/* }}} */

/* {{{ Dividing the output into sections.  */

/* Define this macro if references to a symbol must be treated differently
   depending on something about the variable or function named by the
   symbol (such as what section it is in).

   The macro definition, if any, is executed immediately after the rtl for
   DECL has been created and stored in `DECL_RTL (DECL)'.  The value of the
   rtl will be a `mem' whose address is a `symbol_ref'.

   The usual thing for this macro to do is to record a flag in the
   `symbol_ref' (such as `SYMBOL_REF_FLAG') or to store a modified name
   string in the `symbol_ref' (if one bit is not enough information).
 
   On the Nios, function symbols must be loaded in registers with an @h
   modifier. Set SYMBOL_REF_FLAG bit to 1 for function symbols. If the
   function has been declared in this compilation unit, mark it for short
   call relaxation (bsrr).  */

#define ENCODE_SECTION_INFO(DECL)					\
  if (TREE_CODE (DECL) == FUNCTION_DECL)				\
    {									\
      SYMBOL_REF_FLAG (XEXP (DECL_RTL (DECL), 0)) = 1;			\
      if (TREE_STATIC (DECL)						\
	  || !TREE_PUBLIC (DECL)					\
	  || TREE_ASM_WRITTEN (DECL))					\
	nios_encode_label (XEXP (DECL_RTL (DECL), 0));			\
    }

#define LOCAL_FUNCTION_MARKER	'&'


/* Decode SYM_NAME and store the real name part in VAR, sans the characters
   that encode section info.  Define this macro if `ENCODE_SECTION_INFO' alters
   the symbol's name string.  */

#define STRIP_NAME_ENCODING(VAR, SYM_NAME)				\
  (VAR) = ((SYM_NAME) + ((SYM_NAME)[0] == LOCAL_FUNCTION_MARKER		\
			 || (SYM_NAME)[0] == '*'))


/* A C statement (sans semicolon) to output to the stdio stream STREAM a
   reference in assembler syntax to a label named NAME.  This should add
   `_' to the front of the name, if that is customary on your operating
   system, as it is in most Berkeley Unix systems.  This macro is used in
   `assemble_name'.  */

#undef ASM_OUTPUT_LABELREF
#define ASM_OUTPUT_LABELREF(STREAM, NAME)				\
  do									\
    {									\
      const char *real_name;						\
      STRIP_NAME_ENCODING (real_name, NAME);				\
      asm_fprintf (STREAM, "%U%s", real_name);				\
    }									\
  while (0)


/* A C expression whose value is a string containing the assembler
   operation that should precede instructions and read-only data.  Normally
   `".text"' is right.  */

#define TEXT_SECTION_ASM_OP "\t.text"


/* A C expression whose value is a string containing the assembler
   operation to identify the following data as writable initialized data.
   Normally `".data"' is right.  */

#define DATA_SECTION_ASM_OP "\t.data"


/* If defined, a C expression whose value is a string containing the
   assembler operation to identify the following data as uninitialized
   global data.  If not defined, and neither `ASM_OUTPUT_BSS' nor
   `ASM_OUTPUT_ALIGNED_BSS' are defined, uninitialized global data will be
   output in the data section if `-fno-common' is passed, otherwise
   `ASM_OUTPUT_COMMON' will be used.  */

#define BSS_SECTION_ASM_OP "\t.bss"


/*---------------------------------------------------------------------------
		The Overall Framework of an Assembler File
---------------------------------------------------------------------------*/
/* A C string constant describing how to begin a comment in the target
   assembler language.  The compiler assumes that the comment will end at
   the end of the line.  */
#define ASM_COMMENT_START ";"

/* A C string constant for text to be output before each `asm' statement or
   group of consecutive ones.  Normally this is `"#APP"', which is a
   comment that has no effect on most assemblers but tells the GNU
   assembler that it must check the lines that follow for all valid
   assembler constructs.  */
#define ASM_APP_ON "#APP\n"

/* A C string constant for text to be output after each `asm' statement or
   group of consecutive ones.  Normally this is `"#NO_APP"', which tells
   the GNU assembler to resume making the time-saving assumptions that are
   valid for ordinary compiler output.  */
#define ASM_APP_OFF "#NO_APP\n"


/*---------------------------------------------------------------------------
			      Output of Data
---------------------------------------------------------------------------*/
/* A C statement to output to the stdio stream STREAM an assembler
   instruction to assemble a floating-point constant of `TFmode', `DFmode',
   `SFmode', `TQFmode', `HFmode', or `QFmode', respectively, whose value is
   VALUE. VALUE will be a C expression of type `REAL_VALUE_TYPE'.  Macros
   such as `REAL_VALUE_TO_TARGET_DOUBLE' are useful for writing these
   definitions. */
#define ASM_OUTPUT_DOUBLE(STREAM, VALUE)				\
  do {									\
    if (REAL_VALUE_ISINF (VALUE)					\
	|| REAL_VALUE_ISNAN (VALUE)					\
	|| REAL_VALUE_MINUS_ZERO (VALUE))				\
      {									\
	long t[2];							\
	REAL_VALUE_TO_TARGET_DOUBLE ((VALUE), t);			\
	fprintf (STREAM, "\t.word\t0x%.8lx\n\t.word\t0x%.8lx\n",	\
		 t[0] & 0xffffffff, t[1] & 0xffffffff);			\
      }									\
    else								\
      {									\
	char str[30];							\
	REAL_VALUE_TO_DECIMAL (VALUE, "%.20e", str);			\
	fprintf (STREAM, "\t.double\t%s\n", str);			\
      }									\
  } while (0)


#define ASM_OUTPUT_FLOAT(STREAM, VALUE)					\
  do {									\
    if (REAL_VALUE_ISINF (VALUE)					\
	|| REAL_VALUE_ISNAN (VALUE)					\
	|| REAL_VALUE_MINUS_ZERO (VALUE))				\
      {									\
	  long t;							\
	  REAL_VALUE_TO_TARGET_SINGLE ((VALUE), t);			\
	  fprintf (STREAM, "\t.word\t0x%.8lx\n", t & 0xffffffff);	\
      }									\
    else								\
      {									\
	  char str[30];							\
	  REAL_VALUE_TO_DECIMAL ((VALUE), "%.20e", str);		\
	  fprintf (STREAM, "\t.float\t%s\n", str);			\
      }									\
  } while (0)


/* This is how to output an assembler line defining a `char' constant.  */
#define ASM_OUTPUT_CHAR(STREAM, VALUE) nios_print_int (STREAM, VALUE, 1)

/* This is how to output an assembler line defining a `short' constant.  */
#define ASM_OUTPUT_SHORT(STREAM, VALUE) nios_print_int (STREAM, VALUE, 2)

/* This is how to output an assembler line defining an `int' constant.
   We also handle symbol output here.  */
#define ASM_OUTPUT_INT(STREAM, VALUE) nios_print_int (STREAM, VALUE, 4)

/* A C statement to output to the stdio stream STREAM an assembler
   instruction to assemble a single byte containing the number VALUE.

   This declaration must be present.  */
#define ASM_OUTPUT_BYTE(STREAM, VALUE)					\
  fprintf (STREAM, "\t%s\t0x%x\n", ASM_BYTE_OP, (VALUE))

/* These macros are defined as C string constant, describing the syntax in
   the assembler for grouping arithmetic expressions.  The following
   definitions are correct for most assemblers:

        #define ASM_OPEN_PAREN "("
        #define ASM_CLOSE_PAREN ")"  */
#define ASM_OPEN_PAREN "("
#define ASM_CLOSE_PAREN ")"

/* }}} */

/* {{{ Output and Generation of Labels.  */

/* A C statement (sans semicolon) to output to the stdio stream STREAM the
   assembler definition of a label named NAME.  Use the expression
   `assemble_name (STREAM, NAME)' to output the name itself; before and
   after that, output the additional assembler syntax for defining the
   name, and a newline.  */

#define ASM_OUTPUT_LABEL(STREAM, NAME)					\
  do {									\
    assemble_name (STREAM, NAME);					\
    fputs (":\n", STREAM);						\
  } while (0)


/* A C statement (sans semicolon) to output to the stdio stream STREAM some
   commands that will make the label NAME global; that is, available for
   reference from other files.  Use the expression `assemble_name (STREAM,
   NAME)' to output the name itself; before and after that, output the
   additional assembler syntax for making that name global, and a newline. */

#define ASM_GLOBALIZE_LABEL(STREAM, NAME)				\
  do {									\
    fputs ("\t.globl ", STREAM);					\
    assemble_name (STREAM, NAME);					\
    fputs ("\n", STREAM);						\
  } while (0)


/* A C expression to assign to OUTVAR (which is a variable of type `char *')
   a newly allocated string made from the string NAME and the number
   NUMBER, with some suitable punctuation added.  Use `alloca' to get space
   for the string.

   The string will be used as an argument to `ASM_OUTPUT_LABELREF' to
   produce an assembler label for an internal static variable whose name is
   NAME. Therefore, the string must be such as to result in valid assembler
   code. The argument NUMBER is different each time this macro is executed;
   it prevents conflicts between similarly-named internal static variables
   in different scopes.

   Ideally this string should not be a valid C identifier, to prevent any
   conflict with the user's own symbols.  Most assemblers allow periods or
   percent signs in assembler symbols; putting at least one of these
   between the name and the number will suffice.  */

#define ASM_FORMAT_PRIVATE_NAME(OUTVAR, NAME, NUMBER)			\
  do {									\
    (OUTVAR) = (char *) alloca (strlen ((NAME)) + 12);			\
    sprintf ((OUTVAR), "%s.%ld", (NAME), (long)(NUMBER));		\
  } while (0)


/* A C statement (sans semicolon) to output an
   element in the table of global constructors.  */

#undef ASM_OUTPUT_CONSTRUCTOR
#define ASM_OUTPUT_CONSTRUCTOR(FILE, NAME)                      \
  do                                                            \
    {                                                           \
      ctors_section ();                                         \
      fprintf (FILE, "\t%s\t ", ".nword");                      \
      assemble_name (FILE, NAME);                               \
      fprintf (FILE, "@h\n");                                   \
    }                                                           \
  while (0)


/* A C statement (sans semicolon) to output an
   element in the table of global destructors.  */

#undef ASM_OUTPUT_DESTRUCTOR
#define ASM_OUTPUT_DESTRUCTOR(FILE,NAME)                        \
  do                                                            \
    {                                                           \
      dtors_section ();                                         \
      fprintf (FILE, "\t%s\t ", ".nword");                      \
      assemble_name (FILE, NAME);                               \
      fprintf (FILE, "@h\n");                                    \
    }                                                           \
  while (0)

/* }}} */

/* {{{ Output of Assembler Instructions.  */

/* A C initializer containing the assembler's names for the machine
   registers, each one as a C string constant.  This is what translates
   register numbers in the compiler into assembler language.  */

#define REGISTER_NAMES							\
{									\
  "%r0",   "%g1",   "%g2",   "%g3",   "%g4",   "%g5",   "%g6",   "%g7",	\
  "%o0",   "%o1",   "%o2",   "%o3",   "%o4",   "%o5",   "%sp",   "%o7",	\
  "%l0",   "%l1",   "%l2",   "%l3",   "%l4",   "%l5",   "%l6",   "%l7",	\
  "%i0",   "%i1",   "%i2",   "%i3",   "%i4",   "%i5",   "%fp",   "%i7",	\
  "%ct10", "%ct11", "%ct12", "%ct13", "%ct14", "%ct15"			\
}


/* A C compound statement to output to stdio stream STREAM the assembler
   syntax for an instruction operand X.  X is an RTL expression.

   CODE is a value that can be used to specify one of several ways of
   printing the operand.  It is used when identical operands must be
   printed differently depending on the context.  CODE comes from the `%'
   specification that was used to request printing of the operand.  If the
   specification was just `%DIGIT' then CODE is 0; if the specification was
   `%LTR DIGIT' then CODE is the ASCII code for LTR.

   If X is a register, this macro should print the register's name.  The
   names can be found in an array `reg_names' whose type is `char *[]'.
   `reg_names' is initialized from `REGISTER_NAMES'.

   When the machine description has a specification `%PUNCT' (a `%'
   followed by a punctuation character), this macro is called with a null
   pointer for X and the punctuation character for CODE.  */

#define PRINT_OPERAND(STREAM, X, CODE) nios_print_operand (STREAM, X, CODE)


/* A C compound statement to output to stdio stream STREAM the assembler
   syntax for an instruction operand that is a memory reference whose
   address is X.  X is an RTL expression.

   On some machines, the syntax for a symbolic address depends on the
   section that the address refers to.  On these machines, define the macro
   `ENCODE_SECTION_INFO' to store the information into the `symbol_ref',
   and then check for it here.

   This declaration must be present.  */

#define PRINT_OPERAND_ADDRESS(STREAM, X) nios_print_operand_address (STREAM, X)


/* A C expression which evaluates to true if CODE is a valid punctuation
   character for use in the `PRINT_OPERAND' macro.  If
   `PRINT_OPERAND_PUNCT_VALID_P' is not defined, it means that no punctuation
   characters (except for the standard one, `%') are used in this way.  */

#define PRINT_OPERAND_PUNCT_VALID_P(CODE)	((CODE) == '#')

/* }}} */

/* {{{ Output of dispatch tables.  */

/* This macro should be provided on machines where the addresses in a
   dispatch table are relative to the table's own address.

   The definition should be a C statement to output to the stdio stream
   STREAM an assembler pseudo-instruction to generate a difference between
   two labels. VALUE and REL are the numbers of two internal labels.  The
   definitions of these labels are output using
   `ASM_OUTPUT_INTERNAL_LABEL', and they must be printed in the same way
   here.  For example,

        fprintf (STREAM, "\t.word L%d-L%d\n", VALUE, REL)  */

#define ASM_OUTPUT_ADDR_DIFF_ELT(STREAM, BODY, VALUE, REL)		\
  fprintf (STREAM, "\t.nword .L%d@h-.L%d@h\n", VALUE, REL)


/* This macro should be provided on machines where the addresses in a
   dispatch table are absolute.

   The definition should be a C statement to output to the stdio stream
   STREAM an assembler pseudo-instruction to generate a reference to a
   label.  VALUE is the number of an internal label whose definition is
   output using `ASM_OUTPUT_INTERNAL_LABEL'.  For example,

        fprintf (STREAM, "\t.word L%d\n", VALUE)  */

#define ASM_OUTPUT_ADDR_VEC_ELT(STREAM, VALUE)				\
  fprintf (STREAM, "\t.nword .L%d@h\n", VALUE)

/* }}} */

/* {{{ Assembler Commands for Exception Regions.  */

/* Define this macro to 0 if your target supports DWARF 2 frame unwind
   information, but it does not yet work with exception handling.
   Otherwise, if your target supports this information (if it defines
   `INCOMING_RETURN_ADDR_RTX' and either `UNALIGNED_INT_ASM_OP' or
   `OBJECT_FORMAT_ELF'), GCC will provide a default definition of 1.

   If this macro is defined to 1, the DWARF 2 unwinder will be the default
   exception handling mechanism; otherwise, setjmp/longjmp will be used by
   default.

   If this macro is defined to anything, the DWARF 2 unwinder will be used
   instead of inline unwinders and __unwind_function in the non-setjmp case.  */

/* #define DWARF2_UNWIND_INFO 1 */

/* }}} */

/* {{{ Assembler Commands for Alignment.  */

/* A C statement to output to the stdio stream STREAM an assembler command
   to advance the location counter to a multiple of 2 to the POWER bytes.
   POWER will be a C expression of type `int'.  */

#define ASM_OUTPUT_ALIGN(STREAM, POWER)					\
  fprintf ((STREAM), "\t.p2align %d\n", (POWER))

/* }}} */

/* {{{ Macros Affecting all Debug Formats.  */

/* A C expression that returns the DBX register number for the compiler
   register number REGNO.  In simple cases, the value of this expression
   may be REGNO itself.  But sometimes there are some registers that the
   compiler knows about and DBX does not, or vice versa.  In such cases,
   some register may need to have one number in the compiler and another
   for DBX.

   If two registers have consecutive numbers inside GNU CC, and they can be
   used as a pair to hold a multiword value, then they *must* have
   consecutive numbers after renumbering with `DBX_REGISTER_NUMBER'.
   Otherwise, debuggers will be unable to access such a pair, because they
   expect register pairs to be consecutive in their own numbering scheme.

   If you find yourself defining `DBX_REGISTER_NUMBER' in way that does not
   preserve register pairs, then what you must do instead is redefine the
   actual register numbering scheme.

   This declaration is required.  */

#define DBX_REGISTER_NUMBER(REGNO) (REGNO)


/* A C expression that returns the type of debugging output GNU CC produces
   when the user specifies `-g' or `-ggdb'.  Define this if you have
   arranged for GNU CC to support more than one format of debugging output.
   Currently, the allowable values are `DBX_DEBUG', `SDB_DEBUG',
   `DWARF_DEBUG', `DWARF2_DEBUG', and `XCOFF_DEBUG'.

   The value of this macro only affects the default debugging output; the
   user can always get a specific type of output by using `-gstabs',
   `-gcoff', `-gdwarf-1', `-gdwarf-2', or `-gxcoff'.

   Defined in svr4.h.  */

#undef PREFERRED_DEBUGGING_TYPE
#define PREFERRED_DEBUGGING_TYPE DWARF2_DEBUG


/* Define this macro if GNU CC should produce dwarf version 2 format
   debugging output in response to the `-g' option.

   To support optional call frame debugging information, you must also
   define `INCOMING_RETURN_ADDR_RTX' and either set `RTX_FRAME_RELATED_P'
   on the prologue insns if you use RTL for the prologue, or call
   `dwarf2out_def_cfa' and `dwarf2out_reg_save' as appropriate from
   `FUNCTION_PROLOGUE' if you don't.

   Defined in svr4.h.  */

#ifndef DWARF2_DEBUGGING_INFO
#define DWARF2_DEBUGGING_INFO 
#endif


/* Define this macro if addresses in Dwarf2 debugging info should have
   a size other than POINTER_SIZE / BITS_PER_UNIT.  The macro's value
   should be the size, in bytes, to use for addresses in the debugging
   info.

   This is needed on architectures where addresses are not the same
   size as pointers.  For example, some machines use word addresses
   for instructions, to squeeze in a little more space for code.  */

#define DWARF2_ADDR_SIZE 4

/* }}} */

/* {{{ Miscellaneous Parameters.  */

/* Define REAL_ARITHMETIC to use a software emulator for the target
   floating point mode.  Otherwise the host floating point mode is used. */

#define REAL_ARITHMETIC


/* An alias for a machine mode name.  This is the machine mode that
   elements of a jump-table should have.  */

#define CASE_VECTOR_MODE Pmode


/* Define this macro if operations between registers with integral mode
   smaller than a word are always performed on the entire register.  Most
   RISC machines have this property and most CISC machines do not.  */

#define WORD_REGISTER_OPERATIONS


/* Define this macro to be a C expression indicating when insns that read
   memory in MODE, an integral mode narrower than a word, set the bits
   outside of MODE to be either the sign-extension or the zero-extension of
   the data read.  Return `SIGN_EXTEND' for values of MODE for which the
   insn sign-extends, `ZERO_EXTEND' for which it zero-extends, and `NIL'
   for other modes.

   This macro is not called with MODE non-integral or with a width greater
   than or equal to `BITS_PER_WORD', so you may return any value in this
   case.  Do not define this macro if it would always return `NIL'.  On
   machines where this macro is defined, you will normally define it as the
   constant `SIGN_EXTEND' or `ZERO_EXTEND'.  */

#define LOAD_EXTEND_OP(MODE) ZERO_EXTEND


/* An alias for a tree code that is the easiest kind of division to compile
   code for in the general case.  It may be `TRUNC_DIV_EXPR', `FLOOR_DIV_EXPR',
   `CEIL_DIV_EXPR' or `ROUND_DIV_EXPR'.  These four division operators differ
   in how they round the result to an integer.  `EASY_DIV_EXPR' is used when it
   is permissible to use any of those kinds of division and the choice should
   be made on the basis of efficiency.  */

#define EASY_DIV_EXPR TRUNC_DIV_EXPR


/* A C expression which is nonzero if on this machine it is safe to "convert"
   an integer of INPREC bits to one of OUTPREC bits (where OUTPREC is smaller
   than INPREC) by merely operating on it as if it had only OUTPREC bits.

   On many machines, this expression can be 1.

   When `TRULY_NOOP_TRUNCATION' returns 1 for a pair of sizes for modes for
   which `MODES_TIEABLE_P' is 0, suboptimal code can result.  If this is the
   case, making `TRULY_NOOP_TRUNCATION' return 0 in such cases may improve
   things.  */

#define TRULY_NOOP_TRUNCATION(OUTPREC, INPREC) 1


/* The maximum number of bytes that a single instruction can move quickly
   from memory to memory.  */

#define MOVE_MAX 4


/* A C expression that is nonzero if on this machine the number of bits
   actually used for the count of a shift operation is equal to the number
   of bits needed to represent the size of the object being shifted.  When
   this macro is non-zero, the compiler will assume that it is safe to omit
   a sign-extend, zero-extend, and certain bitwise `and' instructions that
   truncates the count of a shift operation.  On machines that have
   instructions that act on bitfields at variable positions, which may
   include `bit test' instructions, a nonzero `SHIFT_COUNT_TRUNCATED' also
   enables deletion of truncations of the values that serve as arguments to
   bitfield instructions.

   If both types of instructions truncate the count (for shifts) and
   position (for bitfield operations), or if no variable-position bitfield
   instructions exist, you should define this macro.

   However, on some machines, such as the 80386 and the 680x0, truncation
   only applies to shift operations and not the (real or pretended)
   bitfield operations.  Define `SHIFT_COUNT_TRUNCATED' to be zero on such
   machines. Instead, add patterns to the `md' file that include the
   implied truncation of the shift instructions.

   You need not define this macro if it would always have the value of
   zero.  */

#define SHIFT_COUNT_TRUNCATED 1


/* An alias for the machine mode for pointers.  On most machines, define
   this to be the integer mode corresponding to the width of a hardware
   pointer; `SImode' on 32-bit machine or `DImode' on 64-bit machines.  On
   some machines you must define this to be one of the partial integer
   modes, such as `PSImode'.

   The width of `Pmode' must be at least as large as the value of
   `POINTER_SIZE'.  If it is not equal, you must define the macro
   `POINTERS_EXTEND_UNSIGNED' to specify how pointers are extended to
   `Pmode'.  */

#define Pmode ((TARGET_32BIT) ? SImode : HImode)


/* An alias for the machine mode used for memory references to functions
   being called, in `call' RTL expressions.  On most machines this should
   be `QImode'.  */

#define FUNCTION_MODE QImode


/* Define this macro to handle System V style pragmas: #pragma pack and
   #pragma weak.  Note, #pragma weak will only be supported if SUPPORT_WEAK
   is defined.

   Defined in svr4.h.  */

#define HANDLE_SYSV_PRAGMA


/* Indicate how many instructions can be issued at the same time.  */

#define ISSUE_RATE 1


/* A C statement which is executed by the Haifa scheduler at the beginning
   of each block of instructions that are to be scheduled.  STREAM is
   either a null pointer, or a stdio stream to write any debug output to.
   VERBOSE is the verbose level provided by -fsched-verbose-<n>. */

#define MD_SCHED_INIT(STREAM, VERBOSE) nios_sched_init (STREAM, VERBOSE)


/* A C statement which is executed by the Haifa scheduler after it has
   scheduled the ready list to allow the machine description to reorder it
   (for example to combine two small instructions together on VLIW
   machines).  STREAM is either a null pointer, or a stdio stream to write
   any debug output to.  VERBOSE is the verbose level provided by
   -fsched-verbose-=<n>.  READY is a pointer to the ready list of
   instructions that are ready to be scheduled.  N_READY is the number of
   elements in the ready list.  The scheduler reads the ready list in
   reverse order, starting with READY[N_READY-1] and going to READY[0]. */

#define MD_SCHED_REORDER(STREAM, VERBOSE, READY, N_READY, CLOCK, 	\
			 CAN_ISSUE_MORE)				\
  do {									\
    nios_sched_reorder (STREAM, VERBOSE, READY, N_READY);		\
    CAN_ISSUE_MORE = issue_rate;					\
  } while (0)


/* A C statement which is executed by the Haifa scheduler after it has
   scheduled an insn from the ready list.  STREAM is either a null pointer,
   or a stdio stream to write any debug output to.  VERBOSE is the verbose
   level provided by -fsched-verbose-<n>.  INSN is the instruction that was
   scheduled.  MORE is the number of instructions that can be issued in the
   current cycle.  This macro is responsible for updating the value of MORE
   (typically by (MORE)--). */

#define MD_SCHED_VARIABLE_ISSUE(STREAM, VERBOSE, INSN, MORE) 		\
  ((MORE) = nios_sched_variable_issue (STREAM, VERBOSE, INSN, MORE))


#define DONT_USE_BUILTIN_SETJMP

#define CURRENT_FUNCTION_NAME    (                                      \
    ((current_function_decl && DECL_NAME (current_function_decl))       \
     ? IDENTIFIER_POINTER (DECL_NAME (current_function_decl))           \
     : "<unknown>")                                                     \
)

/* }}} */
