;; {{{ Machine Description for the Altera Nios processor.

;; Copyright (C) 1999, 2000 Free Software Foundation, Inc.
;; Contributed by Cygnus Solutions, a Red Hat company

;; This file is part of GNU CC.

;; GNU CC is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2, or (at your option) any later
;; version.

;; GNU CC is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.

;; You should have received a copy of the GNU General Public License along
;; with GNU CC; see the file COPYING.  If not, write to the Free Software
;; Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;; See file "rtl.def" for documentation on define_insn, match_*, et. al.

;; }}}

;; {{{ Attributes.

(define_attr "length" "" (const_int 2))		;; Lengths are in bytes

;; Instruction types
;; 	call - Call instruction.
;; 	branch - Branch instruction.
;; 	return - Return instruction.
;; 	misc - Regular instructions.
(define_attr "type" "call,branch,return,misc" (const_string "misc"))

;; }}} 

;; {{{ Delay Slots.

;; The insn attribute mechanism can be used to specify the requirements for
;; delay slots, if any, on a target machine.  An instruction is said to
;; require a "delay slot" if some instructions that are physically after
;; the instruction are executed as if they were located before it.  Classic
;; examples are branch and call instructions, which often execute the
;; following instruction before the branch or call is performed.
;; 
;; On some machines, conditional branch instructions can optionally "annul"
;; instructions in the delay slot.  This means that the instruction will
;; not be executed for certain branch outcomes.  Both instructions that
;; annul if the branch is true and instructions that annul if the branch is
;; false are supported.
;; 
;; Delay slot scheduling differs from instruction scheduling in that
;; determining whether an instruction needs a delay slot is dependent only
;; on the type of instruction being generated, not on data flow between the
;; instructions.  See the next section for a discussion of data-dependent
;; instruction scheduling.
;; 
;; The requirement of an insn needing one or more delay slots is indicated via
;; the `define_delay' expression.  It has the following form:
;; 
;;
;; (define_delay TEST
;;   [DELAY-1 ANNUL-TRUE-1 ANNUL-FALSE-1
;;    DELAY-2 ANNUL-TRUE-2 ANNUL-FALSE-2
;;    ...])
;; 
;; TEST is an attribute test that indicates whether this `define_delay'
;; applies to a particular insn.  If so, the number of required delay slots
;; is determined by the length of the vector specified as the second
;; argument.  An insn placed in delay slot N must satisfy attribute test
;; DELAY-N. ANNUL-TRUE-N is an attribute test that specifies which insns
;; may be annulled if the branch is true.  Similarly, ANNUL-FALSE-N
;; specifies which insns in the delay slot may be annulled if the branch is
;; false.  If annulling is not supported for that delay slot, `(nil)'
;; should be coded.

(define_attr "ok_for_delay_slot" "false,true"
  (if_then_else (and (eq_attr "type" "!branch,call,return") 
		     (eq_attr "length" "2"))
                (const_string "true")
		(const_string "false")))

(define_delay (eq_attr "type" "call,branch,return")
  [(eq_attr "ok_for_delay_slot" "true") (nil) (nil)])

;; }}}

;; {{{ Moves.

;; Wrap moves in define_expand to prevent memory->memory moves from being
;; generated at the RTL level, which generates better code for most
;; machines which can't do mem->mem moves.

;; {{{ Integer moves.

(define_expand "movqi"
  [(set (match_operand:QI 0 "general_operand" "")
	(match_operand:QI 1 "general_operand" ""))]
  ""
  "
{
  if (!reload_in_progress
      && !reload_completed
      && !register_operand (operands[0], QImode)
      && !register_operand (operands[1], QImode))
    operands[1] = copy_to_mode_reg (QImode, operands[1]);
}")

;; (same as movhi_internal16 below)
;; kv & dvb, 2003.04.17 -- let length of qtr-int reg-to-reg
;; be 4 bytes instead of 2, so it wont get put into a delay
;; slot. (the compiler does an ext8s to promote char to int.)

(define_insn "*movqi_internal"
  [(set (match_operand:QI 0 "general_operand" "=r,r,r,r,m,r")
	(match_operand:QI 1 "general_operand"  "r,P,L,i,r,m"))]
  "register_operand (operands[0], QImode) 
   || register_operand (operands[1], QImode)"
  "* return nios_output_move (operands, insn);"
  [(set_attr "length" "4,2,2,4,16,16")]
)

;;---------------------------------------------------------------------------

(define_expand "movhi"
  [(set (match_operand:HI 0 "general_operand" "")
	(match_operand:HI 1 "general_operand" ""))]
  ""
  "
{
  if (!reload_in_progress
      && !reload_completed
      && !register_operand (operands[0], HImode)
      && !register_operand (operands[1], HImode))
    operands[1] = copy_to_mode_reg (HImode, operands[1]);
}")

;; kv & dvb, 2003.04.17 -- let length of half-int reg-to-reg
;; be 4 bytes instead of 2, so it wont get put into a delay
;; slot. (the compiler does an ext16s to promote short to int.)

(define_insn "*movhi_internal16"
  [(set (match_operand:HI 0 "general_operand" "=r,r,r,r,m,r")
	(match_operand:HI 1 "general_operand"  "r,P,L,i,r,m"))]
  "!TARGET_32BIT
   && (register_operand (operands[0], HImode) 
       || register_operand (operands[1], HImode))"
  "* return nios_output_move (operands, insn);"
  [(set_attr "length" "4,2,2,4,8,8")]
)

;; kv & dms, 2003.05.20 -- let length of half-int reg-to-reg
;; be 4 bytes instead of 2, so it wont get put into a delay
;; slot. (the compiler does an ext16s to promote short to int.)

(define_insn "*movhi_internal32"
  [(set (match_operand:HI 0 "general_operand" "=r,r,r,r,m,r")
	(match_operand:HI 1 "general_operand"  "r,P,L,i,r,m"))]
  "TARGET_32BIT
   && (register_operand (operands[0], HImode)
       || register_operand (operands[1], HImode))"
  "* return nios_output_move (operands, insn);"
  [(set_attr "length" "4,2,2,4,16,16")]
)

;;---------------------------------------------------------------------------

(define_expand "movsi"
  [(set (match_operand:SI 0 "general_operand" "")
	(match_operand:SI 1 "general_operand" ""))]
  ""
  "
{
  if (!reload_in_progress
      && !reload_completed
      && !register_operand (operands[0], SImode)
      && !register_operand (operands[1], SImode))
    operands[1] = copy_to_mode_reg (SImode, operands[1]);
}")


;; Note - it is best to only have one movsi pattern and to handle all the
;; various contingencies by the use of alternatives.  This allows reload
;; the greatest amount of flexibility (since reload will only choose
;; amongst alternatives for a selected insn, it will not replace the insn
;; with another one).

(define_insn "*movsi_internal32"
  [(set (match_operand:SI 0 "general_operand" "=r,r,r,r,r,r,m")
	(match_operand:SI 1 "general_operand"  "r,P,L,N,i,m,r"))]
  "TARGET_32BIT 
   && (register_operand (operands[0], SImode)
       || register_operand (operands[1], SImode))"
  "* return nios_output_move (operands, insn);"
  [(set_attr "length" "2,2,2,4,8,8,8")]
)

(define_insn "*movsi_internal16"
  [(set (match_operand:SI 0 "general_operand" "=r,r,r,r,r,m,r")
	(match_operand:SI 1 "general_operand"  "r,P,L,N,i,r,m"))]
  "!TARGET_32BIT 
   && (register_operand (operands[0], SImode)
       || register_operand (operands[1], SImode))"
  "* return nios_output_move (operands, insn);"
  [(set_attr "length" "2,2,2,4,8,8,8")]
)

(define_split
  [(set (match_operand:SI 0 "general_operand" "")
	(match_operand:SI 1 "general_operand" ""))]
  "!(TARGET_32BIT) && reload_completed"
  [(set (match_dup 2) (match_dup 4))
   (set (match_dup 3) (match_dup 5))]
  "
{
  /* Generate move for least significant word.  */
  operands[2] = nios_subword (operands[0], 0, HImode, SImode);
  operands[4] = nios_subword (operands[1], 0, HImode, SImode);

  /* Generate move for most significant word.  */
  operands[3] = nios_subword (operands[0], 1, HImode, SImode);
  operands[5] = nios_subword (operands[1], 1, HImode, SImode);
}")

;; }}}

;; {{{ Floating Point Moves.

;; Note - Patterns for SF mode moves are compulsory, but
;; patterns for DF are optional, as GCC can synthesise them.

(define_expand "movsf"
  [(set (match_operand:SF 0 "general_operand" "")
	(match_operand:SF 1 "general_operand" ""))]
  ""
  "
{
  if (!reload_in_progress
      && !reload_completed
      && GET_CODE (operands[0]) == MEM
      && GET_CODE (operands[1]) == MEM)
    operands[1] = copy_to_mode_reg (SFmode, operands[1]);
}")

(define_insn "*movsf_internal32"
  [(set (match_operand:SF 0 "general_operand" "=r,r,r,r,&r,m")
	(match_operand:SF 1 "general_operand"  "r,L,N,i,m,r"))]
  "TARGET_32BIT
   && (!memory_operand (operands[0], SFmode)
       || !memory_operand (operands[1], SFmode))"
  "* return nios_output_move (operands, insn);"
  [(set_attr "length" "2,2,4,8,8,8")]
)

;; The early clobbers are needed to fix a regression in
;; gcc.c-torture/execute/ieee/fp-cmp-2.c
;; This avoids the compiler emitting code like 'ldp %l1, [%l1, 0]'

(define_insn "*movsf_internal16"
  [(set (match_operand:SF 0 "general_operand" "=r,r,r,r,m,&r")
	(match_operand:SF 1 "general_operand"  "r,L,N,i,r,m"))]
  "!TARGET_32BIT
   && (!memory_operand (operands[0], SFmode)
       || !memory_operand (operands[1], SFmode))"
  "* return nios_output_move (operands, insn);"
  [(set_attr "length" "2,2,4,8,8,8")]
)

(define_split
  [(set (match_operand:SF 0 "general_operand" "")
	(match_operand:SF 1 "general_operand" ""))]
  "!(TARGET_32BIT) && reload_completed"
  [(set (match_dup 2) (match_dup 4))
   (set (match_dup 3) (match_dup 5))]
  "
{
  /* Generate move for least significant word.  */
  operands[2] = nios_subword (operands[0], 0, HImode, SFmode);
  operands[4] = nios_subword (operands[1], 0, HImode, SFmode);

  /* Generate move for most significant word.  */
  operands[3] = nios_subword (operands[0], 1, HImode, SFmode);
  operands[5] = nios_subword (operands[1], 1, HImode, SFmode);
}")

(define_expand "movdf"
  [(set (match_operand:DF 0 "general_operand" "")
	(match_operand:DF 1 "general_operand" ""))]
  "TARGET_32BIT"
  "
{
  /* One of the ops has to be in a register or 0 */
  if (!register_operand (operand0, DFmode)
      && !register_operand (operand1, DFmode))
    operands[1] = copy_to_mode_reg (DFmode, operand1);
}")

(define_insn "*movdf_internal"
  [(set (match_operand:DF 0 "nonimmediate_operand" "=r,m")
	(match_operand:DF 1 "general_operand" 	   "rim,r"))]
  "((TARGET_32BIT) && ! (memory_operand (operands[0], DFmode)
         && memory_operand (operands[1], DFmode)))"
  "#"
)

(define_split
  [(set (match_operand:DF 0 "general_operand" "")
	(match_operand:DF 1 "general_operand" ""))]
  "(TARGET_32BIT && (reload_completed || reload_in_progress))"

  [(set (match_dup 2) (match_dup 3))
   (set (match_dup 4) (match_dup 5))
  ]

  "{
    /* figure out what precisely to put into operands 2, 3, 4, and 5 */
    nios_split_words (SImode, DFmode, operands);
  }"
)

;; }}}
;; }}}

;; {{{ Bit manipulation.

;; {{{ Sign extensions.

(define_insn "extendqihi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(sign_extend:HI (match_operand:QI 1 "register_operand" "0")))]
  ""
  "sext8\\t%0"
  [(set_attr "length" "2")])

(define_insn "extendqisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(sign_extend:SI (match_operand:QI 1 "register_operand" "0")))]
  "TARGET_32BIT"
  "sext8\\t%0";
  [(set_attr "length" "2")])

(define_insn "extendhisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(sign_extend:SI (match_operand:HI 1 "register_operand" "0")))]
  "TARGET_32BIT"
  "sext16\\t%0"
  [(set_attr "length" "2")])


;; }}}

;; {{{ Zero extensions.

(define_insn "zero_extendqihi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(zero_extend:HI (match_operand:QI 1 "register_operand" "0")))]
  "TARGET_ZERO_EXTEND"
  "ext8s\\t%0, 0"
  [(set_attr "length" "2")])

(define_insn "zero_extendqisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(zero_extend:SI (match_operand:QI 1 "register_operand" "0")))]
  "TARGET_32BIT && TARGET_ZERO_EXTEND"
  "ext8s\\t%0, 0";
  [(set_attr "length" "2")])

(define_insn "zero_extendhisi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(zero_extend:SI (match_operand:HI 1 "register_operand" "0")))]
  "TARGET_32BIT && TARGET_ZERO_EXTEND"
  "ext16s\\t%0, 0"
  [(set_attr "length" "2")])


;; }}}

;; {{{ Bit extractions.

(define_insn "extvsi4"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(sign_extract:SI (match_operand:SI 1 "register_operand" "r")
			 (match_operand:SI 2 "const_int_operand" "L")
			 (match_operand:SI 3 "const_int_operand" "L")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_EXTV"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_EXTV, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "extvhi4"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(sign_extract:HI (match_operand:HI 1 "register_operand" "r")
			 (match_operand:HI 2 "const_int_operand" "L")
			 (match_operand:HI 3 "const_int_operand" "L")))]
  "!TARGET_32BIT && TARGET_USER_OPCODE_EXTV"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_EXTV, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "extzvsi4"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(zero_extract:SI (match_operand:SI 1 "register_operand" "r")
			 (match_operand:SI 2 "const_int_operand" "L")
			 (match_operand:SI 3 "const_int_operand" "L")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_EXTZV"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_EXTZV, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "extzvhi4"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(zero_extract:HI (match_operand:HI 1 "register_operand" "r")
			 (match_operand:HI 2 "const_int_operand" "L")
			 (match_operand:HI 3 "const_int_operand" "L")))]
  "!TARGET_32BIT && TARGET_USER_OPCODE_EXTZV"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_EXTZV, operands, insn);"
  [(set_attr "length" "6")]
)

;; }}}

;; {{{ Bit insertions.

(define_insn "insvsi4"
  [(set (zero_extract:SI (match_operand:SI 0 "register_operand" "=r")
			 (match_operand:SI 1 "const_int_operand" "L")
			 (match_operand:SI 2 "const_int_operand" "L"))
	(match_operand:SI 3 "register_operand" "r"))]
  "TARGET_32BIT && TARGET_USER_OPCODE_INSV"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_INSV, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "insvhi4"
  [(set (zero_extract:HI (match_operand:HI 0 "register_operand" "=r")
			 (match_operand:HI 1 "const_int_operand" "L")
			 (match_operand:HI 2 "const_int_operand" "L"))
	(match_operand:HI 3 "register_operand" "r"))]
  "TARGET_32BIT && TARGET_USER_OPCODE_INSV"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_INSV, operands, insn);"
  [(set_attr "length" "6")]
)

;; }}}

;; {{{ Find first set bit.

(define_insn "ffssi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ffs:SI (match_operand:SI 1 "register_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_FFS"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_FFS, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "ffshi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(ffs:HI (match_operand:HI 1 "register_operand" "r")))]
  "!TARGET_32BIT && TARGET_USER_OPCODE_FFS"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_FFS, operands, insn);"
  [(set_attr "length" "6")]
)

;; }}}

;; }}}

;; {{{ Integer arithmetic.

;; {{{ Addition.

(define_insn "addhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r,r,r")
	(plus:HI (match_operand:HI 1 "register_operand" "%0,%0,%0,%0,%0")
		 (match_operand:HI 2 "nonmemory_operand" "r,L,N,M,I")))]
  "!TARGET_32BIT"
  "@
   add\\t%0, %2
   addi\\t%0, %2
   pfx\\t%%hi(%2)\;addi\\t%0, %%lo(%2)
   subi\\t%0, %n2
   pfx\\t%%hi(%n2)\;subi\\t%0, %%lo(%n2)
   #"
  [(set_attr "length" "2,2,4,2,4")]
)

;; 32-bit addition. Only used on nios-32. In 16-bit mode we use __addsi3.

(define_insn "addsi3_32bit"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r,r")
	(plus:SI (match_operand:SI 1 "register_operand" "%0,%0,%0,%0,%0")
		 (match_operand:SI 2 "nonmemory_operand" "r,L,N,M,I")))]
  "TARGET_32BIT"
  "@
   add\\t%0, %2
   addi\\t%0, %2
   pfx\\t%%hi(%2)\;addi\\t%0, %%lo(%2)
   subi\\t%0, %n2
   pfx\\t%%hi(%n2)\;subi\\t%0, %%lo(%n2)
   #"
  [(set_attr "length" "2,2,4,2,4")]
)


;; We want to avoid using explicit registers; reload won't tell us
;; if it has to spill them and may generate incorrect code in such
;; cases.
;;
;; So we call out to a library routine to perform 32bit add or
;; subtract operations.
;;
;; operand2 must be nonmemory_operand so that we will accept CONST_INTs
;; during initial code generation.

(define_expand "addsi3"
  [(set (match_operand:SI 0 "register_operand" "")
	(plus:SI (match_operand:SI 1 "register_operand" "")
		 (match_operand:SI 2 "nonmemory_operand" "")))]
  ""
  "
{
  if (!TARGET_32BIT && rtx_equal_function_value_matters)
    {
      rtx ret, insns;

      start_sequence ();
      ret = emit_library_call_value (gen_rtx_SYMBOL_REF (Pmode, \"__addsi3\"),
				     NULL_RTX, 1, SImode, 2, operands[1],
				     SImode, operands[2], SImode);
      insns = get_insns ();
      end_sequence ();
      emit_libcall_block (insns, operands[0], ret,
			  gen_rtx_PLUS (SImode, operands[1], operands[2]));
      DONE;
    }
  else if (!TARGET_32BIT)
    FAIL;
}")

(define_insn "addsf3_usr"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(plus:SF (match_operand:SF 1 "register_operand" "0")
		(match_operand:SF 2 "nonmemory_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_FADD"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_FADD, operands, insn);"
  [(set_attr "length" "6")]
)

;; }}}

;; {{{ Subtraction.

(define_insn "subhi3"
  [(set (match_operand:HI 0 "general_operand" "=r,r,r")
	(minus:HI (match_operand:HI 1 "general_operand" "0,0,0")
		  (match_operand:HI 2 "general_operand" "r,L,N")))]
  "!TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_SUB, operands, 0, 2, insn);"
  [(set_attr "length" "2,2,4")]
)


;; 32-bit subtraction. Only used in 32-bit mode. For nios-16 we use
;; __subsi3.

(define_insn "subsi3_32bit"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(minus:SI (match_operand:SI 1 "register_operand" "0,0,0")
		  (match_operand:SI 2 "nonmemory_operand" "r,L,N")))]
  "TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_SUB, operands, 0, 2, insn);"
  [(set_attr "length" "2,2,4")]
)


;; Expander for nios-16. Call __subsi3 when subtracting multi-word
;; operands.

(define_expand "subsi3"
  [(set (match_operand:SI 0 "register_operand" "")
	(minus:SI (match_operand:SI 1 "register_operand" "")
		  (match_operand:SI 2 "register_operand" "")))]
  ""
  "
{
  if (!TARGET_32BIT && rtx_equal_function_value_matters)
    {
      rtx ret, insns;

      start_sequence ();
      ret = emit_library_call_value (gen_rtx_SYMBOL_REF (Pmode, \"__subsi3\"),
				     NULL_RTX, 1, SImode, 2, operands[1],
				     SImode, operands[2], SImode);
      insns = get_insns ();
      end_sequence ();
      emit_libcall_block (insns, operands[0], ret,
			  gen_rtx_MINUS (SImode, operands[1], operands[2]));
      DONE;
    }
  else if (!TARGET_32BIT)
    FAIL;
}")


;; Expander for nios-16. Call __subdi3 when subtracting multi-word
;; operands.

(define_expand "subdi3"
  [(set (match_operand:DI 0 "register_operand" "")
	(minus:SI (match_operand:DI 1 "register_operand" "")
		  (match_operand:DI 2 "register_operand" "")))]
  "!TARGET_32BIT"
  "
{
  if (rtx_equal_function_value_matters)
    {
      rtx ret, insns;

      start_sequence ();
      ret = emit_library_call_value (gen_rtx_SYMBOL_REF (Pmode, \"__subdi3\"),
				     NULL_RTX, 1, DImode, 2, operands[1],
				     DImode, operands[2], DImode);
      insns = get_insns ();
      end_sequence ();
      emit_libcall_block (insns, operands[0], ret,
			  gen_rtx_MINUS (DImode, operands[1], operands[2]));
      DONE;
    }
  else
    FAIL;
}")

(define_insn "subsf3_usr"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(minus:SF (match_operand:SF 1 "register_operand" "0")
		  (match_operand:SF 2 "nonmemory_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_FSUB"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_FSUB, operands, insn);"
  [(set_attr "length" "6")]
)

;; }}}

;; {{{ Multiplication.

(define_expand "mulhisi3_internal"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
        (minus:SI
          (sign_extend:SI (match_operand:HI 1 "register_operand" "%0,r"))
          (sign_extend:SI (match_operand 2 "const_int_operand" "J,K"))))]
  "!TARGET_NEW_MULHISI"
  "")

(define_expand "mulhisi3_internal_using_mult"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
        (mult:SI
          (sign_extend:SI (match_operand:HI 1 "register_operand" "%0,r"))
          (sign_extend:SI (match_operand 2 "const_int_operand" "J,K"))))]
  "TARGET_NEW_MULHISI"
  "")


;; Expander to avoid calling __mulsi3 when multiplying HI operands.

(define_expand "mulhi3"
  [(set (match_operand:HI 0 "register_operand" "")
	(mult:SI (match_operand:HI 1 "general_operand" "")
		  (match_operand:HI 2 "general_operand" "")))]
  "!TARGET_USER_OPCODE_MUL"
  "
{
  if (rtx_equal_function_value_matters)
    {
      rtx ret, insns;

      start_sequence ();
      ret = emit_library_call_value (gen_rtx_SYMBOL_REF (Pmode, \"__mulhi3\"),
				     NULL_RTX, 1, HImode, 2, operands[1],
				     HImode, operands[2], HImode);
      insns = get_insns ();
      end_sequence ();
      emit_libcall_block (insns, operands[0], ret,
			  gen_rtx_MULT (HImode, operands[1], operands[2]));
      DONE;
    }
  else
    FAIL;
}")

(define_insn "mulsi3_usr"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(mult:SI (match_operand:SI 1 "register_operand" "0")
		 (match_operand:SI 2 "nonmemory_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_MUL"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_MUL, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "mulhi3_usr"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(mult:HI (match_operand:HI 1 "register_operand" "0")
		 (match_operand:HI 2 "nonmemory_operand" "r")))]
  "!TARGET_32BIT && TARGET_USER_OPCODE_MUL"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_MUL, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "mulsf3_usr"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(mult:SF (match_operand:SF 1 "register_operand" "0")
		 (match_operand:SF 2 "nonmemory_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_FMUL"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_FMUL, operands, insn);"
  [(set_attr "length" "6")]
)

;; }}}

;; {{{ Division.

(define_insn "divsi3_usr"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(div:SI (match_operand:SI 1 "register_operand" "0")
		 (match_operand:SI 2 "nonmemory_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_DIV"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_DIV, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "divhi3_usr"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(div:HI (match_operand:HI 1 "register_operand" "0")
		 (match_operand:HI 2 "nonmemory_operand" "r")))]
  "!TARGET_32BIT && TARGET_USER_OPCODE_DIV"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_DIV, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "udivsi3_usr"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(udiv:SI (match_operand:SI 1 "register_operand" "0")
		 (match_operand:SI 2 "nonmemory_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_UDIV"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_UDIV, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "udivhi3_usr"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(udiv:HI (match_operand:HI 1 "register_operand" "0")
		 (match_operand:HI 2 "nonmemory_operand" "r")))]
  "!TARGET_32BIT && TARGET_USER_OPCODE_UDIV"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_UDIV, operands, insn);"
  [(set_attr "length" "6")]
)

;; }}}

;; {{{ Modulus.

(define_insn "modsi3_usr"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(mod:SI (match_operand:SI 1 "register_operand" "0")
		 (match_operand:SI 2 "nonmemory_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_MOD"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_MOD, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "modhi3_usr"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(mod:HI (match_operand:HI 1 "register_operand" "0")
		 (match_operand:HI 2 "nonmemory_operand" "r")))]
  "!TARGET_32BIT && TARGET_USER_OPCODE_MOD"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_MOD, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "umodsi3_usr"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(umod:SI (match_operand:SI 1 "register_operand" "0")
		 (match_operand:SI 2 "nonmemory_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_UMOD"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_UMOD, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "umodhi3_usr"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(umod:HI (match_operand:HI 1 "register_operand" "0")
		 (match_operand:HI 2 "nonmemory_operand" "r")))]
  "!TARGET_32BIT && TARGET_USER_OPCODE_UMOD"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_UMOD, operands, insn);"
  [(set_attr "length" "6")]
)

;; }}}

;; {{{ Minimum.

(define_insn "sminsi3_usr"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(smin:SI (match_operand:SI 1 "register_operand" "0")
		 (match_operand:SI 2 "nonmemory_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_SMIN"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_SMIN, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "sminhi3_usr"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(smin:HI (match_operand:HI 1 "register_operand" "0")
		 (match_operand:HI 2 "nonmemory_operand" "r")))]
  "!TARGET_32BIT && TARGET_USER_OPCODE_SMIN"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_SMIN, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "uminsi3_usr"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(umin:SI (match_operand:SI 1 "register_operand" "0")
		 (match_operand:SI 2 "nonmemory_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_UMIN"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_UMIN, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "uminhi3_usr"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(umin:HI (match_operand:HI 1 "register_operand" "0")
		 (match_operand:HI 2 "nonmemory_operand" "r")))]
  "!TARGET_32BIT && TARGET_USER_OPCODE_UMIN"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_UMIN, operands, insn);"
  [(set_attr "length" "6")]
)

;; }}}

;; {{{ Maximum.

(define_insn "smaxsi3_usr"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(smax:SI (match_operand:SI 1 "register_operand" "0")
		 (match_operand:SI 2 "nonmemory_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_SMAX"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_SMAX, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "smaxhi3_usr"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(smax:HI (match_operand:HI 1 "register_operand" "0")
		 (match_operand:HI 2 "nonmemory_operand" "r")))]
  "!TARGET_32BIT && TARGET_USER_OPCODE_SMAX"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_SMAX, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "umaxsi3_usr"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(umax:SI (match_operand:SI 1 "register_operand" "0")
		 (match_operand:SI 2 "nonmemory_operand" "r")))]
  "TARGET_32BIT && TARGET_USER_OPCODE_UMAX"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_UMAX, operands, insn);"
  [(set_attr "length" "6")]
)

(define_insn "umaxhi3_usr"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(umax:HI (match_operand:HI 1 "register_operand" "0")
		 (match_operand:HI 2 "nonmemory_operand" "r")))]
  "!TARGET_32BIT && TARGET_USER_OPCODE_UMAX"
  "* return nios_output_user_opcode (NIOS_USER_OPCODE_UMAX, operands, insn);"
  [(set_attr "length" "6")]
)

;; }}}

;; {{{ Negation.

(define_insn "negsi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(neg:SI (match_operand:SI 1 "register_operand" "0")))]
  "TARGET_32BIT"
  "neg\\t%0")

(define_insn "neghi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(neg:HI (match_operand:HI 1 "register_operand" "0")))]
  "!TARGET_32BIT"
  "neg\\t%0")

;; }}}

;; {{{ Absolute value.

(define_insn "abssi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(abs:SI (match_operand:SI 1 "register_operand" "0")))]
  "TARGET_32BIT"
  "abs\\t%0"
)

(define_insn "abshi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(abs:HI (match_operand:HI 1 "register_operand" "0")))]
  "!TARGET_32BIT"
  "abs\\t%0"
)
;; }}}

;; }}}

;; {{{ Integer Shifts and Rotates.

;; Arithmetic Shift Left.

(define_insn "ashlsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(ashift:SI (match_operand:SI 1 "register_operand" "0,0")
		   (match_operand:SI 2 "nonmemory_operand" "r,L")))]
  "TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_LSL, operands, 0, 2, insn);"
)

(define_insn "ashlhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(ashift:HI (match_operand:HI 1 "register_operand" "0,0")
		   (match_operand:HI 2 "nonmemory_operand" "r,J")))]
  "!TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_LSL, operands, 0, 2, insn);"
)


;; Arithmetic Shift Right.

(define_insn "ashrsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(ashiftrt:SI (match_operand:SI 1 "register_operand" "0,0")
		     (match_operand:SI 2 "nonmemory_operand" "r,L")))]
  "TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_ASR, operands, 0, 2, insn);"
)

(define_insn "ashrhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(ashiftrt:HI (match_operand:HI 1 "register_operand" "0,0")
		     (match_operand:HI 2 "nonmemory_operand" "r,J")))]
  "!TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_ASR, operands, 0, 2, insn);"
)


;; Logical Shift Right.

(define_insn "lshrsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(lshiftrt:SI (match_operand:SI 1 "register_operand" "0,0")
		     (match_operand:SI 2 "nonmemory_operand" "r,L")))]
  "TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_LSR, operands, 0, 2, insn);"
)

(define_insn "lshrhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(lshiftrt:HI (match_operand:HI 1 "register_operand" "0,0")
		     (match_operand:HI 2 "nonmemory_operand" "r,J")))]
  "!TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_LSR, operands, 0, 2, insn);"
)

;; }}}

;; {{{ Integer Logical operations.

;; Note on lengths: The nios does not have immediate opcodes for `and',
;; `or', and `xor'. The only way to distinguish immediate operands is using
;; a `pfx' before the operation. Therefore, immediate operands need two
;; instructions (length == 4).

;; Logical AND.
;; `O' constraint checks for AND-to-ANDN conversion opportunity.

(define_insn "andsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(and:SI (match_operand:SI 1 "register_operand" "%0,%0,%0")
		(match_operand:SI 2 "nonmemory_operand" "r,N,O")))]
  "TARGET_32BIT"
  "@
   and\\t%0, %2
   pfx\\t%%hi(%2)\;and\\t%0, %%lo(%2)
   pfx\\t%%hi(%N2)\;andn\\t%0, %%lo(%N2)
   #"
  [(set_attr "length" "2,4,4")]
)

(define_insn "andhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(and:HI (match_operand:HI 1 "register_operand" "%0,%0,%0")
		(match_operand:HI 2 "nonmemory_operand" "r,L,N")))]
  "!TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_AND, operands, 0, 2, insn);"
  [(set_attr "length" "2,4,4")]
)

;; Special patterns for ANDN when the first operand of "and" is inverted.

(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r")
        (and:SI (not:SI (match_operand:SI 1 "register_operand" "r"))
	        (match_operand:SI 2 "register_operand" "0")))]
  "TARGET_32BIT && TARGET_ANDN"
  "andn\\t%0, %1"
  [(set_attr "length" "2")]
)

(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r")
        (and:HI (not:HI (match_operand:HI 1 "register_operand" "r"))
	        (match_operand:HI 2 "register_operand" "0")))]
  "!TARGET_32BIT && TARGET_ANDN"
  "andn\\t%0, %1"
  [(set_attr "length" "2")]
)


;; Inclusive OR.

(define_insn "iorsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(ior:SI (match_operand:SI 1 "register_operand" "%0,%0,%0")
		(match_operand:SI 2 "nonmemory_operand" "r,L,N")))]
  "TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_OR, operands, 0, 2, insn);"
  [(set_attr "length" "2,4,4")]
)

(define_insn "iorhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(ior:HI (match_operand:HI 1 "register_operand" "%0,%0,%0")
		(match_operand:HI 2 "nonmemory_operand" "r,L,N")))]
  "!TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_OR, operands, 0, 2, insn);"
  [(set_attr "length" "2,4,4")]
)


;; Exclusive OR, 32-bit integers.

(define_insn "xorsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(xor:SI (match_operand:SI 1 "register_operand" "%0,%0,%0")
		(match_operand:SI 2 "nonmemory_operand" "r,L,N")))]
  "TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_XOR, operands, 0, 2, insn);"
  [(set_attr "length" "2,4,4")]
)

(define_insn "xorhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(xor:HI (match_operand:HI 1 "register_operand" "%0,%0,%0")
		(match_operand:HI 2 "nonmemory_operand" "r,L,N")))]
  "!TARGET_32BIT"
  "* return nios_output_2op_insn (NIOS_XOR, operands, 0, 2, insn);"
  [(set_attr "length" "2,4,4")]
)


;; One's complement.

(define_insn "one_cmplsi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(not:SI (match_operand:SI 1 "register_operand" "0")))]
  "TARGET_32BIT"
  "not\\t%0"
)

(define_insn "one_cmplhi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(not:HI (match_operand:HI 1 "register_operand" "0")))]
  "!TARGET_32BIT"
  "not\\t%0"
)

;; }}}

;; {{{ Comparisons.

;; Store the comparison operands and emit RTL for compare-and-branch in the
;; conditional branching instructions.

(define_expand "cmpsi"
  [(set (reg:CC 0)
	(compare:CC (match_operand:SI 0 "register_operand" "")
		    (match_operand:SI 1 "nonmemory_operand" "")))]
  "TARGET_32BIT"
  "
{
  nios_compare_op0 = operands[0];
  nios_compare_op1 = operands[1];
  DONE;
}")

(define_expand "cmphi"
  [(set (reg:CC 0)
	(compare:CC (match_operand:HI 0 "register_operand" "")
		    (match_operand:HI 1 "nonmemory_operand" "")))]
  "!TARGET_32BIT"
  "
{
  nios_compare_op0 = operands[0];
  nios_compare_op1 = operands[1];
  DONE;
}")

;; }}}

;; {{{ Branches.

;; {{{ Expanders.

;; define_expands called by the machine independent part of the compiler to
;; allocate a new comparison register.  Each of these named patterns must
;; be present, and they cannot be amalgamated into one pattern.

(define_expand "beq"
  [(set (pc)
	(if_then_else (eq (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  /* Set operands from the compare expander.  */
  operands[1] = nios_compare_op0;
  operands[2] = nios_compare_op1;
  /* Fall through and generate default code.  */
}")


(define_expand "bne"
  [(set (pc)
        (if_then_else (ne (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  /* Set operands from the compare expander.  */
  operands[1] = nios_compare_op0;
  operands[2] = nios_compare_op1;
  /* Fall through and generate default code.  */
}")

(define_expand "blt"
  [(set (pc)
        (if_then_else (lt (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  /* Set operands from the compare expander.  */
  operands[1] = nios_compare_op0;
  operands[2] = nios_compare_op1;
  /* Fall through and generate default code.  */
}")

(define_expand "ble"
  [(set (pc)
	(if_then_else (le (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  /* Set operands from the compare expander.  */
  operands[1] = nios_compare_op0;
  operands[2] = nios_compare_op1;
  /* Fall through and generate default code.  */
}")

(define_expand "bgt"
  [(set (pc)
        (if_then_else (gt (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  /* Set operands from the compare expander.  */
  operands[1] = nios_compare_op0;
  operands[2] = nios_compare_op1;
  /* Fall through and generate default code.  */
}")

(define_expand "bge"
  [(set (pc)
        (if_then_else (ge (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  /* Set operands from the compare expander.  */
  operands[1] = nios_compare_op0;
  operands[2] = nios_compare_op1;
  /* Fall through and generate default code.  */
}")

(define_expand "bltu"
  [(set (pc)
        (if_then_else (ltu (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  /* Set operands from the compare expander.  */
  operands[1] = nios_compare_op0;
  operands[2] = nios_compare_op1;
  /* Fall through and generate default code.  */
}")

(define_expand "bleu"
  [(set (pc)
        (if_then_else (leu (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  /* Set operands from the compare expander.  */
  operands[1] = nios_compare_op0;
  operands[2] = nios_compare_op1;
  /* Fall through and generate default code.  */
}")

(define_expand "bgtu"
  [(set (pc)
        (if_then_else (gtu (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  /* Set operands from the compare expander.  */
  operands[1] = nios_compare_op0;
  operands[2] = nios_compare_op1;
  /* Fall through and generate default code.  */
}")

(define_expand "bgeu"
  [(set (pc)
        (if_then_else (geu (match_dup 1) (match_dup 2))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "
{
  /* Set operands from the compare expander.  */
  operands[1] = nios_compare_op0;
  operands[2] = nios_compare_op1;
  /* Fall through and generate default code.  */
}")

;; }}}

;; {{{ Actual branches.  

;; We must allow for the (label_ref) and the (pc) to be
;; swapped.  If they are swapped, it reverses the sense of the branch.

(define_insn "branch_true_32"
  [(set (pc)
	(if_then_else 
	  (match_operator 0 "comparison_operator" 
			    [(match_operand:SI 1 "register_operand" "r,r,r")
			     (match_operand:SI 2 "nonmemory_operand" "r,L,N")])
	  (label_ref (match_operand 3 "" ""))
	  (pc)))]
  "TARGET_32BIT"
  "* return nios_output_cond_branch (1, insn, operands);"
  [(set_attr "type" "branch")
   (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 3) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 3) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 12)))]
)

(define_insn "branch_false_32"
  [(set (pc)
	(if_then_else 
	  (match_operator 0 "comparison_operator" 
			    [(match_operand:SI 1 "register_operand" "r,r,r")
			     (match_operand:SI 2 "nonmemory_operand" "r,L,N")])
	  (pc)
	  (label_ref (match_operand 3 "" ""))))]
  "TARGET_32BIT"
  "* return nios_output_cond_branch (0, insn, operands);"
  [(set_attr "type" "branch")
   (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 3) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 3) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 12)))]
)


(define_insn "branch_true_16"
  [(set (pc)
	(if_then_else 
	  (match_operator 0 "comparison_operator" 
			    [(match_operand:HI 1 "register_operand" "r,r,r")
			     (match_operand:HI 2 "nonmemory_operand" "r,L,N")])
	  (label_ref (match_operand 3 "" ""))
	  (pc)))]
  "!TARGET_32BIT"
  "* return nios_output_cond_branch (1, insn, operands);"
  [(set_attr "type" "branch")
   (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 3) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 3) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 12)))]
)

(define_insn "branch_false_16"
  [(set (pc)
	(if_then_else 
	  (match_operator 0 "comparison_operator" 
			    [(match_operand:HI 1 "register_operand" "r,r,r")
			     (match_operand:HI 2 "nonmemory_operand" "r,L,N")])
	  (pc)
	  (label_ref (match_operand 3 "" ""))))]
  "!TARGET_32BIT"
  "* return nios_output_cond_branch (0, insn, operands);"
  [(set_attr "type" "branch")
   (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 3) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 3) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 12)))]
)

;; }}}

;; {{{ Bit test branch instructions. 

;; For each mode (SI, HI) we need 4 patterns:
;; one for EQ and true-branch, one for EQ and false-branche, one for NE and
;; true branch, and, one for NE and false-branch.

(define_insn "bittest_1_true_32"
  [(set (pc)
	(if_then_else
	     (ne (zero_extract:SI (match_operand:SI 0 "register_operand" "r")
				  (const_int 1)
				  (match_operand 1 "const_int_operand" "L"))
	     (const_int 0))
	 (label_ref (match_operand 2 "" ""))
	 (pc)))]
  "TARGET_32BIT"
  "* return nios_output_bit_test_branch (1, insn, operands, NE);"
  [(set_attr "type" "branch")
   (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 2) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 2) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 12)))]
)

(define_insn "bittest_1_false_32"
  [(set (pc)
	(if_then_else
	     (ne (zero_extract:SI (match_operand:SI 0 "register_operand" "r")
				  (const_int 1)
				  (match_operand 1 "const_int_operand" "L"))
	     (const_int 0))
	 (pc)
	 (label_ref (match_operand 2 "" ""))))]
  "TARGET_32BIT"
  "* return nios_output_bit_test_branch (0, insn, operands, NE);"
  [(set_attr "type" "branch")
   (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 2) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 2) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 12)))]
)

(define_insn "bittest_0_true_32"
  [(set (pc)
	(if_then_else
	     (eq (zero_extract:SI (match_operand:SI 0 "register_operand" "r")
				  (const_int 1)
				  (match_operand 1 "const_int_operand" "L"))
	     (const_int 0))
	 (label_ref (match_operand 2 "" ""))
	 (pc)))]
  "TARGET_32BIT"
  "* return nios_output_bit_test_branch (1, insn, operands, EQ);"
  [(set_attr "type" "branch")
   (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 2) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 2) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 12)))]
)

(define_insn "bittest_0_false_32"
  [(set (pc)
	(if_then_else
	     (eq (zero_extract:SI (match_operand:SI 0 "register_operand" "r")
				  (const_int 1)
				  (match_operand 1 "const_int_operand" "L"))
	     (const_int 0))
	 (pc)
	 (label_ref (match_operand 2 "" ""))))]
  "TARGET_32BIT"
  "* return nios_output_bit_test_branch (0, insn, operands, EQ);"
  [(set_attr "type" "branch")
   (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 2) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 2) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 12)))]
)


(define_insn "bittest_1_true_16"
  [(set (pc)
	(if_then_else
	     (ne (zero_extract:HI (match_operand:HI 0 "register_operand" "r")
				  (const_int 1)
				  (match_operand 1 "const_int_operand" "L"))
	     (const_int 0))
	 (label_ref (match_operand 2 "" ""))
	 (pc)))]
  "!TARGET_32BIT"
  "* return nios_output_bit_test_branch (1, insn, operands, NE);"
  [(set_attr "type" "branch")
   (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 2) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 2) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 12)))]
)

(define_insn "bittest_1_false_16"
  [(set (pc)
	(if_then_else
	     (ne (zero_extract:HI (match_operand:HI 0 "register_operand" "r")
				  (const_int 1)
				  (match_operand 1 "const_int_operand" "L"))
	     (const_int 0))
	 (pc)
	 (label_ref (match_operand 2 "" ""))))]
  ""
  "* return nios_output_bit_test_branch (0, insn, operands, NE);"
  [(set_attr "type" "branch")
   (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 2) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 2) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 12)))]
)


(define_insn "bittest_0_true_16"
  [(set (pc)
	(if_then_else
	     (eq (zero_extract:HI (match_operand:HI 0 "register_operand" "r")
				  (const_int 1)
				  (match_operand 1 "const_int_operand" "L"))
	     (const_int 0))
	 (label_ref (match_operand 2 "" ""))
	 (pc)))]
  "!TARGET_32BIT"
  "* return nios_output_bit_test_branch (1, insn, operands, EQ);"
  [(set_attr "type" "branch")
   (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 2) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 2) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 12)))]
)

(define_insn "bittest_0_false_16"
  [(set (pc)
	(if_then_else
	     (eq (zero_extract:HI (match_operand:HI 0 "register_operand" "r")
				  (const_int 1)
				  (match_operand 1 "const_int_operand" "L"))
	     (const_int 0))
	 (pc)
	 (label_ref (match_operand 2 "" ""))))]
  "!TARGET_32BIT"
  "* return nios_output_bit_test_branch (0, insn, operands, EQ);"
  [(set_attr "type" "branch")
   (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 2) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 2) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 12)))]
)

;; }}}

;; }}}

;; {{{ Call and branch instructions.

;; Subroutine call instruction returning no value.  Operand 0 is the
;; function to call; operand 1 is the number of bytes of arguments pushed
;; (in mode `SImode', except it is normally a `const_int'); operand 2 is
;; the number of registers used as operands.
;;
;; On most machines, operand 2 is not actually stored into the RTL pattern.
;; It is supplied for the sake of some RISC machines which need to put this
;; information into the assembler code; they can put it in the RTL instead
;; of operand 1.

(define_expand "call"
  [(parallel [(call (match_operand 0 "" "")
		    (match_operand 1 "" ""))
	      (clobber (match_dup 2))])]
  ""
  "
{
  rtx call_reg;
  rtx link_reg;
  rtx sym;

  link_reg = gen_rtx_REG (Pmode, GPR_LINK);

  if (! register_operand (operands[0], Pmode))
    {
      sym = XEXP (operands[0], 0);

      /* If the function symbol is not for a local function, force a copy
	 into a new register. Otherwise, just reserve a register for
	 `bsrr'.  */
      if (! nios_local_function_operand (sym, Pmode))
	call_reg = copy_to_mode_reg (Pmode, sym);
      else
	call_reg = gen_reg_rtx (Pmode);
    }
  else
    {
      sym = NULL_RTX;
      call_reg = operands[0];
    }

  /* If the function symbol is defined within this compilation unit,
     we might be able to use a shorter call sequence with `bsr', so we
     emit `bsrr' which will be relaxed by the assembler.  */
  if (nios_local_function_operand (sym, Pmode))
    emit_call_insn (gen_call_internal_sym (sym, call_reg, link_reg));
  else
    emit_call_insn (gen_call_internal (call_reg, operands[1], link_reg));

  DONE;
}")


(define_insn "call_internal"
  [(call (mem:QI (match_operand 0 "register_operand" "r"))
	 (match_operand 1 "" ""))
   (clobber (match_operand 2 "register_operand" "=l"))]
  ""
  "call\\t%0%#"
  [(set_attr "type" "call")]
)

(define_insn "call_internal_sym"
  [(call (mem:QI (match_operand 0 "nios_local_function_operand" ""))
	 (match_operand 1 "register_operand" "r"))
   (clobber (match_operand 2 "register_operand" "=l"))]
  ""
  "bsrr\\t%1, %F0%#"
  [(set_attr "type" "call")
   (set_attr "length" "10")]
)



;; Subroutine call instruction returning a value.  Operand 0 is the hard
;; register in which the value is returned.  There are three more operands, the
;; same as the three operands of the `call' instruction (but with numbers
;; increased by one).
;;
;; Subroutines that return `BLKmode' objects use the `call' insn.

(define_expand "call_value"
  [(parallel [(set (match_operand 0 "" "")
		   (call (match_operand 1 "general_operand" "")
		         (match_operand 2 "general_operand" "")))
	      (clobber (match_dup 3))])]
  ""
  "
{
  rtx call_reg;
  rtx link_reg;
  rtx sym;

  link_reg = gen_rtx_REG (Pmode, GPR_LINK);

  if (! register_operand (operands[1], Pmode))
    {
      sym = XEXP (operands[1], 0);

      /* If the function symbol is not for a local function, force a copy
	 into a new register. Otherwise, just reserve a register for
	 `bsrr'.  */
      if (! nios_local_function_operand (sym, Pmode))
	call_reg = copy_to_mode_reg (Pmode, sym);
      else
	call_reg = gen_reg_rtx (Pmode);
    }
  else
    {
      sym = NULL_RTX;
      call_reg = operands[1];
    }

  /* If the function symbol is defined within this compilation unit,
     we might be able to use a shorter call sequence with `bsr', so we
     emit `bsrr' which will be relaxed by the assembler.  */

  if (nios_local_function_operand (sym, Pmode))
    emit_call_insn (gen_call_value_internal_sym (operands[0], 
						 sym, 
						 call_reg, 
						 link_reg));
  else
    emit_call_insn (gen_call_value_internal (operands[0],
					     call_reg,
					     operands[2], 
					     link_reg));

  DONE;
}")


(define_insn "call_value_internal"
  [(set (match_operand 0 "register_operand" "=r")
	(call (mem:QI (match_operand 1 "register_operand" "r"))
	      (match_operand 2 "" "")))
   (clobber (match_operand 3 "register_operand" "=l"))]
  ""
  "call\\t%1%#"
  [(set_attr "type" "call")]
)

(define_insn "call_value_internal_sym"
  [(set (match_operand 0 "register_operand" "=r")
	(call (mem:QI (match_operand 1 "nios_local_function_operand" ""))
	      (match_operand 2 "register_operand" "r")))
   (clobber (match_operand 3 "register_operand" "=l"))]
  ""
  "bsrr\\t%2, %F1%#"
  [(set_attr "type" "call")
   (set_attr "length" "10")]
)

;; Direct jump.

(define_insn "jump"
 [(set (pc) (label_ref (match_operand 0 "" "")))]
 ""
 "* return nios_output_branch (insn, operands);"
 [(set_attr "type" "branch")
  (set (attr "length") 
        (if_then_else (and (ge (minus (match_dup 0) (pc)) (const_int -1024)) 
		           (le (minus (match_dup 0) (pc)) (const_int 1023)))
		      (const_int 2) 
		      (const_int 8)))]
)


;; Indirect jump through a register.

(define_insn "indirect_jump"
  [(set (pc) (match_operand 0 "register_operand" "r"))]
  ""
  "jmp\\t%0%#"
  [(set_attr "type" "branch")
   (set_attr "length" "4")]
)

;; Instruction to jump to a variable address.  This is a low-level
;; capability which can be used to implement a dispatch table when there is
;; no `casesi' pattern.  Either the 'casesi' pattern or the 'tablejump'
;; pattern, or both, MUST be present in this file.
;;
;; This pattern requires two operands: the address or offset, and a label
;; which should immediately precede the jump table.  If the macro
;; `CASE_VECTOR_PC_RELATIVE' is defined then the first operand is an offset
;; which counts from the address of the table; otherwise, it is an absolute
;; address to jump to.  In either case, the first operand has mode `Pmode'.
;;
;; The `tablejump' insn is always the last insn before the jump table it
;; uses. Its assembler code normally has no need to use the second operand,
;; but you should incorporate it in the RTL pattern so that the jump
;; optimizer will not delete the table as unreachable code.

(define_insn "tablejump"
  [(set (pc) (match_operand 0 "register_operand" "r"))
   (use (label_ref (match_operand 1 "" "")))]
  ""
  "jmp\\t%0%#"
  [(set_attr "type" "branch")]
)

;; }}}

;; {{{ Miscellaneous patterns.

;; No operation, needed in case the user uses -g but not -O.

(define_insn "nop"
  [(const_int 0)]
  ""
  "nop"
)

;; Pseudo instruction that prevents the scheduler from moving code above this
;; point.

(define_insn "blockage"
  [(unspec_volatile [(const_int 0)] 0)]
  ""
  ""
)

;; Trap instruction to allow usage of the __builtin_trap function

(define_insn "trap"
  [(trap_if (const_int 1) (const_int 0))]
  ""
  "trap")
;; }}}
