/* Start of header file for FS2 Extensions to the MDI specification  (mdiext.h) */

#ifndef MDIEXT_Specification_Definitions
#define MDIEXT_Specification_Definitions

/*
	This is the FS2 extensions to the MDI specification.
*/

/*
	From mdi.h:

    To build MDILib:
        Define MDI_LIB before #include "mdi.h"
        Include mdi.def in the link on Windows hosts.

    To build an MDI application (debugger):
        Compile mdiinit.c and include it in your link
        Make a call to 
           int MDIInit(char *MDIdllpathandname, HMODULE *handle)
        to explicitly load the specified MDILib before making any other MDI calls.
*/

#include "windows.h" // dvb 2003, how did this ever work?

#include "mdi.h"	//need standard defines

/* Values for cpuType member of MdiHardwareConfigT */

#define MDI_CPUTYPE_UNKNOWN     0
#define MDI_CPU_NIOS32          1
#define MDI_CPU_NIOS16          2

/* Structure for MDIGetHardwareConfig */
typedef struct {
  MDIUint32 cpuType;             // cpu type (see definitions above)
  MDIUint32 numSwBkpts;          // number of software breakpoints
  MDIUint32 numInstBkpts;        // number of instruction breakpoints
  MDIUint32 numDataBkpts;        // number of data breakpoints (triggers)
  MDIUint32 instAddrBits;        // number of bits in the instruction memory address
  MDIUint32 dataAddrBits;        // number of bits in the data memory address
  MDIUint32 dataBits;            // width of data bus
  MDIUint32 dbrkTrace;           // Do data bkpts have trace control capability? 0=no, 1=yes
  MDIUint32 dbrkPairs;           // Can data bkpts be combined in pairs? 0=no, 1=yes
  MDIUint32 dataTrace;           // Does the hardware have data trace capability? 0=no, 1=yes
  MDIUint32 onchipTrace;         // number of words in on-chip trace (0 if none)
  MDIUint32 offchipTrace;        // number of words in off-chip trace (0 if none)
  MDIUint32 irqNumber;           // Interrupt number assigned to hardware by SOPC Builder
} MDIHardwareConfigT;

typedef enum {
	NIOS_BYTEBLASTER,
	NIOS_BLACKBOX,
	NIOS_BLACKBOX_T,
	NIOS_USER_CHOICE,
} MDI_NIOS_TYPE;

typedef MDIInt32  (__stdcall *MDICBOUTPUT)(MDIHandleT Device, MDIInt32 Type, char *Buffer, MDIInt32 Count);
typedef MDIInt32 (__stdcall *MDICBINPUT)(MDIHandleT Device, MDIInt32 Type, MDIInt32 Mode, char **Buffer, MDIInt32 *Count);


/* Function Prototypes */

#ifdef __cplusplus
extern "C" {
#endif

#if defined( MDI_LIB )
/* MDILib, do extern function declarations */
#define yf(str) extern int __stdcall str
#elif defined( MDILOAD_DEFINE )
/* mdiinit.c, do function pointer defintions */
#define yf(str) int (__stdcall *str)
#else
/* debugger, do extern function pointer declarations */
#define yf(str) extern int (__stdcall *str)
#endif


/****** Get Hardware Configuration ******/
	yf(MDIGetHardwareConfig)(MDIHandleT, MDIHardwareConfigT *config);
/* MDIFS2ReadSetup assumes debugger will allocate a buffer at least 128 characters in length! */
	yf(MDIFs2ReadSetup)(const char *, char **value);
	yf(MDIFs2WriteSetup)(const char *, const char *value);
	yf(MDIConnectDialog)(HWND parent, MDI_NIOS_TYPE boxType);
	yf(MDILoggingFunctions)(MDICBINPUT input, MDICBOUTPUT output);
	
#undef yf

#ifdef __cplusplus
}
#endif

#endif

/* End of header file for MDIEXT (mdiext.h) */

