/* nios simulator support code
   Copyright (C) 1998, 1999 Free Software Foundation, Inc.
   Contributed by Cygnus Solutions.

This file is part of the GNU simulators.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#define WANT_CPU
#define WANT_CPU_NIOS16BF


#include "sim-main.h"
#include "cgen-mem.h"
#include "cgen-ops.h"

static void nios16bf_h_status_set_no_cwp (SIM_CPU *current_cpu, UHI newval);

/* Convert gdb dedicated register number to actual dr reg number.  */

static int
decode_gdb_dr_regnum (int gdb_regnum)
{
#ifdef FIXME

#endif
  abort ();
}

/* The contents of BUF are in target byte order.  */

int
nios16bf_fetch_register (SIM_CPU *current_cpu, int rn, unsigned char *buf, int len)
{
  if (FIRST_GR_REGNUM <= rn && rn <= LAST_GR_REGNUM)
    {
      SETTHI (buf, nios16bf_h_m16_gr_get (current_cpu, rn - FIRST_GR_REGNUM));
      return sizeof (HI);
    }
  else if (rn == PC_REGNUM)
    {
      SETTUSI (buf, nios16bf_h_pc_get (current_cpu));
      return sizeof (USI);
    }
  else if (FIRST_CTL_REGNUM <= rn && rn <= LAST_CTL_REGNUM)
    {
      SETTUHI (buf, nios16bf_h_m16_ctl_get (current_cpu, 
					     rn - FIRST_CTL_REGNUM));
      return sizeof (UHI);
    }
  else 
    return 0;
}

/* The contents of BUF are in target byte order.  */

int
nios16bf_store_register (SIM_CPU *current_cpu, int rn, unsigned char *buf, int len)
{
  if (FIRST_GR_REGNUM <= rn && rn <= LAST_GR_REGNUM)
    {
      nios16bf_h_m16_gr_set (current_cpu, rn, GETTHI (buf));
      return sizeof (HI);
    }
  else if (rn == PC_REGNUM)
    {
      nios16bf_h_pc_set (current_cpu, GETTUSI (buf));
      return sizeof (USI);
    }
  else if (FIRST_CTL_REGNUM <= rn && rn <= LAST_CTL_REGNUM)
    {
      nios16bf_h_m16_ctl_set (current_cpu, 
			       rn - FIRST_CTL_REGNUM,
			       GETTUHI (buf));
      return sizeof (UHI);
    }
  else
    return 0;
}

/* Do a pseudo power-on-reset.
   USERLAND_P is non-zero to prepare to run a user-land program.  */

void
nios16bf_cold_reset (SIM_CPU *current_cpu, int userland_p)
{
  int i;

  /* Initialize the status register so TE bit is on and IPRI is 63. 
     This has to be careful as we don't want to set the CWP such
     that the old window is "swapped out". */
  nios16bf_h_status_set_no_cwp (current_cpu, 0xfe00);

  /* Initialize cwp directly (bypassing SET_H_CWP and SET_H_STATUS) as we just
     want to initialize things, not "swap" the current window out.  */
  CPU (h_cwp) = 31;

  /* Set high window to 31 and low window to 1. */

  SET_H_M16_CTL (2, (31 << 5) | (1));

  SET_H_PC (0);

  for (i = 0; i < 32; ++i)
    SET_H_M16_GR (i, 0);
}

/* Do a warm reset (the reset trap).  */

void
nios16bf_warm_reset (SIM_CPU *current_cpu)
{
  /* FIXME: unimplemented yet */
}

/* Cover fns to access the status bits.  */

UHI
nios16bf_h_status_get_handler (SIM_CPU *current_cpu)
{
  int status = (  (GET_H_CBIT () << 0)
	        | (GET_H_ZBIT () << 1)
	        | (GET_H_VBIT () << 2)
	        | (GET_H_NBIT () << 3)
		| (GET_H_CWP  () << 4)
		| (GET_H_IPRI () << 9)
		| (GET_H_TEBIT() << 15)
	       );

  return status;
}

void
nios16bf_h_status_set_handler (SIM_CPU *current_cpu, UHI newval)
{
  int status = newval;

  SET_H_CBIT ((status & 1) != 0);
  SET_H_ZBIT ((status & 2) != 0);
  SET_H_VBIT ((status & 4) != 0);
  SET_H_NBIT ((status & 8) != 0);
  SET_H_TEBIT((status & 0x8000) != 0);
  SET_H_CWP  ((status & 0x01F0) >> 4);
  SET_H_IPRI ((status & 0x7E00) >> 9);
}

static void
nios16bf_h_status_set_no_cwp (SIM_CPU *current_cpu, UHI newval)
{
  int status = newval;

  SET_H_CBIT ((status & 1) != 0);
  SET_H_ZBIT ((status & 2) != 0);
  SET_H_VBIT ((status & 4) != 0);
  SET_H_NBIT ((status & 8) != 0);
  SET_H_TEBIT((status & 0x8000) != 0);
  SET_H_IPRI ((status & 0x7E00) >> 9);
}

void
nios16bf_h_cwp_set_handler (SIM_CPU *current_cpu, UQI new)
{
  int old = GET_H_CWP ();

  if (new >= GET_NWINDOWS ())
    abort ();

  CPU (h_cwp) = new;

  if (GET_H_WBIT ())
    SET_H_OLD_CWP (old);
  else
    {
      /* Swap current values out of `current_regs'.
         Do this even if old == new.  */
      nios16bf_swapout_regwin (current_cpu, old);

      /* Swap new values into `current_regs'.  */
      if (old != new)
        nios16bf_swapin_regwin (current_cpu, new);
    }
}

/* register window functions */

/* Swap out the values in `current_regs'.  */

void
nios16bf_swapout_regwin (SIM_CPU *current_cpu, int win)
{
  int n = 8 * sizeof (HI);

  memcpy (REAL_OREGS (current_cpu, win), CURRENT_OREGS (current_cpu), n);
  memcpy (REAL_LREGS (current_cpu, win), CURRENT_LREGS (current_cpu), n);
  memcpy (REAL_IREGS (current_cpu, win), CURRENT_IREGS (current_cpu), n);
}

/* Swap int values for `current_regs'.  */

void
nios16bf_swapin_regwin (SIM_CPU *current_cpu, int win)
{
  int n = 8 * sizeof (HI);

  memcpy (CURRENT_OREGS (current_cpu), REAL_OREGS (current_cpu, win), n);
  memcpy (CURRENT_LREGS (current_cpu), REAL_LREGS (current_cpu, win), n);
  memcpy (CURRENT_IREGS (current_cpu), REAL_IREGS (current_cpu, win), n);
}

/* Create a new window.  We assume there is room.  */

void
nios16bf_save_regwin (SIM_CPU *current_cpu)
{
  SET_H_CWP (NEXT_WIN (GET_H_CWP ()));
}

/* Pop a window.  We assume no traps possible.  */

void
nios16bf_restore_regwin (SIM_CPU *current_cpu)
{
  SET_H_CWP (PREV_WIN (GET_H_CWP ()));
}

/* Flush the register windows to memory.
   This is necessary, for example, when we want to walk the stack in gdb.
   NO_ERRORS_P is non-zero if memory faults must be avoided.  This is important
   when returning to gdb, the processor has "stopped".

   ??? At present we only handle user programs.  */

void
nios16bf_flush_regwins (SIM_CPU *current_cpu, IADDR pc, int no_errors_p)
{
  int i, win;
  int count = GET_NWINDOWS ();
  int hi_window = HI_WINDOW(GET_H_M16_CTL (H_CTL_WVALID));

  /* Flush the current window cache.  */
  nios16bf_swapout_regwin (current_cpu, GET_H_CWP ());

  /* For each register window that is marked valid, flush it to memory.
     We start at the current window and move upwards on the stack.  */

  for (i = 0, win = GET_H_CWP (); i < count, win <= hi_window; i++, win = PREV_WIN (win))
    {
      nios16bf_flush_regwin (current_cpu, pc, win, no_errors_p);
    }
}

/* Flush a register window's contents out to memory.
   This function is used in three contexts:
   - to handle TRAP_WINDOW_OVERFLOW exceptions, raised by `save' instructions
   - to handle a TRAP_FLUSH_REGWIN exception, which flushes all the currently
     valid windows
   - to spill the register windows to memory when the debugger stops the
     program, so the debugger can see the contents of windows other than
     the current one.

   In all but the last context, any errors encountered should be
   signalled normally.  However, if a window's SP points to bogus
   memory, we don't want to signal an error, because that would
   inhibit debugging; the user's code might put things right again
   later.  So, no_errors_p is true when this function is called to
   suspend the simulator.  */
void
nios16bf_flush_regwin (SIM_CPU *cpu, IADDR pc, int win, int no_errors_p)
{
  int i;
  UHI sp,fp,addr;
  /* Fetch pointers to lregs and iregs for this frame.  */
  HI *lregs = REAL_LREGS (cpu, win);
  HI *iregs = REAL_IREGS (cpu, win);
  SIM_DESC sd = CPU_STATE (cpu);

  /* Fetch values of sp,fp for this frame.  */
  sp = REAL_OREGS (cpu, win) [6];
  fp = REAL_IREGS (cpu, win) [6];

  /* Exit early if there'd be a memory fault but we can't have any errors.  */
  if (no_errors_p)
    {
      /* Check if sp and fp indicate a proper save may not have been done.  */
      if (fp <= sp
	  || fp - sp < 16 * sizeof (HI))
	return;
      /* sp misaligned? */
      if (sp & 1)
	return;
    }

  /* Use sim_core_write_aligned_N here to handle endian conversions.  */

  addr = sp;
  for (i = 0; i < 8; i++)
    {
      if (no_errors_p)
	{
	  char reg[2];
	  SETTHI (reg, lregs[i]);
	  if (sim_core_write_buffer (sd, cpu, write_map, reg, addr, 2) == 0)
	    return;
	}
      else
	sim_core_write_aligned_2 (cpu, pc, write_map, addr, lregs[i]);
      addr += 2;
    }
  for (i = 0; i < 8; i++)
    {
      if (no_errors_p)
	{
	  char reg[2];
	  SETTHI (reg, iregs[i]);
	  if (sim_core_write_buffer (sd, cpu, write_map, reg, addr, 2) == 0)
	    return;
	}
      else
	sim_core_write_aligned_2 (cpu, pc, write_map, addr, iregs[i]);
      addr += 2;
    }
}


/* Restore the contents of a single register window.

   This function is very similar to nios16bf_load_regwin, except that
   it is designed for use when resuming the simulator after being
   stopped by the debugger.  It differs from ..._load_regwin as
   follows:

   - It never reports a memory access error, but silently fails
     instead.  This isn't ideal, but I think it's an appropriate
     consequence of GDB not having access to the simulator's true
     state, i.e., the windows.

   - It uses each window's SP to find the save area, not the FP, since
     we're not in the middle of executing a "restore" instruction when
     we call this function.

   The two functions are different enough that it seemed reasonable to
   not bother to share code.  */

static void
nios16bf_reload_regwin (SIM_CPU *cpu, int win)
{
  SIM_DESC sd = CPU_STATE (cpu);

  /* Fetch value of sp for spilled window.  */
  UHI addr = REAL_OREGS (cpu, win) [6];

  /* Fetch pointers to lregs and iregs for this frame.  */
  HI *lregs = REAL_LREGS (cpu, win);
  HI *iregs = REAL_IREGS (cpu, win);

  int i;

  /* If restoring the memory would encounter any errors, just return
     early.  Check for errors early.  */

  /* sp misaligned?  */
  if (addr & 1)
    return;

  /* Use sim_core_read_buffer, which returns an error code if
     something goes wrong, instead of sim_core_read_aligned_N, which
     would raise an exception.  */
  for (i = 0; i < 8; i++, addr += 2)
    {
      char reg[2];
      if (sim_core_read_buffer (sd, cpu, read_map, reg, addr, 2) == 0)
	return;
      lregs[i] = GETTHI (reg);
    }
  for (i = 0; i < 8; i++, addr += 2)
    {
      char reg[2];
      if (sim_core_read_buffer (sd, cpu, read_map, reg, addr, 2) == 0)
	return;
      iregs[i] = GETTHI (reg);
    }
}


/* Reload all the register windows from the stack.  
   Whenever we suspend execution, we save all the registers to their
   stack slots so that GDB can see them.  Before resuming execution,
   we reload those contents, so that any changes GDB made will take
   effect.  */
void
nios16bf_reload_regwins (SIM_CPU *current_cpu)
{
  /* The current window has its proper contents in CURRENT_[ILO]REGS,
     so we don't mess with that.  We restore all higher-numbered
     windows (i.e., those corresponding to stack frames below the
     current one) from memory, up to HI_WINDOW.  */
  UHI wvalid = GET_H_M16_CTL (H_CTL_WVALID);
  int cwp = GET_H_CWP ();

  if (WINDOW_VALID_P (cwp, wvalid))
    {
      int i;
  
      /* In order to discover where each window was saved, we need to
	 have the proper values installed in the next window's
	 REAL_OREGS, so nios16bf_load_regwin can find the appropriate SP.

	 To get things started, we have to flush the CURRENT regs out
	 to the REAL regs; any register modifications GDB has made
	 affect CURRENT, not REAL.  */
  
      nios16bf_swapout_regwin (current_cpu, cwp);

      for (i = PREV_WIN (cwp); WINDOW_VALID_P (i, wvalid); i = PREV_WIN (i))
	nios16bf_reload_regwin (current_cpu, i);
    }
}



/* Restore the contents of a single register window.
   This function is used as the handler for TRAP_WINDOW_UNDERFLOW in
   the user environment.  */
void
nios16bf_load_regwin (SIM_CPU *cpu, IADDR pc, int win)
{
  int i;

  /* Fetch value of sp for spilled window.  The window we want is actually
     hi_window + 1, which doesn't exist.  This is because we shifted  
     all windows left when we spilled.  Thus, to get the sp, we
     use the fp of win which is hi_window. */

  UHI addr = REAL_IREGS (cpu, win) [6];
  /* Fetch pointers to lregs and iregs for this frame.  */
  HI *lregs = REAL_LREGS (cpu, win);
  HI *iregs = REAL_IREGS (cpu, win);
  SIM_DESC sd = CPU_STATE (cpu);

  /* Use sim_core_read_aligned_N here to handle endian conversions.  */
      
  for (i = 0; i < 8; i++)
    {
      lregs[i] = sim_core_read_aligned_2 (cpu, pc, read_map, addr);
      addr += 2;
    }
  for (i = 0; i < 8; i++)
    {
      iregs[i] = sim_core_read_aligned_2 (cpu, pc, read_map, addr);
      addr += 2;
    }
}

/* Save/restore insns.  */

/* Handle the save instruction.  */

void
nios16bf_do_save (SIM_CPU *current_cpu, IADDR pc, UQI i8v)
{
  int oldwin,newwin,fp,wim;

  /* FIXME: Watch for stack overflow if user prog.  */

  /* Determine new window number and see if it is valid. */
  oldwin = GET_H_CWP ();
  newwin = NEXT_WIN (oldwin);

  if (! WINDOW_VALID_P (newwin, GET_H_M16_CTL (H_CTL_WVALID)))
    nios16bf_window_overflow (current_cpu, pc);

  /* Switch to the new window.  */
  nios16bf_save_regwin (current_cpu);
  fp = GET_H_M16_GR (H_GR_FP);
  SET_H_M16_GR (H_GR_SP, fp - i8v * 2);

  if (TRACE_INSN_P (current_cpu)) /* FIXME */
    {
      trace_result (current_cpu, "sp", 'x', GET_H_M16_GR (H_GR_SP));
      trace_result (current_cpu, "fp", 'x', GET_H_M16_GR (H_GR_FP));
      trace_result (current_cpu, "cwp", 'x', GET_H_CWP ());
    }
}

/* Handle the restore instruction.  */

void
nios16bf_do_restore (SIM_CPU *current_cpu, IADDR pc)
{
  int oldwin,newwin,wim;

  /* Determine new window number and see if it is valid. */
  oldwin = GET_H_CWP ();
  newwin = PREV_WIN (oldwin);

  if (! WINDOW_VALID_P (newwin, GET_H_M16_CTL (H_CTL_WVALID)))
    nios16bf_window_underflow (current_cpu, pc);

  /* Switch to the previous window.  */
  nios16bf_restore_regwin (current_cpu);

  if (TRACE_INSN_P (current_cpu)) /* FIXME */
    {
      trace_result (current_cpu, "sp", 'x', GET_H_M16_GR (H_GR_SP));
      trace_result (current_cpu, "fp", 'x', GET_H_M16_GR (H_GR_FP));
      trace_result (current_cpu, "cwp", 'x', GET_H_CWP ());
    }
}

/* Handle the skps instruction. */
int
nios16bf_check_cc_mask (SIM_CPU *current_cpu, UQI i4w)
{
  switch (i4w)
    {
    case CC_C:
      return (GET_H_CBIT ());
    case CC_NC:
      return (GET_H_CBIT () == 0);
    case CC_NZ:
      return (GET_H_ZBIT () == 0);
    case CC_Z:
      return (GET_H_ZBIT ());
    case CC_GE:
      return ((GET_H_NBIT () ^ GET_H_VBIT ()) == 0);
    case CC_LT:
      return (GET_H_NBIT () ^ GET_H_VBIT ());
    case CC_GT:
      return ((GET_H_ZBIT () | (GET_H_NBIT () ^ GET_H_VBIT ())) == 0);
    case CC_LE:
      return (GET_H_ZBIT () | (GET_H_NBIT () ^ GET_H_VBIT ()));
    case CC_LS:
      return (GET_H_CBIT () | GET_H_ZBIT ());
    case CC_HI:
      return ((GET_H_CBIT () | GET_H_ZBIT ()) == 0);
    case CC_MI:
      return (GET_H_NBIT ());
    case CC_PL:
      return (GET_H_NBIT () == 0);
    case CC_V:
      return (GET_H_VBIT ());
    case CC_NV:
      return (GET_H_VBIT () == 0);
    }

  /* should never get here */
  abort ();
  return 0;
}
      

#if WITH_PROFILE_MODEL_P

/* FIXME: Some of these should be inline or macros.  Later.  */

/* Initialize cycle counting for an insn.
   FIRST_P is non-zero if this is the first insn in a set of parallel
   insns.  */

void
nios16bf_model_insn_before (SIM_CPU *cpu, int first_p)
{
  MODEL_NIOS16_1_DATA *d = CPU_MODEL_DATA (cpu);
  d->load_regs_pending = 0;
}

/* Record the cycles computed for an insn.
   LAST_P is non-zero if this is the last insn in a set of parallel insns,
   and we update the total cycle count.
   CYCLES is the cycle count of the insn.  */

void
nios16bf_model_insn_after (SIM_CPU *cpu, int last_p, int cycles)
{
  PROFILE_DATA *p = CPU_PROFILE_DATA (cpu);
  MODEL_NIOS16_1_DATA *d = CPU_MODEL_DATA (cpu);

  PROFILE_MODEL_TOTAL_CYCLES (p) += cycles;
  PROFILE_MODEL_CUR_INSN_CYCLES (p) = cycles;
  d->load_regs = d->load_regs_pending;
}

static INLINE int
check_load_stall (SIM_CPU *cpu, int regno)
{
  const MODEL_NIOS16_1_DATA *d = CPU_MODEL_DATA (cpu);
  UINT load_regs = d->load_regs;

  if (regno != -1
      && (load_regs & (1 << regno)) != 0)
    {
      PROFILE_DATA *p = CPU_PROFILE_DATA (cpu);
      ++ PROFILE_MODEL_LOAD_STALL_CYCLES (p);
      if (TRACE_INSN_P (cpu))
	cgen_trace_printf (cpu, " ; Load stall.");
      return 1;
    }
  else
    return 0;
}

int
nios16bf_model_nios16_1_u_exec (SIM_CPU *cpu, const IDESC *idesc,
			    int unit_num, int referenced)
{
  int cycles = idesc->timing->units[unit_num].done;
  return cycles;
}



#endif /* WITH_PROFILE_MODEL_P */
