/* nios32 trap support
   Copyright (C) 1999 Cygnus Solutions.  */

#define WANT_CPU
#define WANT_CPU_NIOS32BF

#include "sim-main.h"
#include "targ-vals.h"
#include "cgen-mem.h"
#include "cgen-ops.h"
#include "trap32.h"

static int nios32bf_spill_counter = 0;
static void nios32bf_shift_windows_up (SIM_CPU *current_cpu);

/* Indicate a window overflow has occured.  */

void
nios32bf_window_overflow (SIM_CPU *cpu, IADDR pc)
{
  SIM_DESC sd = CPU_STATE (cpu);

  if (STATE_ENVIRONMENT (sd) == OPERATING_ENVIRONMENT)
    nios32bf_hw_trap (cpu, pc, TRAP_WINDOW_OVERFLOW);
  else
    nios32bf_hw_trap (cpu, pc, TRAP_SIM_SPILL);
}

/* Indicate a window underflow has occured.  */

void
nios32bf_window_underflow (SIM_CPU *cpu, IADDR pc)
{
  SIM_DESC sd = CPU_STATE (cpu);

  if (STATE_ENVIRONMENT (sd) == OPERATING_ENVIRONMENT)
    nios32bf_hw_trap (cpu, pc, TRAP_WINDOW_UNDERFLOW);
  else
    nios32bf_hw_trap (cpu, pc, TRAP_SIM_FILL);
}

void
nios32bf_invalid_insn (SIM_CPU * cpu, IADDR pc)
{
  nios32bf_hw_trap (cpu, pc, TRAP_ILLEGAL_INSN);
}

void
nios32bf_core_signal (SIM_DESC sd, SIM_CPU *cpu, sim_cia pc,
		     unsigned int map, int nr_bytes, address_word addr,
		     transfer_type transfer, sim_core_signals sig)
{
  nios32bf_hw_trap (cpu, pc,
		   map == exec_map
		   ? TRAP_INSTRUCTION_ACCESS
		   : TRAP_DATA_ACCESS);
}

/* Handle hardware generated traps when --environment=operating.  */

static void
nios32bf_hw_trap_oper (SIM_CPU *current_cpu, IADDR pc, TRAP_TYPE trap)
{
  SIM_DESC sd = CPU_STATE (current_cpu);
  IADDR new_pc = VECBASE_ADDR + (trap << 1);

  nios32bf_h_m32_ctl_set (current_cpu, H_CTL_ISTATUS, 
                       nios32bf_h_status_get (current_cpu));
  nios32bf_h_tebit_set (current_cpu, 0);
  nios32bf_h_ipri_set (current_cpu, trap);
  nios32bf_h_cwp_set (current_cpu, 
                       NEXT_WIN (nios32bf_h_cwp_get (current_cpu)));
  nios32bf_h_m32_gr_set (current_cpu, H_GR_LINK, 
                          pc + 2);

  nios32bf_h_pc_set (current_cpu, new_pc);

  sim_engine_restart (CPU_STATE (current_cpu), current_cpu, NULL, new_pc);
}

/* when spilling, shift all windows up by one to leave space at bottom */
static void
nios32bf_shift_windows_up (SIM_CPU *current_cpu)
{
  HI wvalid = nios32bf_h_m32_ctl_get (current_cpu, H_CTL_WVALID);
  int hi_window = HI_WINDOW (wvalid);
  int lo_window = LO_WINDOW (wvalid);
  int i;

  /* shift each register window up by one..quicker to do this via 
     memcpy loop than overlapping memmove */
  for (i = hi_window; i > lo_window; --i)
    {
      memcpy (&current_cpu->win_regs[i],&current_cpu->win_regs[i-1], 16 * sizeof(SI));
    }
}
      
/* Handle hardware generated traps when --environment=user.  */

static void
nios32bf_hw_trap_user (SIM_CPU *current_cpu, IADDR pc, TRAP_TYPE trap)
{
  SIM_DESC sd = CPU_STATE (current_cpu);

  switch (trap)
    {
    case TRAP_SIM_SPILL :
      /* The LO_LIMIT has been exceeded. */
      {
	UQI win = nios32bf_h_cwp_get (current_cpu);
	HI wvalid = nios32bf_h_m32_ctl_get (current_cpu, H_CTL_WVALID);
	int hi_window = HI_WINDOW(wvalid);
	
	++nios32bf_spill_counter;

	/* There's no need to flush `current_regs' here, `hi_window' can't
	   refer to it.  */
	nios32bf_flush_regwin (current_cpu, pc, hi_window, 0 /* error ok (?) */);

	/* Shift all windows up by one. */
	nios32bf_shift_windows_up (current_cpu);

        CPU (h_cwp) = win + 1;
	return;
      }

    case TRAP_SIM_FILL :
      /* The HI_LIMIT has been exceeded */
      {
	UQI win = (nios32bf_h_cwp_get (current_cpu));
	HI wvalid = nios32bf_h_m32_ctl_get (current_cpu, H_CTL_WVALID);
	int hi_window = HI_WINDOW(wvalid);

	if (--nios32bf_spill_counter < 0)
	  {
	    sim_io_eprintf (sd, "Register window underflow has occurred\n");
	    sim_engine_halt (sd, current_cpu, NULL, pc, sim_stopped, SIM_SIGABRT);
	  }
	/* Load caller's caller's window.
	   There's no need to flush `current_regs' as `hi_window' can't
	   refer to it.  */
	nios32bf_load_regwin (current_cpu, pc, hi_window);

	CPU (h_cwp) = win - 1;
	return;
      }
    }

  sim_io_eprintf (sd, "Received trap %d\n", trap);
  sim_engine_halt (sd, current_cpu, NULL, pc, sim_stopped, SIM_SIGABRT);
}

/* Handle hardware generated traps.  */

void
nios32bf_hw_trap (SIM_CPU *current_cpu, IADDR pc, TRAP_TYPE trap)
{
  SIM_DESC sd = CPU_STATE (current_cpu);

  if (STATE_ENVIRONMENT (sd) == OPERATING_ENVIRONMENT)
    nios32bf_hw_trap_oper (current_cpu, pc, trap);
  else
    nios32bf_hw_trap_user (current_cpu, pc, trap);
}

/* Handle the trap insn when --environment=operating.  */

static IADDR
nios32bf_sw_trap_oper (SIM_CPU *current_cpu, IADDR pc, UQI trap)
{
  SIM_DESC sd = CPU_STATE (current_cpu);
  IADDR trap_addr;

  trap += TRAP_INSTRUCTION;
  trap_addr = VECBASE_ADDR + (trap << 1);

  /* ??? Quick hack to have breakpoints work with gdb+"target sim" until
     other things are working.  */
  if (trap == TRAP_BREAKPOINT)
    sim_engine_halt (sd, current_cpu, NULL, pc, sim_stopped, SIM_SIGTRAP);

  if (! nios32bf_h_tebit_get (current_cpu) || trap >= nios32bf_h_ipri_get (current_cpu))
    {
      int o0 = nios32bf_h_m32_gr_get (current_cpu, H_GR_O0);
      int o1 = nios32bf_h_m32_gr_get (current_cpu, H_GR_O1);

      if (o1 == LIBGLOSS_EXIT_MAGIC)
	sim_engine_halt (sd, current_cpu, NULL, pc, sim_exited, o0);
      else
	{
	  sim_io_eprintf (sd, "Unexpected program termination, pc=0x%x\n",
			  (int) pc);
	  sim_engine_halt (sd, current_cpu, NULL, pc,
			   sim_signalled, SIM_SIGABRT);
	}
    }

  nios32bf_h_tebit_set (current_cpu, 0);
  nios32bf_h_m32_ctl_set (current_cpu, H_CTL_ISTATUS,
                       nios32bf_h_status_get (current_cpu));
  nios32bf_h_cwp_set (current_cpu, 
                       NEXT_WIN (nios32bf_h_cwp_get (current_cpu)));
  nios32bf_h_ipri_set (current_cpu, trap);
  nios32bf_h_m32_gr_set (current_cpu, H_GR_LINK, 
                          pc + 2);

  return trap_addr;
}

/* Subroutine of nios32bf_do_trap to read target memory.  */

static int
syscall_read_mem (host_callback *cb, CB_SYSCALL *sc,
		  unsigned long taddr, char *buf, int bytes)
{
  SIM_DESC sd = (SIM_DESC) sc->p1;
  SIM_CPU *cpu = (SIM_CPU *) sc->p2;

  return sim_core_read_buffer (sd, cpu, read_map, buf, taddr, bytes);
}

/* Subroutine of nios32bf_do_trap to write target memory.  */

static int
syscall_write_mem (host_callback *cb, CB_SYSCALL *sc,
		   unsigned long taddr, const char *buf, int bytes)
{
  SIM_DESC sd = (SIM_DESC) sc->p1;
  SIM_CPU *cpu = (SIM_CPU *) sc->p2;

  return sim_core_write_buffer (sd, cpu, write_map, buf, taddr, bytes);
}

/* Handle the trap insn when --environment=user.  */

static IADDR
nios32bf_sw_trap_user (SIM_CPU *current_cpu, IADDR pc, UQI trap)
{
  SIM_DESC sd = CPU_STATE (current_cpu);
  IADDR new_pc = pc + 2;

  trap += TRAP_INSTRUCTION;

  switch (trap)
    {
    case TRAP_SYSCALL :
      /* FIXME: Later make trap number runtime selectable.  */
      {
	CB_SYSCALL s;

	CB_SYSCALL_INIT (&s);
	s.func = nios32bf_h_m32_gr_get (current_cpu, 8);
	s.arg1 = nios32bf_h_m32_gr_get (current_cpu, 9);
	s.arg2 = nios32bf_h_m32_gr_get (current_cpu, 10);
	s.arg3 = nios32bf_h_m32_gr_get (current_cpu, 11);
	if (s.func == TARGET_SYS_exit)
	  {
	    /* Tell sim_resume program called exit().  */
	    sim_engine_halt (sd, current_cpu, NULL, pc, sim_exited, s.arg1);
	  }
	s.p1 = (PTR) sd;
	s.p2 = (PTR) current_cpu;
	s.read_mem = syscall_read_mem;
	s.write_mem = syscall_write_mem;
	cb_syscall (STATE_CALLBACK (CPU_STATE (current_cpu)), &s);
	nios32bf_h_m32_gr_set (current_cpu, 10, s.errcode);
	nios32bf_h_m32_gr_set (current_cpu, 8, s.result);
	nios32bf_h_m32_gr_set (current_cpu, 9, s.result2);
	break;
      }

    case TRAP_BREAKPOINT :
      sim_engine_halt (sd, current_cpu, NULL, pc, sim_stopped, SIM_SIGTRAP);

    case TRAP_DIVIDE_0 :
      sim_engine_halt (sd, current_cpu, NULL, pc, sim_stopped, SIM_SIGFPE);

    case TRAP_FLUSH_REGWIN :
      nios32bf_flush_regwins (current_cpu, pc, 0 /* error ok */);
      break;

    default :
      sim_io_eprintf (sd, "Unsupported trap %d\n", trap);
      sim_engine_halt (sd, current_cpu, NULL, pc, sim_stopped, SIM_SIGILL);
    }

  return new_pc;
}

/* Called from the semantic code to handle the trap instruction.  */

IADDR
nios32bf_sw_trap (SIM_CPU *current_cpu, IADDR pc, UQI trap)
{
  SIM_DESC sd = CPU_STATE (current_cpu);
  IADDR new_pc;

  if (STATE_ENVIRONMENT (sd) == OPERATING_ENVIRONMENT)
    new_pc = nios32bf_sw_trap_oper (current_cpu, pc, trap);
  else
    new_pc = nios32bf_sw_trap_user (current_cpu, pc, trap);
  return new_pc;
}

/* Handle the tret insn.  */

IADDR
nios32bf_do_tret (SIM_CPU *current_cpu, IADDR pc, IADDR new_pc)
{
  nios32bf_h_m32_ctl_set (current_cpu, H_CTL_STATUS, 
                       nios32bf_h_m32_ctl_get (current_cpu, H_CTL_ISTATUS));

  if (TRACE_INSN_P (current_cpu)) /* FIXME */
    {
      trace_result (current_cpu, "sp", 'x', nios32bf_h_m32_gr_get (current_cpu, H_GR_SP));
      trace_result (current_cpu, "fp", 'x', nios32bf_h_m32_gr_get (current_cpu, H_GR_FP));
      trace_result (current_cpu, "cwp", 'x', nios32bf_h_cwp_get (current_cpu));
    }

  return new_pc << 1;
}
