	.equ TRAP_SYS,0
	.macro start
	.text
failmsg:
	.ascii "fail\n"
passmsg:
	.ascii "pass\n"
	.text
	.global _start
_start:
	pfx %hi(0xbf00)
	movi %sp,%lo(0xbf00)
	.endm

; Exit with return code
	.macro exit rc
        movi %r9,\rc
        movi %r8,#1
        trap TRAP_SYS
	nop
	.endm

; Pass the test case
	.macro pass
	clr %r12
        movi %r11,#5
        pfx %hi(passmsg)
        movi %r10,%lo(passmsg)
        movi %r9,#1
        movi %r8,#5
        trap TRAP_SYS
	nop
	exit  #0
	.endm

; Fail the testcase
	.macro fail
	clr %r12
        movi %r11,#5
        pfx %hi(failmsg)
        movi %r10,%lo(failmsg)
        movi %r9,#1
        movi %r8,#5
        trap TRAP_SYS
	nop
	exit  #1
	.endm

; Load an immediate value into a general register
	.macro mvi_h_gr reg val
	.if (\val >= 0) && (\val <= 0x1f)
	movi \reg,\val
	.else
	pfx %hi(\val)
	movi \reg,%lo(\val)
	.if (\val > 0xffff) || (\val < 0)
	pfx %xhi(\val)
	movhi \reg,%xlo(\val)
	.endif
	.endif
	.endm

; Load an immediate value into a control register
	.macro mvi_h_ctl ctl_no val
	.if (\ctl_no == 3)
	pfx %hi(mvi_ctl\@@h)
	movi %r31,%lo(mvi_ctl\@@h)
	.endif
	.if (\val >= 0) && (\val <= 0x1f)
	movi %r1,\val
	.else
	pfx %hi(\val)
	movi %r1,%lo(\val)
	.endif
	pfx \ctl_no
	wrctl %r1
mvi_ctl\@:
	.endm

; Load a control register from another register
	.macro mvr_h_ctl ctl_no reg
	mov %r1,\reg
	pfx \ctl_no
	wrctl %r1
	.endm

; Load a register from a control register
	.macro ldr_h_ctl ctl_no reg
	pfx \ctl_no
	rdctl %r1
	mov \reg,%r1
	.endm

; Load a general register with an address
	.macro	set_addr_h_gr reg val
	pfx %hi(\val)
	movi \reg,%lo(\val)
	.endm

; Load a general register into another general register
	.macro mvr_h_gr targ src
	mov \targ,\src
	.endm

; Store an immediate into a word in memory
	.macro mvi_h_mem addr val
	mvi_h_gr  %r1,\val
	mvr_h_mem \addr,%r1
	.endm

; Store a register into a word in memory
	.macro mvr_h_mem addr reg
	st [\addr],\reg
	.endm

; Store the current ps on the stack
	.macro save_ps
	rdctl %r1
	st [%sp],%r1
	.endm

; Load a word value from memory
	.macro ldmem_h_gr reg addr
	pfx %hi(\addr)
	movi %r1,%lo(\addr)
	ld \reg,[%r1]
	.endm

; Add 2 general registers
	.macro add_h_gr reg1 reg2
	add \reg1,\reg2
	.endm

; Increment a register by an immediate
	.macro inci_h_gr reg inc
	mvi_h_gr %r1,\inc
	add \reg,%r1
	.endm

; Test the value of an immediate against a general register
	.macro test_h_gr reg val
	.if (\val >= 0) && (\val <= 0x1f)
	cmpi \reg,\val
	.else
	.if (\val > 0xffff) || (\val < 0)
	pfx %hi(\val)
	movi %r2,%lo(\val)
	pfx %xhi(\val)
	movhi %r2,%xlo(\val)
	cmp \reg,%r2
	.else
	pfx %hi(\val)
	cmpi \reg,%lo(\val)
	.endif
	.endif
	beq test_gr\@
	nop
	fail
test_gr\@:
	.endm

; Load an immediate value into a control register
	.macro test_h_ctl ctl_no val
	pfx \ctl_no
	rdctl %r1
	.if (\val >= 0) && (\val <= 0x1f)
	cmpi %r1,\val
	.else
	pfx %hi(\val)
	cmpi %r1,%lo(\val)
	.endif
	beq test_ctl\@
	nop
	fail
test_ctl\@:
	.endm

; compare two general registers
	.macro testr_h_gr reg1 reg2
	cmp \reg1,\reg2
	beq testr_gr\@
	nop
	fail
testr_gr\@:
	.endm

; Test the link register for a specified address
	.macro	test_link_addr addr
	pfx %hi(\addr@h)
	movi %r1,%lo(\addr@h)
	addi %r1,2
	testr_h_gr %r1,%r15
	.endm

; Compare an immediate with word in memory
	.macro test_h_mem addr val
	ldmem_h_gr %r5,\addr
	test_h_gr %r5,\val
	.endm

; Compare an immediate with a byte in memory
        .macro test_h_memb addr val
        ldmem_h_gr %r5,\addr
        set_addr_h_gr %r6,\addr
        ext8d %r5,%r6
        test_h_gr %r5,\val
        .endm


; Compare an address with a word in memory
	.macro test_h_mem_a addr addr2
	ldmem_h_gr %r5,\addr
	set_addr_h_gr %r6,\addr2
	testr_h_gr %r5,%r6
	.endm

; Compare a general register with word in memory
	.macro testr_h_mem reg addr
	ldmem_h_gr %r5,\addr
	testr_h_gr %r5,\reg
	.endm

; Set the condition codes
	.macro set_cc mask
	rdctl %r1
	pfx 0
	andn %r1,0xf
	pfx 0
	or %r1,\mask
	wrctl %r1
	.endm

; Set the current window pointer
	.macro set_cwp cwp
	rdctl %r1
	pfx %hi(0x1f0)
	movi %r2,%lo(0x1f0)
	andn %r1,%r2
	movi  %r2,\cwp
	lsli %r2,4
	or %r1,%r2
	wrctl %r1
	.endm

; Set the trap-enable bit
	.macro set_te val
	rdctl %r1
	pfx %hi(0x7fff)
	and %r1,%lo(0x7fff)
	.if (\val == 1)
	pfx %hi(0x8000)
	or %r1,%lo(0x8000)
	.endif
	wrctl %r1
	.endm

; Test the current window pointer
	.macro test_cwp cwp
	rdctl %r1
	pfx %hi(0x1f0)
	movi %r2,%lo(0x1f0)
	and %r1,%r2
	lsri %r1,4
	cmpi %r1,\cwp
	beq test_cwp\@
	nop
	fail
test_cwp\@:
	.endm

; Test the condition codes
	.macro test_cc N V Z C
	.if (\N == 1)
	bp fail\@
	nop
	.else
	bn fail\@
	nop
	.endif
	.if (\Z == 1)
	bne fail\@
	nop
	.else
	beq fail\@
	nop
	.endif
	.if (\V == 1)
	bnv fail\@
	nop
	.else
	bv fail\@
	nop
	.endif
	.if (\C == 1)
	bnc fail\@
	nop
	.else
	bc fail\@
	nop
	.endif
	br test_cc\@
	nop
fail\@:
	fail
test_cc\@:
	.endm

; Ensure branch taken with delay slot
	.macro take_branch_d opcode val
	\opcode take_brd\@
	mvi_h_gr %r1,\val
	fail
take_brd\@:
	test_h_gr %r1,\val
	.endm

; Ensure branch not taken
	.macro no_branch opcode
	\opcode no_brf\@
	nop
	br	no_brs\@
	nop
no_brf\@:
	fail
no_brs\@:
	.endm

	.macro no_branch_d opcode val
	\opcode no_brdf\@
	movi	%r1,\val
	br	no_brds\@
	nop
no_brdf\@:
	fail
no_brds\@:
	test_h_gr %r1,\val
	.endm


